addpath('C:\Code\Matlab\Master\Spring')

%% OD-probe
name_file = '20230208_None_None_Glucose_B2.mat';
exp_data = load(name_file, 'VarMem');   % OD-calibration

NIR_raw = exp_data.VarMem.Instrument.NIR_smooth; % use smoothed data
t_NIR = exp_data.VarMem.Instrument.t/3600;       % to hours
%% Manual sampling
filename = 'Lab_results.xlsx';

lab_result_table = readtable(filename, 'Sheet','Results_3ml_B2');

lag = 6; % old: 5   % How many time points in lag phase !! for raw data we have 3*measuremets
stat = 12;  %old 12% At what time point the stationary phase starts

lab_results = table2array(lab_result_table(:,3:10)); % 1:13, 3:10 for A2

dryweight = [lab_results(:,5:7)-lab_results(:,2:4)]; % Triplicates

time = lab_results(:,8); 

cell_density = dryweight*1000/3; % g/L, 3 mL samples -> x g in 3mL  -> y g/L

mean_cell_density = mean(cell_density,2);
std_cell_density = std(cell_density,0, 2);

%mean_OD = mean(OD,2);
%std_OD = std(OD,0, 2);

%% Get corresponding NIR data
NIR = zeros(1, length(time));
time_N = time;
for i = 1:(length(mean_cell_density)-1)
    k = find(t_NIR <= time_N(1), 1, 'last');
    NIR(i) = NIR_raw(k);
    time_N = time_N(2:end);
    disp(k)
    disp(time_N)
    
end
% last iteration
k = find(t_NIR <= time_N(1), 1, 'last');
NIR(end) = NIR_raw(k);

%% Regression
% R_squared <=> ordinary R^2
% lag phase
[intercept_lag, slope_lag, x_lag, y_lag, RMSE_lag, R_sq_lag] = regression( NIR(1:lag), mean_cell_density(1:lag));
% log phase
%[intercept_log, slope_log, x_log, y_log, RMSE_log, R_sq_log] = regression( NIR(lag+1:stat-1), mean_cell_density(lag+1:stat-1));
[intercept_log, slope_log, x_log, y_log, RMSE_log, R_sq_log] = regression( NIR(lag:stat-1), mean_cell_density(lag:stat-1));
% stationary
[intercept_sta, slope_sta, x_sta, y_sta, RMSE_sta, R_sq_sta] = regression( NIR(stat:end), mean_cell_density(stat:end));
%% Plot

color = colorHex();
% Plot
figure(1)
hold on
rectangle(Position=[0.1,0, 7.2, 1.5], FaceColor=[0.9 0.9 0.9], EdgeColor=[0.9 0.9 0.9])
rectangle(Position=[7.2,0, 16-7.2, 1.5], FaceColor='#DFDFFF', EdgeColor=[0.9 0.9 0.9])
rectangle(Position=[16,0, 10, 1.5], FaceColor='#FFBCBC', EdgeColor=[0.9 0.9 0.9])
set(gca, "Layer", "top")
text(1.8, 0.45, 'Lag phase')
text(10, 0.45, 'Log phase')
text(16.5, 0.45, 'Stationary phase')
colororder({'k','k'})
plot(time,NIR, 'x', "Color", color.b)
xlim([0 22])
ylim([0 1.25])
xlabel("Time [h]")
ylabel("NIR")
yyaxis right
ylim([0 28])
plot(time,cell_density, 'x', 'Color', color.r )
ylabel('CDW [g/L]')
%errorbar(time,OD, std_OD,'x')
legend('NIR', 'CDW', 'Location', 'SouthEast')
title('NIR & CDW measurements over time')


figure(2)
hold on
plot( NIR(1:lag), mean_cell_density(1:lag),'x', 'Color', color.b)
%plot( NIR(lag+1:stat-1),mean_cell_density(lag+1:stat-1),  'x', 'Color', color.r)
plot( NIR(lag:stat-1),mean_cell_density(lag:stat-1),  'x', 'Color', color.r)
plot( NIR(stat:end),mean_cell_density(stat:end), 'x', 'Color', color.p)
plot(x_lag,y_lag, 'Color', color.b)
plot(x_log,y_log, 'Color', color.r)
plot(x_sta,y_sta, 'Color', color.p)
ylabel("Dryweight [g/L]")
xlabel("NIR")
errorbar(NIR, mean_cell_density, std_cell_density, 'LineStyle', 'none')
linear_lag = strcat('NIR = ',num2str(intercept_lag), ' + ', num2str(slope_lag), 'x');
linear_log = strcat('NIR = ',num2str(intercept_log), ' + ', num2str(slope_log), 'x');
linear_sta = strcat('NIR = ',num2str(intercept_sta), ' + ', num2str(slope_sta), 'x');
%errorbar(mean_cell_density,mean_OD, std_cell_density, std_OD, 'both', 'x')
legend('Lag', 'Log', 'Stationary', linear_lag, linear_log, linear_sta, 'Location', 'NorthWest') % SouthWest

% Errors
str = {[strcat('RMSE = [', string(RMSE_lag), ' , ', string(RMSE_log), ' , ', string(RMSE_sta), ']' )], [strcat('R^2 = [' , string(R_sq_lag), ' , ',string(R_sq_log), ' , ', string(R_sq_sta), ']')]};
%annotation('textbox',  [0.17 0.66 0.3 0.25] , 'String', str, 'FitBoxToText','on', 'FontSize', 9)
annotation('textbox',  [0.525 0.003 0.3 0.25] , 'String', str, 'FitBoxToText','on', 'FontSize', 9, 'LineWidth', 0.4)

title('NIR-cell dry weight calibration curve')

% NIR to CDW 
%a = NIR(1:lag)*slope_lag + intercept_lag;
a = NIR(1:lag+3)*slope_lag + intercept_lag;
%b = NIR(lag+1:stat-1)*slope_log + intercept_log;
b = NIR(lag:end)*slope_log + intercept_log;
figure(3)
hold on
%plot(time(1:lag),a, "Color", color.b)
plot(time(1:lag+3),a, "Color", color.b)
%plot(time(lag+1:stat-1),b, "Color", color.b)
plot(time(lag:end),b, "Color", color.r)
ylim([0 28])
%ylabel("NIR")
xlabel("Time [h]")
%yyaxis right
plot(time,cell_density, 'x', 'Color', [0.7 0.7 0.7] )
ylabel("Dryweight [g/L]")
legend("NIR lag", "NIR log", "CDW exp")
title('NIR to CDW comparison')
%% cftool
%cftool()


%CDW_NIR =  a1*sin(b1*NIR+c1) + a2*sin(b2*NIR+c2);
       a1 =       -15212.545  ;
       b1 =       2.0532  ;
       c1 =       2.5893  ;
       a2 =       15197.4245  ;
       b2 =       2.055  ;
       c2 =       2.5887 ;
%% Function
function [intercept, slope, x_reg, y_reg, RMSE, R_squared] = regression(x, y)
% x: Cell_density
% y: OD
x_points = length(x);
% Linear regression
linreg = fitlm(x, y)

intercept = linreg.Coefficients.Estimate(1);
slope = linreg.Coefficients.Estimate(2);
if x_points > 5
    x_reg = linspace(x(1)-0.01, x(end)+0.08, 20);%x_points);
else
    x_reg = linspace(x(1)-0.15, x(end)+0.1, 20);%x_points);
end
y_reg = slope*x_reg + intercept;

RMSE = linreg.RMSE;
R_squared = linreg.Rsquared.Ordinary;

% Simple regression
%X = [ones(length(cell_density),1) cell_density];
%b = X\OD;
%yCalc = X*b;
end
