addpath('C:\Code\Matlab\Master\Spring')
addpath('C:\Code\Matlab\Master\Spring\Results')
%% OD-probe
name_file = '20230208_None_None_Glucose_B2.mat';
exp_data = load(name_file, 'VarMem');   % OD-calibration

NIR_raw = exp_data.VarMem.Instrument.NIR_smooth; % use smoothed data
t_NIR = exp_data.VarMem.Instrument.t/3600;       % to hours
%% Manual sampling
filename = 'Lab_results.xlsx';

lab_result_table = readtable(filename, 'Sheet','Results_3ml_B2');

lab_results = table2array(lab_result_table(:,3:10)); % 1:13, 3:10 for A2

dryweight = [lab_results(:,5:7)-lab_results(:,2:4)]; % Triplicates

time = lab_results(:,8); 

cell_density = dryweight*1000/3; % g/L, 3 mL samples -> x g in 3mL  -> y g/L

mean_cell_density = mean(cell_density,2);
std_cell_density = std(cell_density,0, 2);

%% Get corresponding NIR data
NIR = zeros(1, length(time));
time_N = time;
for i = 1:(length(mean_cell_density)-1)
    k = find(t_NIR <= time_N(1), 1, 'last');
    NIR(i) = NIR_raw(k);
    time_N = time_N(2:end);
    %disp(k)
    %disp(time_N)  
end
% last iteration
k = find(t_NIR <= time_N(1), 1, 'last');
NIR(end) = NIR_raw(k);

createFit(NIR', mean_cell_density, std_cell_density, time, NIR_raw, t_NIR);



function [fitresult, gof] = createFit(NIR, mean_cell_density, std_cell_density, time, NIR_raw, t_NIR)
%CREATEFIT(NIR,MEAN_CELL_DENSITY)
%  Create a fit.
%
%  Data for 'NIR CDW' fit:
%      X Input : NIR
%      Y Output: mean_cell_density
%  Output:
%      fitresult : a fit object representing the fit.
%      gof : structure with goodness-of fit info.
%
%  See also FIT, CFIT, SFIT.

%  Auto-generated by MATLAB on 28-Feb-2023 08:23:08


%% Fit: 'NIR CDW'.
[xData, yData] = prepareCurveData( NIR, mean_cell_density );

% Set up fittype and options.
ft = fittype( 'sin2' );
opts = fitoptions( 'Method', 'NonlinearLeastSquares' );
opts.Display = 'Off';
opts.Lower = [-Inf 0 0 -Inf 0 -Inf];
opts.StartPoint = [15.6104359296021 2.69984270546183 0.9 10.2340847941493 5.39968541092366 0.9];
% Fit model to data.
[fitresult, gof] = fit( xData, yData, ft, opts );

% Fitresult
a1 = fitresult.a1;  b1= fitresult.b1;  c1 = fitresult.c1;  
a2 = fitresult.a2;  b2=fitresult.b2;   c2 = fitresult.c2;

c = newline;
regression = strcat('CDW = ',num2str(a1), ' sin( ', num2str(b1), 'NIR  +', num2str(c1), ') + ' , num2str(a2), 'sin( ', num2str(b2), 'NIR + ', num2str(c2), ')');

disp(regression)
NIR_CDW = a1*sin(b1*NIR+c1) + a2*sin(b2*NIR+c2);

% Plot fit with data.
color = colorHex();
figure(1)
%figure( 'Name', 'NIR-CDW calibration curve' );
plot( fitresult, xData, yData);
%errorbar(NIR, mean_cell_density, std_cell_density, 'LineStyle', 'none')
legend( 'Experimental data', 'Fit', 'Location', 'SouthEast', 'Interpreter', 'none' );
legend boxoff  
% Label axes
xlabel( 'NIR', 'Interpreter', 'none' );
ylabel( 'CDW [g/L]', 'Interpreter', 'none' );
ylim([0 25])
title('NIR-CDW calibration curve')

% Errors
str = {[strcat('R^2 = ', string(gof.rsquare))], [strcat('RMSE = ' , string(gof.rmse))]};
annotation('textbox',  [0.15 0.56 0.3 0.25] , 'String', str, 'FitBoxToText','on', 'FontSize', 9, 'LineWidth', 0.4, 'LineStyle', '--')
annotation('textbox',  [0.15 0.66 0.3 0.25] , 'String', regression, 'FitBoxToText','on', 'FontSize', 9, 'LineWidth', 0.4, 'LineStyle', '--')


drawnow

figure(2)
box off
hold on
colororder({'k','k'})
%figure( 'Name', 'NIR to CDW comparison' );
plot(time, NIR_CDW,'-o', 'Color', color.p, 'MarkerEdgeColor', color.p)
%scatter(time, NIR_CDW, 'o', 'Color', color.p)
ylim([0 30])
xlabel("Time [h]")
plot(time, mean_cell_density, 'x', 'Color', color.r)
errorbar(time, mean_cell_density, std_cell_density, 'LineStyle', 'none', 'Color', color.o)
ylabel("Dryweight [g/L]")
yyaxis right
plot(time, NIR, '+', 'MarkerEdgeColor', color.b)
plot( t_NIR, NIR_raw, 'Color',color.p)
%ylim([0 0.1])
ylabel("NIR")
title("NIR to CDW comparison")
legend("NIR to CDW", "Mean CDW", '', 'NIR', 'All NIR','Location', 'SouthEast')
legend boxoff  
annotation('textbox',  [0.15 0.65 0.3 0.25] , 'String', regression, 'FitBoxToText','on', 'FontSize', 9, 'LineWidth', 0.4, 'LineStyle', '--') %'LineWidth', 0.4
drawnow

end

