addpath('C:\Code\Matlab\Master\Spring')

filename = 'Lab_results.xlsx';

lab_result_table = readtable(filename, 'Sheet','Results_3ml_B2'); % B2

lag = 5;    % How many time points in lag phase !! for raw data we have 3*measuremets
stat = 12;  % At what time point the stationary phase starts

lab_results = table2array(lab_result_table(:,3:10)); % 1:13, 3:10 for A2

dryweight = [lab_results(:,5:7)-lab_results(:,2:4)]; % Triplicates

OD = lab_results(:,1);

time = lab_results(:,8); 

cell_density = dryweight*1000/3; % g/L, 3 mL samples!

mean_cell_density = mean(cell_density,2);
std_cell_density = std(cell_density,0, 2);


%% Regression
% R_squared <=> ordinary R^2
% lag phase
[intercept_lag, slope_lag, x_lag, y_lag, RMSE_lag, R_sq_lag] = regression( OD(1:lag), mean_cell_density(1:lag));
% log phase
[intercept_log, slope_log, x_log, y_log, RMSE_log, R_sq_log] = regression( OD(lag+1:stat-1), mean_cell_density(lag+1:stat-1));
% stationary
[intercept_sta, slope_sta, x_sta, y_sta, RMSE_sta, R_sq_sta] = regression( OD(stat:end), mean_cell_density(stat:end));
%% Plot

color = colorHex();
% Plot
figure(1)
hold on
rectangle(Position=[0.1,0.5, 6.5, 90], FaceColor=[0.9 0.9 0.9], EdgeColor=[0.9 0.9 0.9])
rectangle(Position=[6.5,0.5, 16-6.5, 90], FaceColor='#DFDFFF', EdgeColor=[0.9 0.9 0.9])
rectangle(Position=[16,0.5, 10, 90], FaceColor='#FFBCBC', EdgeColor=[0.9 0.9 0.9])
set(gca, "Layer", "top")
text(1.5, 30, 'Lag phase')
text(9.5, 30, 'Log phase')
text(16.5, 30, 'Stationary phase')
colororder({'k','k'})
plot(time,OD, 'x', "Color", color.b)
ylim([0 85])
%ylim([0 9])
xlim([0 22])
xlabel("Time [h]")
ylabel("OD600")
yyaxis right
ylim([0 26])
plot(time,cell_density, 'x', 'Color', color.r )
ylabel('CDW [g/L]')
%errorbar(time,OD, std_OD,'x')
legend('OD', 'CDW', 'Location', 'SouthEast')
title('OD & CDW measurements over time')


figure(2)
hold on
plot( OD(1:lag), mean_cell_density(1:lag),'x', 'Color', color.b)
plot( OD(lag+1:stat-1),mean_cell_density(lag+1:stat-1),  'x', 'Color', color.r)
plot(  OD(stat:end),mean_cell_density(stat:end), 'x', 'Color', color.p)
plot(x_lag,y_lag, 'Color', color.b)
plot(x_log,y_log, 'Color', color.r)
plot(x_sta,y_sta, 'Color', color.p)
ylabel("Dryweight [g/L]")
xlabel("OD600")
errorbar(OD, mean_cell_density, std_cell_density, 'LineStyle', 'none')
linear_lag = strcat('CDW = ',num2str(intercept_lag), ' + ', num2str(slope_lag), 'x');
linear_log = strcat('CDW = ',num2str(intercept_log), ' + ', num2str(slope_log), 'x');
linear_sta = strcat('CDW = ',num2str(intercept_sta), ' + ', num2str(slope_sta), 'x');
%errorbar(mean_cell_density,mean_OD, std_cell_density, std_OD, 'both', 'x')
legend('Lag', 'Log', 'Stationary', linear_lag, linear_log, linear_sta, 'Location', 'SouthEast')

% Errors
str = {[strcat('RMSE = [', string(RMSE_lag), ' , ', string(RMSE_log), ' , ', string(RMSE_sta), ']' )], [strcat('R^2 = [' , string(R_sq_lag), ' , ',string(R_sq_log), ' , ', string(R_sq_sta), ']')]};
annotation('textbox',  [0.15 0.65 0.3 0.25] , 'String', str, 'FitBoxToText','on', 'FontSize', 9, 'LineWidth', 0.4)

title('OD-cell dry weight calibration curve')


%% Function
function [intercept, slope, x_reg, y_reg, RMSE, R_squared] = regression(x, y)
% x: Cell_density
% y: OD
x_points = length(x);
% Linear regression
linreg = fitlm(x, y)

intercept = linreg.Coefficients.Estimate(1);
slope = linreg.Coefficients.Estimate(2);
x_reg = linspace(x(1)-4, x(end)+4, 20);%x_points);
y_reg = slope*x_reg + intercept;

RMSE = linreg.RMSE;
R_squared = linreg.Rsquared.Ordinary;

% Simple regression
%X = [ones(length(cell_density),1) cell_density];
%b = X\OD;
%yCalc = X*b;
end
%% cftool
%cftool()