%% General function to call OPC and to start the biochemical process
function [savefile]=main_meas_v1_CDW_calibration_A2(simulation)
% change later to Runme.m
% the main function for control 
% everything is saved in SAVEFILE in the EXPERIMENTS folder.
% "par.filter_step" determined how often (how many while loop) before the EKF is
% running. for example, if step = 5, the EKF will run calculation every
% iterationTime * 5, i.e. every fifth loop iteration. 
% default: Preferable step=1 (each time step)
% including reading the automativ HPLC measurement from the Chromeleon
% files in 'C:\Users\Public\Documents\Lucullus\import\chromeleon'
% The procedure check the time in the beginning of the experiments (clk)
% and only check for files created after this time. 
% The output is Instrument.HPLC.SugVal.Man/Xyl/Glu/Ara
% if you call the function with value, then simulation ==1 which implies
% not reading data from real-time

%% LOG CHANGE of version 5.
% This is the update version of main_meas_casadi_v2_3_MPC
% This code follows the biomass setpoint


%% LOG CHANGE V 2.1
% V 2.2
% Has a smoothing alogwrith 'Lowess' with window size 80 in the Instruent
% Anaysis function
% The instrument.NIR is the noisy measurements and the instrument.NIR_smooth 
% is after smoothing. instrument.CDWcu is calculated from the smoothing.
% the function InstrumentAnalysis has moved after GetInstrumentData, Before
% the HPLC beacuse the HPLC uses Intrument.CDWcu from measurements.

% V2.3
% Added VarMem.Instrument.HPLC.SampleIteration in order to be able to
% simulate offline EKF for analysis.
% VarMem.Instrument.HPLC.SampleIteration is the iteration in which a sample
% was taken.

% V2.5 
% New Cleaning procedure


% V3_1
% implemented sample database VarMem.Instrument.SampleDataBase that stores
% all the samples that were taken during the experiments. Then when new
% file is written to the folder, check its barcode, and compare it to the
% list of barcodes in the database. If finds new barcode, add it.
% HPLC measurements: check from the SamplesDataBase WHEN (timestamp) the
% barcode was taken and which bioreactor it is...

% V3.2
% fix bug with the date UTC and summmer winter time
% fixed bug with HPLC name measurement
% enabled full FLUSH with external pumps + checked that it is working
% fixed bug with the set point sugar levels 
% fixed bug with reading of the HPLC file values, also when ara xyl and man
% are not present it will work (they get nan values)
% fixed bug with reading from NUMERA value for clock and sample name
% TODO: reverse flush pump should start 4 min after the SDU port changes
% value in NUMERA

% CO2 is smoothed

% V3_4
% safeguard against reset of balance
    
%% CHECK:
% check whether the NIR and the cdw are correctly measured and defined.

%% Get feeding profile
% !OBS par.iterationtime set to 60s and par.dt set to 1/60 as below
disp('Get feed profile')
load('Values_feed.mat')
Values_feeding_profile = Values_feed;
time_profile = Values_feeding_profile.time;
feed_profile = Values_feeding_profile.feed;


%% Connect to the OPC server
if nargin==1
     %% For computer in-silico simulations, uncomment the next line:
 %   load ('Data_20210908_EKF_sim')
    %load ('tempNadi')
    simulation=1;
else
    simulation=0;
    %% For real OPC time use this option
    % Connect to OPC Server
    da = opcda('localhost', 'Securecell.OpcBridgeServer.DA.1'); 
    connect(da); 
    BioReactor = 'A2';
    [readGrpA2, writeGrpA2, indexesA2] = findReactorTags(BioReactor, da);
    %[readGrpHPLC,~, indexesHPLC] = findSampleInfo(reactor, da);
    data = read(readGrpA2);
end

close all


%% Initialize the code
% The following creates the file name to be saved at each iteration (saving
% the data), according to the current date
  

clk=clock;
clk_0=fix(clk);  
      % INITIAL clock time update, 
clk_now=clock;
clk_now=fix(clk_now);
%% Define the initial loop
i=1;
VarMem=  [];

if simulation <1
    % Initialize the memory    
    % TODO 
    
    iteration= i;
    
    par=GetParValues(BioReactor,[],clk_0);
    % Get the instrument information
    [Instrument0]=GetInstData(data,indexesA2,clk_now,par,iteration);
    par=GetParValues(BioReactor,Instrument0,clk_0);
    
    Instrument0.V = par.V0; %Initial reactor volume (user defined) 
   
    % Define the initial conditions
    x0= [par.V0 par.X0 par.S0 par.CO20];
    % Initializing for the first estimator

    %% BEGIN SANITY CHECK LIST!
   % sprintf('Running sannity check... Please Wait\n----------------------------------\n')
    % check the following in Instrument0
    % Instrument.Balance_A2
   % if isempty
    % Instrument.Balance_B2
    % Balance C
    % Balance D
    % Instrument.NIR
    % Instrument.Air_flow= double(data(indexesA2.measurements.AirFlow).Value);
    % Instrument.CO2
    % Instrument.O2
    % Instrument.Pump_1
    % Instrument.Pump_2
    % Instrument.Acid_pump
    % Instrument.Base_pump
    % Instrument.pO2
    % Instrument.NowSampleBarCode
    % Instrument.NowSampleName
    % Instrument.NowSampleBioreactor
    
    
    % xhatukf = xhat; % only if using UKF
    %Q_ukf = diag([0.0001 0.0000004 0.000002 0.001]);
else
%     load('20220329_EKF_PID_Glucose_A2.mat','Instrument0');
%    BioReactor = 'A2';
%     par=GetParValues(BioReactor,Instrument0,clk_0);
%     Instrument0.V = par.V0; %Initial reactor volume (user defined) 
%     Filter.P=par.P; % this is for the first EKF run at i==1
%     % Define the initial conditions
%     par.S0=9.5;
%    x0= [par.V0 par.X0 par.S0 par.CO20];
%     % Initializing for the first estimator
% 
%     Filter.xhat= x0;
end

% if you stop the code in the middle of experiments and want to proceed 
%from same place, you can:
% 1. Run the code until here (debug here)
% 2. load the most recent SaveFile by the following command:
% load (SaveFile,'VarMem','i','Instrument0','Instrument','par','x0','MPC')
% For example:
% load ('20220928_EKF_PID_Glucose_A2.mat','VarMem','i','Instrument0','Instrument','par','x0','MPC')
% 3. continue, remove the debug point

%% Enter in the loop
 for i = 1:990 % Change later to while
    tic
    
    % DATA EXTRACTION
    % Read the data
     
        if i==1 
           ProcTime= 0;
           iteration=1;
        else
           iteration=i;
           % We need to update the ProcTime always!!!   - Done in SaveMemory         

%            ProcTime= (clk_now*[0 0 60*60*24 60*60 60 1]'-par.clk_0*[0 0 60*60*24 60*60 60 1]');
%            Instrument.ProcTime= ProcTime;

        end       
        
        data = read(readGrpA2);
        %get HPLC data from OPC Lucullus
        %dataHPLC = read(readGrpHPLC);        
    
       % Extract data from the instrument
       [Instrument]=GetInstData(data,indexesA2,clk_now,par,iteration);
       
       %% CLEANING REVERSE FLUSH
       
       if iteration>1
           [C] = intersect(VarMem.Instrument.SampleDataBase(:,1),Instrument.NowSampleBarCode,'stable');
           if not(isempty(Instrument.NowSampleBarCode)) && isempty(C)
               % new sample was just taken
               disp('FLUSHING....')
               %write(writeGrpA2.Item(indexesA2.setpoints.Feed_Ext_3_Setpoint), {6})
               write(writeGrpA2.Item(indexesA2.setpoints.Feed_Ext_3_Setpoint), {10})
               % the value 5 is about 10mL in 30 second... about 10 dropse
           else
               write(writeGrpA2.Item(indexesA2.setpoints.Feed_Ext_3_Setpoint), {0})
           end
       end
%%
       % Get HPLC measurements from file, if exist:
       [Instrument,VarMem]=GetHPLCFromFile(Instrument,clk_now,par,VarMem,iteration);
       
       %% Instrumentation additional analysis additional
       [Instrument,par]= Analysis_instrumentation(Instrument,Instrument0,par,iteration,VarMem);
       
       %% Measurement vector with HPLC data and the outlier identification algorithm
       % [measurements, VarMem,Instrument]= HPLC_data(Instrument,VarMem,iteration,ProcTime);
       [measurements,VarMem,Instrument]= HPLC_data(Instrument,VarMem,iteration,Instrument.ProcTime,par);
       

%% Control action Feed
    if i > 990
        feed_i = 0;
    else
        % CSTR
        feed_i = feed_profile(i); %L/h 
        
    end
    % Fin = Fout
    Controller.Fin= feed_i;

    % Use the calibration data to transform Flow into % pump   
    Controller.Fin2pump= par.pump.m1*feed_i+par.pump.b2;

     % Fin
     write(writeGrpA2.Item(indexesA2.setpoints.Feed_Ext_1_Setpoint), {Controller.Fin2pump});


%% Save data into VarMem function
% Input is the Intrument0 also and initialize answer to not erase data
if i==1
    % First iteration
    VarMem=SaveVar2Memory(i,par,measurements,Instrument0,Instrument,Controller,VarMem);
else % send VarMem to the function
    VarMem=SaveVar2Memory(i,par,measurements,Instrument0,Instrument,Controller,VarMem);
end

    
%%  Plotting with an external function
plotting_VarMem(VarMem,par,iteration);
usedTime = toc;
save(par.Info.SaveFile)

%remove the HPLC measurement before the next iteration, because each HPLC
%measurement only used once.
  Instrument.HPLC.gluc=[];
    Instrument.HPLC.xyl= [];
    Instrument.HPLC.man= [];
    Instrument.HPLC.ara= [];
     Instrument.HPLC.TimeStamp=[];
% Note: it must be Iteration time> TOC    g
% Otherwise it does not pause
 
% IMPORTANT: update of the new clock must be BEFORE the pause, otherwise
% any new HPLC file will not be recognized (because the comparison with the
% time stamp of the new file and the clock). 
% ONLY HERE!
clk_now=clock;
clk_now=fix(clk_now);

fprintf('iteration=%d',i);
%pause(par.iterationTime-usedTime);
pause(10) % CHANGE 
disp(i)
for z=1:log10(i)+11
    fprintf('\b');
end

i = i+1; % advance one step


end

disconnect(da)
delete(da)


end

% Function to obtain the parameters of the instrument
function par=GetParValues(BioReactor,Instrument0,clk_0)
% All the parameters of the reaction are packed here!

% % information about the experiment, file and folder
  par.Info.ChromeleonFolder ='C:\Users\LAB\Desktop\ChromeleonResult\'; % chromeleon folder!
  par.Info.Substrate = 'Glucose';
  par.Info.Control = 'None';
  par.Info.Filter = 'None';
  par.Info.date = datestr(now,'yyyymmdd');
  par.Info.BioReactor=BioReactor;
  par.Info.SaveFile=sprintf('%s_%s_%s_%s_%s.mat',par.Info.date,par.Info.Filter,par.Info.Control,par.Info.Substrate,par.Info.BioReactor);
  par.Info.LoopMethod= 'Feedback_from_model';
  %par.Info.ChromeFile=sprintf('%s\\%s_%s_%s_%s.mat',par.Info.ChromeleonFolder,par.Info.date,par.Info.Filter,par.Info.Control,par.Info.Substrate,par.Info.BioReactor);

  % The initial time in the system is:
% Compute it as hrs always
% Take into consideration the hours, minutes and seconds of today.
par.clk_0= clk_0;
switch clk_0(2)
    case {1,3,5,7,8,10,12}
        par.mth=31;
    case 2
        par.mth=28; % CAREFULL!!! it can be 29 in extended years
    case {4,6,9,11}
        par.mth=30;
end

% The initial time is in hrs
par.initial_time= clk_0*[0 par.mth*60*60*24 60*60*24 60*60 60 1]'/3600;
par.iterationTime =60; % in seconds

par.dt= 1/60; % the discrete time step in hours

% the delay time in minutes from the sample taken to the chromeleon wrote
% the file:
% OBSOLETE
par.NUMERA_HPLC_SamplingTime=55; % in minutes
par.NUMERA_HPLC_SamplingTime=par.NUMERA_HPLC_SamplingTime * 60; % convert to seconds

% "step" determinefd hsm often (how many while loop) before the EKF is
% running. for example, if step = 5, the EKF will run calculation every
% iterationTime * 5, i.e. every fifth loop iteration. 
% default: Preferable step=1 (each time step)
% Initial conditions:
par.V0= 1.5;
par.S0= 15; % initial g/L sugar in the reactor

if nargin>1 && not(isempty(Instrument0)) % if there is Instrument0   
% Initial measurement of CO2
par.CO2inAir_volpercent = Instrument0.CO2; 
par.CO2inAir_volpercent = 0.04; 


% Density of CO2 in air
CO2_density = 1.77;

% No CO2 at the beginning
CO20= 0;
CarbDiox = CO20/par.V0;
par.CO20= 0;

% safeguard of the balances A,B,C and D
% if the balances reset by mistake, it will be ignored.
% the baseline weight is first zero. if a reset happens, the baselineWeight
% will be changed in the Analysis_Instrumentation function.
if isfield(Instrument0,'Balance_A2') %  'Balance_A' CHANGE ->'Balance_A2' ? 
par.baselineWeight.A2=Instrument0.Balance_A2; %  'Balance_A' CHANGE ->'Balance_A2' ? 
par.baselineWeight.B2=Instrument0.Balance_B2; % par.baselineWeight.B -> B2
else
par.baselineWeight.C=Instrument0.Balance_C;
par.baselineWeight.D=Instrument0.Balance_D;
end

% Biomass corelation
% This needs to be re-evaluated
if Instrument0.NIR>= 0.52
    CDWcu = 22.187 * Instrument0.NIR - 5.0991 ;
else
    CDWcu = 11.124 * Instrument0.NIR + 0.66116 ;
end

% No extra big NIR values (probe disconnected)
if Instrument0.NIR >= 5
    CDWcu= 0.66116;
end

% No negative NIR values (no biomasss in the system)
if Instrument0.NIR <= 0
    CDWcu= 0.66116;
end     

% Biomass concentration
par.X0= CDWcu; % CDW in g/L (OD 1)

% Concentration of the feeding
par.Sin= 200; % g/L 

% set points:
% Use only constant setpoints
par.sp.S= 15; % sugar set point, mg/L

% Set point of the Volume controller
par.sp.V= 1.55;

% Set point of the CDWcu
par.sp.CDWcu= 17;


% Parameters from the JPC paper
% Wild type C Glutamicum 
 par.Y_XS   = 0.42042;   % g g^-1
 par.mu_max = 0.19445;   % h^-1
 par.Ks     = 0.007;   % g L^-1 
 par.Y_CO2X = 0.543;     % g g^-1
 par.kd     = 0.006;     % h^-1

% new parameter estimation 2022 1st of March 
% Pediocin producing strains
%par.Y_XS   = 0.3753;   % g g^-1
%par.mu_max = 0.0460;   % h^-1
%par.Ks     = 0.036;   % g L^-1 
%par.Y_CO2X = 0.4361;     % g g^-1
%par.kd     = 4 *10^-4;     % h^-1


% Calibration value to merge NIR and CO2
par.alpha = 0.6655;
   

% Parameter for the pumps
% linear conversion of the flow to opening of external WM pumps
par.pump.m1= 45.0196; % Pump 1 %Prevoiusly 37.03
par.pump.b1= 0;

% Parameters for pump 2
par.pump.m2= 45.0196; % pump 2
par.pump.b2= 0;



end

end

% Function to find the OPC ports and the indexes
function [readGrp, writeGrp, indexes] = findReactorTags(reactor, da)

% Connect to OPC Server 
%da = opcda('localhost', 'Securecell.OpcBridgeServer.DA.1'); 
%connect(da); 
itmIDs = serveritems(da); 


reactorIDs = [];

% Find all values related to the reactor
for i = 1:length(itmIDs)
    str = [char(itmIDs(i)), ' '];   % Convert cell string to char, add space
                                    % for easier parsing (making sure we
                                    % always have 2 characters)
    if length(reactor) == 2
        if str(1:2) == reactor
            reactorIDs = [reactorIDs; itmIDs(i)];
        end                         
    elseif reactor == 'A' || reactor == 'B'
        if str(1) == reactor && str(2) ~= '2'
            reactorIDs = [reactorIDs; itmIDs(i)];
        end
    else
        if str(1) == reactor
            reactorIDs = [reactorIDs; itmIDs(i)];
        end
    end
end



%  {'Numera.Numera.SEQUENCE_1.TimeStampSec_NU'                   }
%     {'Numera.Numera.SEQUENCE_1.TimeStampHour_NU'  

if isempty(reactorIDs)
    error("No tags were found for reactor """+convertCharsToStrings(reactor)+""". Have you inserted a valid reactor name? Use characters and capital letters, like 'A'.")
end

% Additional instrumentation
balances = reactorIDs(contains(reactorIDs, 'Balance'));
buglab = reactorIDs(contains(reactorIDs, 'Buglab'));
HPLC= reactorIDs(contains(reactorIDs,'HPLC'));

% Take the time stamp sf the last sample of Numera to HPLC
HPLC=[HPLC ; 
    {'Numera.Numera.SEQUENCE_1.TimeStampYear_NU'};
    {'Numera.Numera.SEQUENCE_1.TimeStampMonth_NU'};
    {'Numera.Numera.SEQUENCE_1.TimeStampDay_NU'};
    {'Numera.Numera.SEQUENCE_1.TimeStampHour_NU'};
    {'Numera.Numera.SEQUENCE_1.TimeStampMinute_NU'};% 
    {'Numera.Numera.SEQUENCE_1.TimeStampSec_NU'}];


measurements = reactorIDs(contains(reactorIDs, 'Measured Values'));
controlOutputs = reactorIDs(contains(reactorIDs, 'Controller Output'));
setpoints = reactorIDs(contains(reactorIDs, 'Setpoints'));
sysDevice = reactorIDs(contains(reactorIDs, 'System Device.SystemSubDev'));

% Digging out the timers
HPLC = HPLC(2:end);
measurements = measurements(2:end);
controlOutputs = controlOutputs(2:end);
sysDevice = sysDevice(2:end);
setpoints = setpoints(2:end);

% Adding the scales and HPLC to measurements
Balance_A2 = balances(contains(balances, {'Balance A2.Balance Ports.Weight'}));
Balance_B2 = balances(contains(balances, {'Balance B2.Balance Ports.Weight'}));
Balance_C = balances(contains(balances, {'Balance C.Balance Ports.Weight'}));
Balance_D = balances(contains(balances, {'Balance D.Balance Ports.Weight'}));
measurements = [measurements; Balance_A2; Balance_B2;Balance_C;Balance_D; HPLC];

% Adding the buglab to measurements
buglab_measurement = buglab(contains(buglab, {'Result Raw'}));
measurements = [measurements; buglab_measurement];

% Create Read Group 
readGrp = addgroup(da, ['ReadGroup' ,reactor]); 


%% SEQUENCE TWO OF THE HPLC DATA
NUMERA=itmIDs(contains(itmIDs,'Numera.Numera.SEQUENCE_2'));
NUMERA=NUMERA(2:end);

% For HPLC reactor identification of SampleManagmenet 
% important to know from which reactor the file with the barcode was taken
SampleBarcode=itmIDs(contains(itmIDs,'SampleBarcode'));
SampleName=itmIDs(contains(itmIDs,'SampleName'));

% this loop is to secure the name of the reactor is registered only from the
% first field of the list of possible reactors (because the letters D, E, A
% can also be par of the names in ItemID): 
for z=1:length(SampleBarcode)
    
    SBC=SampleName(contains(SampleName{z}(1:2),reactor));
    if not(isempty(SBC))
        SBCidx=z;
    end
end
SampleName=SampleName(SBCidx);
SampleBarcode=SampleBarcode(SBCidx);

% Adding Numera Sequence 2 information
measurements = [measurements; NUMERA;SampleBarcode;SampleName];



% Add Tags 
readID = [measurements; controlOutputs; sysDevice];
itm = additem(readGrp, readID);
writeGrp = addgroup(da, ['WriteGroup', reactor]);
writeID = setpoints;
itm = additem(writeGrp, writeID);

% Making a structure with the indexes of each variable of interest
measurementIndexes = struct;
measurementIndexes.ExitO2 = find(contains(measurements, 'ExitO2'));
measurementIndexes.GasMix = find(contains(measurements, 'GasMix'));
measurementIndexes.OD = find(contains(measurements, 'OD'));
measurementIndexes.Temperature = find(contains(measurements, 'Temperature'));
measurementIndexes.Feed = find(contains(measurements, 'Feed'));
if length(measurementIndexes.Feed) > 1
    measurementIndexes.Feed_Ext_1 = find(contains(measurements, 'Ext_1'));
    measurementIndexes.Feed_Ext_2 = find(contains(measurements, 'Ext_2'));
    measurementIndexes.Feed_Ext_3 = find(contains(measurements, 'Ext_3'));
    measurementIndexes.Feed_Ext_4 = find(contains(measurements, 'Ext_4'));
    measurementIndexes.Feed_Ext_5 = find(contains(measurements, 'Ext_5'));
    % Make a list of only the IDs of "feed" pumps
    feedMeasurements = measurements(measurementIndexes.Feed);
    % Find the only index that is not "Ext"
    measurementIndexes.Feed = measurementIndexes.Feed ...
        (~contains(feedMeasurements, 'Ext'));
end
measurementIndexes.O2Flow = find(contains(measurements, 'O2Flow'));
measurementIndexes.N2Flow = find(contains(measurements, 'N2Flow'));
measurementIndexes.AirFlow = find(contains(measurements, 'AirFlow'));
measurementIndexes.ExitCO2 = find(contains(measurements, 'ExitCO2'));
measurementIndexes.pH = find(contains(measurements, 'Values.pH'));
measurementIndexes.GMFlow = find(contains(measurements, 'GMFlow'));
measurementIndexes.pO2 = find(contains(measurements, 'pO2'));
measurementIndexes.Antifoam = find(contains(measurements, 'Antifoam'));
measurementIndexes.Stirrer = find(contains(measurements, 'Stirrer'));
measurementIndexes.Balance_A2 = find(contains(measurements, 'Balance A2'));
measurementIndexes.Balance_B2 = find(contains(measurements, 'Balance B2'));
measurementIndexes.Balance_C = find(contains(measurements, 'Balance C'));
measurementIndexes.Balance_D = find(contains(measurements, 'Balance D'));


% Add the HPLC information
measurementIndexes.HPLC_gluc= find(contains(measurements,'Measurements_HPLC.Glucose'));
measurementIndexes.HPLC_ara= find(contains(measurements,'Measurements_HPLC.Arabinose'));
measurementIndexes.HPLC_xyl= find(contains(measurements,'Measurements_HPLC.Xylose'));
measurementIndexes.HPLC_man= find(contains(measurements,'Measurements_HPLC.Mannose'));
measurementIndexes.HPLC_year= find(contains(measurements,'TimeStampYear_NU'));
measurementIndexes.HPLC_month= find(contains(measurements,'TimeStampMonth_NU'));
measurementIndexes.HPLC_day= find(contains(measurements,'TimeStampDay_NU'));
measurementIndexes.HPLC_hour= find(contains(measurements,'TimeStampHour_NU'));
measurementIndexes.HPLC_minute= find(contains(measurements,'TimeStampMinute_NU'));
measurementIndexes.HPLC_sec= find(contains(measurements,'TimeStampSec_NU'));

% Add the Buglab
measurementIndexes.Buglab = find(contains(measurements, 'Buglab'));

% Add the Seq 2 from Numera
measurementIndexes.NumeraIndexes.analyzer_number= find(contains(measurements,'ANALYZER_NUMBER'));
measurementIndexes.NumeraIndexes.seq2_start =find(contains(measurements,'SEQ2_START'));
measurementIndexes.NumeraIndexes.seq2_state =find(contains(measurements,'SEQ2_STATE'));
measurementIndexes.NumeraIndexes.sampling_done =find(contains(measurements,'SAMPLING_DONE'));
measurementIndexes.NumeraIndexes.analyzer_transf_type =find(contains(measurements,'ANALYZER_TRANSF_TYPE'));
measurementIndexes.NumeraIndexes.sample_ready= find(contains(measurements,'SAMPLE_READY'));
measurementIndexes.NumeraIndexes.seq2_stop= find(contains(measurements,'SEQ2_STOP'));

measurementIndexes.SampleBarcode=find(contains(measurements,SampleBarcode));
measurementIndexes.SampleName=find(contains(measurements,SampleName));

controlOutputIndexes = struct;
controlOutputIndexes.GMFlow_Output = find(contains(controlOutputs, 'GMFlow_Output')) + length(measurements);
controlOutputIndexes.Stirrer_Output = find(contains(controlOutputs, 'Stirrer_Output')) + length(measurements);
controlOutputIndexes.AirFlow_Output = find(contains(controlOutputs, 'AirFlow_Output')) + length(measurements);
controlOutputIndexes.OD_Output = find(contains(controlOutputs, 'OD_Output')) + length(measurements);
controlOutputIndexes.Temperature_Output = find(contains(controlOutputs, 'Temperature_Output')) + length(measurements);
controlOutputIndexes.GasMix_Output = find(contains(controlOutputs, 'GasMix_Output')) + length(measurements);
controlOutputIndexes.pH_Output = find(contains(controlOutputs, 'pH_Output')) + length(measurements);
controlOutputIndexes.Feed_Output = find(contains(controlOutputs, 'Feed_Output')) + length(measurements);
controlOutputIndexes.pO2_Output = find(contains(controlOutputs, 'pO2_Output')) + length(measurements);
controlOutputIndexes.ExitCO2_Output = find(contains(controlOutputs, 'ExitCO2_Output')) + length(measurements);
controlOutputIndexes.N2Flow_Output = find(contains(controlOutputs, 'N2Flow_Output')) + length(measurements);
controlOutputIndexes.Antifoam_Output = find(contains(controlOutputs, 'Antifoam_Output')) + length(measurements);
controlOutputIndexes.O2Flow_Output = find(contains(controlOutputs, 'O2Flow_Output')) + length(measurements);
controlOutputIndexes.ExitO2_Output = find(contains(controlOutputs, 'ExitO2_Output')) + length(measurements);
controlOutputIndexes.PumpValAcid = find(contains(controlOutputs, 'Pump_Val_Acid')) + length(measurements);
controlOutputIndexes.PumpValBase = find(contains(controlOutputs, 'Pump_Val_Base')) + length(measurements);

systemIndexes = struct;
systemIndexes.StepTime = find(contains(sysDevice, 'StepTime')) + ...
    length(measurements) + length(controlOutputs);
systemIndexes.ProcTime = find(contains(sysDevice, 'ProcTime')) + ...
    length(measurements) + length(controlOutputs);

setpointIndexes = struct;
setpointIndexes.Antifoam_Setpoint = find(contains(setpoints, 'Antifoam_Setpoint'));
setpointIndexes.Temperature_Setpoint = find(contains(setpoints, 'Temperature_Setpoint'));
setpointIndexes.pO2_Setpoint = find(contains(setpoints, 'pO2_Setpoint'));
setpointIndexes.Stirrer_Setpoint = find(contains(setpoints, 'Stirrer_Setpoint'));
setpointIndexes.GMFlow_Setpoint = find(contains(setpoints, 'GMFlow_Setpoint'));
setpointIndexes.O2Flow_Setpoint = find(contains(setpoints, 'O2Flow_Setpoint'));
setpointIndexes.pH_Setpoint = find(contains(setpoints, 'pH_Setpoint'));
setpointIndexes.AirFlow_Setpoint = find(contains(setpoints, 'AirFlow_Setpoint'));
setpointIndexes.N2Flow_Setpoint = find(contains(setpoints, 'N2Flow_Setpoint'));
setpointIndexes.GasMix_Setpoint = find(contains(setpoints, 'GasMix_Setpoint'));
setpointIndexes.Feed_Setpoint = find(contains(setpoints, 'Feed'));

if length(setpointIndexes.Feed_Setpoint) > 1
    setpointIndexes.Feed_Ext_1_Setpoint = find(contains(setpoints, 'Ext_1'));
    setpointIndexes.Feed_Ext_2_Setpoint = find(contains(setpoints, 'Ext_2'));
    setpointIndexes.Feed_Ext_3_Setpoint = find(contains(setpoints, 'Ext_3'));
    setpointIndexes.Feed_Ext_4_Setpoint = find(contains(setpoints, 'Ext_4'));
    setpointIndexes.Feed_Ext_5_Setpoint = find(contains(setpoints, 'Ext_5'));

    % Make a list of only the IDs of "feed" pumps
    feedSetpoints = setpoints(setpointIndexes.Feed_Setpoint);
    
    % Find the only index that is not "Ext"
    setpointIndexes.Feed_Setpoint = setpointIndexes.Feed_Setpoint ...
        (~contains(feedSetpoints, 'Ext'));
end



indexes = struct;
indexes.system = systemIndexes;
indexes.setpoints = setpointIndexes;
indexes.measurements = measurementIndexes;
indexes.controlOutputs = controlOutputIndexes;
end

% Function to extract the information of the measurements
function [Instrument]=GetInstData(data,indexesA2,clk_now,par,iteration)
% clk is the clock information taken in the beginning of the expetiments
% here you can add what you want to Instrument struct (as fields)
% Define data for the balances, 

if ~isnan(indexesA2.measurements.Balance_A2)==1
    Instrument.Balance_A2= double(data(indexesA2.measurements.Balance_A2).Value);
    Instrument.Balance_B2= double(data(indexesA2.measurements.Balance_B2).Value);

end

if ~isnan(indexesA2.measurements.Balance_C)==1
    Instrument.Balance_C= double(data(indexesA2.measurements.Balance_C).Value);
    Instrument.Balance_D= double(data(indexesA2.measurements.Balance_D).Value);
end

%%%%%%%%%%%%%%%%%%%
%%%% NIR Acquisition: Smoothing, then linear conversion to Cell dry weight CDw
%%%%%%%%
%%%%
% Measurements of NIR, Buglab;
%Instrument.Buglab= double(data(indexesA2.measurements.Buglab).Value);
%acquisition
Instrument.NIR= double(data(indexesA2.measurements.OD).Value);
% smoothing and conversion is conducted in Analysis_instrumentation()

% Measurements of Air flow CO2 and O2 
% XXX Remove the 60 after discussion with Nadi
Instrument.Air_flow= double(data(indexesA2.measurements.AirFlow).Value);
Instrument.CO2 = double(data(indexesA2.measurements.ExitCO2).Value);
Instrument.O2= double(data(indexesA2.measurements.ExitO2).Value); % 

% Read the data of the pumps
Instrument.Pump_1 = double(data(indexesA2.measurements.Feed_Ext_1).Value);
Instrument.Pump_2 = double(data(indexesA2.measurements.Feed_Ext_2).Value); 
 
% NOT NEEDED!!
% Aqcuire the information from Numera
% Instrument.Numera.Analyzer_number = double(data(indexesA2.measurements.NumeraIndexes.analyzer_number).Value);
% Instrument.Numera.seq2_start= double(data(indexesA2.measurements.NumeraIndexes.seq2_start).Value);
% Instrument.Numera.seq2_state= double(data(indexesA2.measurements.NumeraIndexes.seq2_state).Value);
% Instrument.Numera.sampling_done= double(data(indexesA2.measurements.NumeraIndexes.sampling_done).Value);
% Instrument.Numera.analyzer_transfer_type= double(data(indexesA2.measurements.NumeraIndexes.analyzer_transf_type).Value);
% Instrument.Numera.sample_ready= double(data(indexesA2.measurements.NumeraIndexes.sample_ready).Value);
% Instrument.Numera.seq2_stop= double(data(indexesA2.measurements.NumeraIndexes.seq2_stop).Value);

% Instrument time
if iteration>1
        Instrument.t=iteration*par.iterationTime; %calculate the current time from i and par.iterationtime
        %Instrument.time= Instrument.time(iteration-1) + par.iterationTime;
        Instrument.ProcTime= fix(clock)*[0 par.mth*60*60*24 60*60*24 60*60 60 1]'-par.clk_0*[0 par.mth*60*60*24 60*60*24 60*60 60 1]';        
    else % i==1
        Instrument.t=0; % first iteration time is close to zero
        %Instrument.time= 0;
        Instrument.ProcTime= 0;
end

% Acid and Base outputs
Instrument.Acid_pump= double(data(indexesA2.controlOutputs.PumpValAcid).Value);
Instrument.Base_pump= double(data(indexesA2.controlOutputs.PumpValBase).Value);

% pO2 measurement
Instrument.pO2= double(data(indexesA2.measurements.pO2).Value);

%% Sample HANDLING:
% The samples taken by NUMERA: 
% 1. read the now sample
Instrument.NowSampleBarCode=(data(indexesA2.measurements.SampleBarcode).Value);
Instrument.NowSampleName=(data(indexesA2.measurements.SampleName).Value);
% get the sample time
NUmth=double(data(indexesA2.measurements.HPLC_month).Value);
NUd=double(data(indexesA2.measurements.HPLC_day).Value);
NUh=double(data(indexesA2.measurements.HPLC_hour).Value);
% IMPORTANT!! because NUMERA reports the hour in UTC, mhile Norway clock is
% CET+1 -- we need to add another hour.
% SUMMER TIME: +2 WINTER: +1
t1=datetime('today','TimeZone','Europe/Oslo');
utc2GMTOslo=tzoffset(t1);

NUh=NUh+hours(utc2GMTOslo);
NUm=double(data(indexesA2.measurements.HPLC_minute).Value);
NUs=double(data(indexesA2.measurements.HPLC_sec).Value);

Instrument.NowSampleProcTime=[0 NUmth NUd NUh NUm NUs]*[0 par.mth*60*60*24 60*60*24 60*60 60 1]'-par.clk_0*[0 par.mth*60*60*24 60*60*24 60*60 60 1]';
if Instrument.NowSampleProcTime<0 % in case you started matlab and it found older sample from lucullus.
    Instrument.NowSampleProcTime=0;
end

if isempty(Instrument.NowSampleBarCode)
   % Instrument.NowSampleBioreactor= '';
else
    Instrument.NowSampleBioreactor=Instrument.NowSampleName(1:2);
end


end

% Obtain further analysis of the instrumentation, except HPLC
function[Instrument,par] = Analysis_instrumentation(Instrument,Instrument0,par,iteration,VarMem)
       % We need to divide by 1000 to have it in kg, 
       % Minus sign because the balance reduces its weight when consumed
       if iteration==1
           % first iteration, no memory of the flow
           Instrument.Volume_in= 0;
           Instrument.Volume_out= 0;
           Instrument.Flow_in = 0;
           Instrument.Flow_out = 0;
           Instrument.CO2_measurement = 0;  
           Instrument.V= par.V0;
           
           % CO2 analysis
           % Initial condition of air in the system (no overvalue of CO2, in case of wrong calibration)           
           if Instrument.CO2>0.06
               CO2inAir_volpercent= 0.04;
           else
               CO2inAir_volpercent = Instrument.CO2;
           end                       
                      
           % CO2 density kg/m3 and in g/L
           CO2_density = 1.77; % g/L
           
           % Volumetric flow of CO2
           % The Percentage value the volume of CO2
           Volume_CO2 = (((Instrument.CO2-CO2inAir_volpercent)));           
           if Volume_CO2<0
               Volume_CO2=0;
           end
           
           % CO2 excreted in units to check
           Instrument.CO2_excreted = (Volume_CO2*CO2_density); % % CO2/hr
           
           % The follow is the CO2 in % for the estimator
           % The total volume of the reactor is V_total:
           V_total= 2.7;
           
           % XXX Here multiply by 60, and in the air flow ignore the 60
           %Instrument.CO2_measurement_EKF = 60 * Instrument.CO2_excreted/((V_total-Instrument.V).*100);
           Instrument.CO2_measurement = 60 * Instrument.CO2_excreted/((V_total-Instrument.V).*100);
          
           % At the beginning no CO2 produced
           Instrument.CER=0;
           Instrument.CO2_stoichiometric = 0;           
           
           % NIR data as such
           NIR= Instrument.NIR;
           % No smoothing of first point
           Instrument.NIR_smooth= NIR;
           if NIR>= 0.52
               CDWcu = 22.187 *NIR - 5.0991 ; % MUST BE UPDATED
           else
               CDWcu = 11.124 *NIR + 0.66116 ;
           end
           % No extra big NIR values when disconnected
           if NIR >= 5
               CDWcu= 0.66116;
           end
           % No negative NIR values
           if NIR <= 0
               CDWcu= 0.66116;
           end
           % Store the CDW information
           Instrument.CDWcu= CDWcu;            
           
           % At the beginning no Substrate added
           Instrument.S_added= 0;          
           
       else % iterations > 1
           
           % Compute the balances values
           if isfield(Instrument,'Balance_A2')==1
               % Zero the balances
               % this code is written so that the initial balance A2 is 0,
               % and is negatively decreasing
               % FILTER algorithm works like this:
               % 1. Balance A2 must be monotonically decreasing. 
               % 2. diff values less than 0.02 are ignored (keep the old
               % value)
               diffBalanceA2=Instrument.Balance_A2-VarMem.Instrument.Balance_A2(iteration-1);
               % This difference should either decrease or stay the same,
               % never increase.
               if diffBalanceA2>0
                   % values increased, keep the previous value
                   Instrument.Balance_A2=VarMem.Instrument.Balance_A2(iteration-1);
               elseif abs(diffBalanceA2)<0.05
                % the change is too small, it is noise. keep the previous
                % value
                  Instrument.Balance_A2=VarMem.Instrument.Balance_A2(iteration-1);
           
               end
               
               % Balance B2 (out balance)
                diffBalanceB2=Instrument.Balance_B2-VarMem.Instrument.Balance_B2(iteration-1);
               % This difference should either increase or stay the same,
               % never decrease.
               if diffBalanceB2<0
                   % values decreased, keep the previous value
                   Instrument.Balance_B2=VarMem.Instrument.Balance_B2(iteration-1);
               elseif abs(diffBalanceB2)<0.05
                % the change is too small, it is noise. keep the previous
                % value
                  Instrument.Balance_B2=VarMem.Instrument.Balance_B2(iteration-1);
               end
               
               if abs(diffBalanceA2)>100
                   % We have a reset of the balances. This needs a
                   % correction to the initial Intrument0
                    par.baselineWeight.A2= par.baselineWeight.A2 + diffBalanceA2;      
               end
               if abs(diffBalanceB2)>100
                   % We have a reset of the balances. This needs a
                   % correction to the initial Intrument0
                    par.baselineWeight.B2= par.baselineWeight.B2 + diffBalanceB2;
               end
               
               Instrument.Volume_in = -(Instrument.Balance_A2 - par.baselineWeight.A2)/1080;        
               Instrument.Volume_out = (Instrument.Balance_B2 - par.baselineWeight.B2)/1000;
               
           end
           
           
           if isfield(Instrument,'Balance_C')==1
                % Instrument.Balance_C=Instrument.Balance_C+par.baselineWeight;
                diffBalanceC=Instrument.Balance_C-VarMem.Instrument.Balance_C(iteration-1);
               % This difference should either decrease or stay the same,
               % never increase.
               if diffBalanceC>0 && abs(diffBalanceC)<100
                       Instrument.Balance_C=VarMem.Instrument.Balance_C(iteration-1);
                  % end
                  
               elseif abs(diffBalanceC)<0.1
                % the change is too small, it is noise. keep the previous
                % value
                  Instrument.Balance_C=VarMem.Instrument.Balance_C(iteration-1);
               end

              
               % Balance D (out balance)
                diffBalanceD=Instrument.Balance_D-VarMem.Instrument.Balance_D(iteration-1);
               % This difference should either decrease or stay the same,
               % never increase.
               if diffBalanceD<0
                   % values increased, keep the previous value
                   Instrument.Balance_D=VarMem.Instrument.Balance_D(iteration-1);
               elseif abs(diffBalanceD)<0.1
                % the change is too small, it is noise. keep the previous
                % value
                  Instrument.Balance_D=VarMem.Instrument.Balance_D(iteration-1);
               end
     
               if abs(diffBalanceC)>100
                   % We have a reset of the balances. This needs a
                   % correction to the initial Intrument0
                    par.baselineWeight.C= par.baselineWeight.C + diffBalanceC;      
               end
               
               if abs(diffBalanceD)>100
                   % We have a reset of the balances. This needs a
                   % correction to the initial Intrument0
                    par.baselineWeight.D= par.baselineWeight.D + diffBalanceD;
               end
               Instrument.Volume_in = -(Instrument.Balance_C - par.baselineWeight.C)/1000;        
               Instrument.Volume_out = (Instrument.Balance_D -  par.baselineWeight.D)/1000;
           end
           
             %regardless of the balances values, the volumes cannot be negative:
             if Instrument.Volume_in<0
                 Instrument.Volume_in=0;
             end
             
              if Instrument.Volume_out<0
                 Instrument.Volume_out=0;
             end
               
           % Volume estimation from the measurements
           Instrument.V= par.V0 + Instrument.Volume_in - Instrument.Volume_out;
           
           % Delta of time in hrs
           Delta= (Instrument.ProcTime-VarMem.ProcTime(iteration-1))/3600;                 
           
           % Compute the flows  
           % Current value minus the previous value divided by the delta in time.
           Instrument.Flow_in = (Instrument.Volume_in - VarMem.Instrument.Volume_in(iteration-1))/ Delta;
           Instrument.Flow_out = (Instrument.Volume_out - VarMem.Instrument.Volume_out(iteration-1))/ Delta;    
           
           % Eliminate data that is only noise
           if iteration>1             
               
               % from analysis of the balances (diff(balance_A2) we found
               % that the flow during 30 seconds is not less than 0.012. 
               % Eliminate noisy data from Flow in
               if Instrument.Flow_in < 0.012 || isnan(Instrument.Flow_in) % 5 mL/h
                   Instrument.Flow_in= 0;
               end               
               if Instrument.Flow_in>2.5 % saturation, because the pumps has max capacity, so this removes noise in the balances.
                   Instrument.Flow_in=2.5;
               end
               
               % Eliminate noisy data from Flow out
               if Instrument.Flow_out < 0.012 || isnan(Instrument.Flow_out)
                   Instrument.Flow_out= 0;
               end
               if Instrument.Flow_out>2.5 % saturation, because the pumps has max capacity, so this removes noise in the balances.
                   Instrument.Flow_out=2.5;
               end
           end
           
           % CO2 computation
           % Take First point
          % CO2inAir_volpercent= VarMem.Instrument.CO2(1);
       
           % CO2 density kg/m3 and in g/L
           CO2_density = 1.77; % g/L
           
           % Volumetric flow of CO2
           % The Percentage value the volume of CO2
           Volume_CO2 = (((Instrument.CO2-par.CO2inAir_volpercent)));           
           if Volume_CO2<0
               Volume_CO2=0;
           end
                                           
           % CO2 excreted in units to check
           Instrument.CO2_excreted = (Volume_CO2*CO2_density); % % CO2/hr
           
           % The follow is the CO2 in % for the estimator
           % The total volume of the reactor is V_total:
           V_total= 2.7;
           Instrument.CO2_measurement = 60 * Instrument.CO2_excreted/((V_total-Instrument.V).*100);
           if Instrument.CO2_measurement < 0
              % how is it possible?
               Instrument.CO2_measurement = 0 ;
           end
           
           % Stoichiometric CO2
           Instrument.CER = Instrument.Air_flow*(Volume_CO2*CO2_density); % g CO2/hr
           CarbDiox= VarMem.Instrument.CO2_stoichiometric(iteration-1);
           
           % Delta of time in hrs
           Delta= (Instrument.ProcTime-VarMem.ProcTime(iteration-1))/3600;                                
           
           % Make the integral gCO2/L 
           Instrument.CO2_stoichiometric = CarbDiox + Instrument.CER*Delta;    
       
           %%%% NIR Handling: Smoothing, then linear conversion to Cell dry weight CDw
           %%%%%%%%
           %%%%
          
           % smooth the whole vector in the NIR
           NIR_analysis = [VarMem.Instrument.NIR Instrument.NIR];
           
           % Gaussian filter
           Instrument.NIR_smooth= smoothdata(NIR_analysis,'lowess',80);
           Instrument.NIR_smooth=Instrument.NIR_smooth(end);
           
           NIR= Instrument.NIR_smooth;
           if NIR>= 0.52
               CDWcu = 22.187 *NIR - 5.0991 ;
           else
               CDWcu = 11.124 *NIR + 0.66116 ;
           end
           % No extra big NIR values when disconnected
           if NIR >= 5
               CDWcu= 0.66116;
           end
           % No negative NIR values
           if NIR <= 0
               CDWcu= 0.66116;
           end
           % Store the CDW information
           Instrument.CDWcu= CDWcu;
           
           % Substrate added in the bioprocess:
           Instrument.S_added = (Instrument.Volume_in)*par.Sin;

           
       end
end

% Function to obtain HPLC data from File
function [Instrument,VarMem]=GetHPLCFromFile(Instrument,clk,par,VarMem,iteration)

  %% Sample data to store in the Numera system
    if iteration ==1
        % nan is because intersect needs to compare to something
        VarMem.Instrument.SampleDataBase(1,1)={'0'};
        VarMem.Instrument.SampleDataBase(1,2)={nan};
        VarMem.Instrument.SampleDataBase(1,3)={nan};
        VarMem.Instrument.SampleDataBase(1,4)={nan};
    
    % 1. Is there a new sample??
    elseif isempty(Instrument.NowSampleBarCode)
        % no --> do nothing
        
    else
        %yes --> compare it to the database
        
        %2. Compare to SampleDataBase
        % the structure of the database is cell matrix m x 4, where m is the number
        % of samples taken from the beginning of the process. the columns are
        % Barcode ,  timestamp,  reactor,  name
         L=length(VarMem.Instrument.SampleDataBase(:,1));
        [C] = intersect(VarMem.Instrument.SampleDataBase(:,1),Instrument.NowSampleBarCode,'stable');
        if isempty(C)
            % new sample does not exist in the database. add the sample
            VarMem.Instrument.SampleDataBase(L+1,1)={Instrument.NowSampleBarCode};
            VarMem.Instrument.SampleDataBase(L+1,2)={Instrument.NowSampleProcTime};
            VarMem.Instrument.SampleDataBase(L+1,3)={Instrument.NowSampleBioreactor};
            VarMem.Instrument.SampleDataBase(L+1,4)={Instrument.NowSampleName};
        end
     
        
    end
%% Aqcuire HPLC measurements
%Read from the Chromeleon Folder, and search for a new file using the
%iteration clock clk

  [file,FileSampleBarCode]=FindNewChromeleonFiles_HPLC(clk,par);
  
  % need to remove the .txt from the barcode:
  FileSampleBarCode=FileSampleBarCode(1:end-4);
  
  % the time file was genrated is the end of HPLC measurement. Understood
  Instrument.HPLC.gluc=[];
  Instrument.HPLC.xyl=[];
  Instrument.HPLC.man=[];
  Instrument.HPLC.ara=[];
  
  if not(isempty(file))
     
     %if strcmp(Instrument.SampleName(1:2),par.Info.BioReactor)
     
     % Now compare the Sample Barcode the the database of all the samples
     % taken, and extract from the SampleDataBase he reactor and the
     % timestamp:
     [C,IA] = intersect(VarMem.Instrument.SampleDataBase(:,1),FileSampleBarCode,'stable');
     if not(isempty(C))
         
          % is this the correct reactor?
          IsBioReactor=VarMem.Instrument.SampleDataBase{IA,3};
          if strcmp(IsBioReactor,par.Info.BioReactor)
              %It is the correct!
          %extract the data from the file
          % the format is Instrument.HPLC.Gluc, Instrument.HPLC.Man,
          % Instrument.HPLC.Ara etc
          Instrument.HPLC = ReadSugarValuesFromChromeleonFile(file);
          Instrument.HPLC.TimeStamp= VarMem.Instrument.SampleDataBase{IA,2};
          end
     else
         disp('ERROR! new chromeleon file was created and found, but it is missing in the SampleDataBase!')
         %disp('debugging...')
         %keyboard
     end
      
  end
end

% Function to analyze HPLC data
function[measurements,VarMem,Instrument]= HPLC_data(Instrument,VarMem,iteration,ProcTime,par)

i= iteration;
% If you do not want to use the possible previous HPLC information
% First iteration is zero
if i==1
    Instrument.HPLC.gluc=[];
    Instrument.HPLC.xyl= [];
    Instrument.HPLC.man= [];
    Instrument.HPLC.ara= [];
end

% Define the vector state measurements
% Case 1) No measurement
% Case 2) With HPLC measurement
% If there is no HPLC measurement then the structure is empty
% Then we have 3 measurements    
if isempty(Instrument.HPLC.gluc)
    % Case 1 
    % Initialize the algorithm
    tidsec= [];
    measurements = [Instrument.V, Instrument.CDWcu, nan, Instrument.CO2_measurement];
    
    % Save the vectors as NaN
    VarMem.Instrument.HPLC.ara(i)=nan;
    VarMem.Instrument.HPLC.gluc(i)=nan;
    VarMem.Instrument.HPLC.man(i)=nan;
    VarMem.Instrument.HPLC.xyl(i)=nan;
%     VarMem.Instrument.HPLC.year(i)=nan;
%     VarMem.Instrument.HPLC.month(i)=nan;
%     VarMem.Instrument.HPLC.day(i)=nan;
%     VarMem.Instrument.HPLC.hour(i)=nan;
%     VarMem.Instrument.HPLC.minute(i)=nan;
%     VarMem.Instrument.HPLC.sec(i)=nan;
    VarMem.Instrument.HPLC.timeHPLC(i)=nan;
    VarMem.Instrument.HPLC.idxTimeStamp(i)=nan;
    VarMem.Instrument.HPLC.TimeStamp(i)= nan;
    VarMem.Instrument.HPLC.SampleIteration=nan;
else
    % Case 2
    % Define the HPLC measurement time in seconds        
    % The time stamp of the measurement from the Chromeleon file stamp, in
    % seconds.
    tidsec= Instrument.HPLC.TimeStamp; % Time of the Numera sample of HPLC, since the beginning of ProcTime. In seconds 
    % Confirm that the time of the sampling of the HPLC is lower than the ProcTime
    if Instrument.ProcTime < tidsec || tidsec<0
        % keyboard
        display('Previous HPLC measurement is found in Lucullus, or ProcTime is lower than HPLC measurement (not correct)')
        %keyboard
    end
    
        % Find the set of indexes that have a higher time than tidsec
        idxTimeBackTemp= find(VarMem.ProcTime>tidsec);
        % Here are two indeces: idxTimeBack is how many index to go back in
        % VarMem, this is for the backwards calculation of the EKF, and 
        % idxTimeStamp is where in VarMem the TimeStamp of the HPLC belongs
        Instrument.HPLC.idxTimeStamp= idxTimeBackTemp(1);
        Instrument.HPLC.idxTimeBack= length(VarMem.ProcTime) - idxTimeBackTemp(1);
        
        % Construct the vector of measurement with the HPLC data
        Instrument.HPLC.first_measurement= [VarMem.measurements(Instrument.HPLC.idxTimeStamp,1:2) Instrument.HPLC.gluc VarMem.measurements(Instrument.HPLC.idxTimeStamp,4)];            
        
        % Replace the specific vector in the Memory of measurements
        VarMem.measurements(Instrument.HPLC.idxTimeStamp,:)= Instrument.HPLC.first_measurement;
        VarMem.Instrument.HPLC.timeHPLC(Instrument.HPLC.idxTimeStamp) = tidsec;
        
        % saving the HPLC values that belong to the past, 
        % Here is the index where the information corresponds
        VarMem.Instrument.HPLC.ara(Instrument.HPLC.idxTimeStamp)=Instrument.HPLC.ara;
        VarMem.Instrument.HPLC.gluc(Instrument.HPLC.idxTimeStamp)=Instrument.HPLC.gluc;
        VarMem.Instrument.HPLC.man(Instrument.HPLC.idxTimeStamp)=Instrument.HPLC.man;
        VarMem.Instrument.HPLC.xyl(Instrument.HPLC.idxTimeStamp)=Instrument.HPLC.xyl;    
        VarMem.Instrument.HPLC.idxTimeStamp(Instrument.HPLC.idxTimeStamp)=Instrument.HPLC.idxTimeStamp;
        VarMem.Instrument.HPLC.TimeStamp(Instrument.HPLC.idxTimeStamp)= Instrument.HPLC.TimeStamp;
        VarMem.Instrument.HPLC.SampleIteration=iteration;
        % Please note that the present i-iteration has no HPLC data
        measurements= [Instrument.V, Instrument.CDWcu, nan, Instrument.CO2_measurement];
    %end
end    

end

% Function for saving the information
function VarMem=SaveVar2Memory(i,par,measurements,Instrument0,Instrument,Controller,VarMem)
    % VarMem is the variables in memory, all in struct with fields related
    
    % TIme
    % Fixed the time with the incorporation of VarMeM above
    if i>1
        VarMem.Instrument.t(i)=i*par.iterationTime; %calculate the current time from i and par.iterationtime
        VarMem.time(i)= VarMem.time(i-1) + par.iterationTime;
        
        Proctime= clock*[0 par.mth*60*60*24 60*60*24 60*60 60 1]'-par.clk_0*[0 par.mth*60*60*24 60*60*24 60*60 60 1]';
        VarMem.ProcTime(i)= Proctime;
    else % i==1
        VarMem(i).Instrument.t=0; % first iteration time is close to zero
        VarMem(1).time= 0;
        VarMem.ProcTime(i)= 0;
    end
    % HPLC:
    % it was already saved in the function HPLC_data, but it is repeated
    % here for the sake of the last iteration and plotting (removing this
    % will fail to plot)
    VarMem.Instrument.HPLC.ara(i)=nan;
    VarMem.Instrument.HPLC.gluc(i)=nan;
    VarMem.Instrument.HPLC.man(i)=nan;
    VarMem.Instrument.HPLC.xyl(i)=nan;
%     VarMem.Instrument.HPLC.year(i)=nan;
%     VarMem.Instrument.HPLC.month(i)=nan;
%     VarMem.Instrument.HPLC.day(i)=nan;
%     VarMem.Instrument.HPLC.hour(i)=nan;
%     VarMem.Instrument.HPLC.minute(i)=nan;
%     VarMem.Instrument.HPLC.sec(i)=nan;
    VarMem.Instrument.HPLC.timeHPLC(i)=nan;
    VarMem.Instrument.HPLC.idxTimeStamp(i)=nan;
    VarMem.Instrument.HPLC.TimeStamp(i)= nan;  
    
    % Infors towers
    VarMem.Instrument.Flow_in(i) = Instrument.Flow_in;
    VarMem.Instrument.Flow_out(i) = Instrument.Flow_out;    
    VarMem.Instrument.Pump_1(i) =Instrument.Pump_1;
    VarMem.Instrument.Pump_2(i) = Instrument.Pump_2;
    VarMem.Instrument.Air_flow(i) =Instrument.Air_flow;
    VarMem.Instrument.CO2(i)= Instrument.CO2;
    VarMem.Instrument.CO2_measurement(i)= Instrument.CO2_measurement;
    VarMem.Instrument.O2(i)=Instrument.O2;
   
     VarMem.Instrument.NIR(i) = Instrument.NIR;
    VarMem.Instrument.NIR_smooth(i) = Instrument.NIR_smooth;
    VarMem.Instrument.CDWcu(i) = Instrument.CDWcu;
    
    VarMem.Instrument.Volume_in(i) =Instrument.Volume_in;
    VarMem.Instrument.Volume_out(i) =Instrument.Volume_out;
    VarMem.Instrument.V(i)= Instrument0.V + Instrument.Volume_in - Instrument.Volume_out;
    VarMem.Instrument.Flow_in(i) =Instrument.Flow_in;
    VarMem.Instrument.Flow_out(i) =Instrument.Flow_out;
    VarMem.Instrument.CO2_stoichiometric(i)= Instrument.CO2_stoichiometric;    
    VarMem.Instrument.CER(i)= Instrument.CER;    
    
    % Save the available balances
    if isfield(Instrument,'Balance_A2')==1
        VarMem.Instrument.Balance_A2(i)= Instrument.Balance_A2;
        VarMem.Instrument.Balance_B2(i)= Instrument.Balance_B2;
    end
    
    if isfield(Instrument,'Balance_C')==1
        VarMem.Instrument.Balance_C(i)= Instrument.Balance_C;
        VarMem.Instrument.Balance_D(i)= Instrument.Balance_D;
    end
    
    % VarMem.Instrument.CarbDiox(i)=   Instrument.CarbDiox;    
    % VarMem.Instrument.Buglab(i)= Instrument.Buglab;        
    % The vector of state measurements
    VarMem.measurements(i,:)=measurements;    
        
    % Acid and Base outputs
    VarMem.Instrument.Acid_pump(i)= Instrument.Acid_pump;
    VarMem.Instrument.Base_pump(i)= Instrument.Base_pump;
    
    % pO2 measurement
    VarMem.Instrument.pO2(i)= Instrument.pO2;
    

    % Information from the Numera based on sequence 2
%     VarMem.Instrument.Numera.Analyzer_number(i) =Instrument.Numera.Analyzer_number;
%     VarMem.Instrument.Numera.seq2_start(i)= Instrument.Numera.seq2_start; 
%     VarMem.Instrument.Numera.seq2_state(i)= Instrument.Numera.seq2_state;
%     VarMem.Instrument.Numera.sampling_done(i)= Instrument.Numera.sampling_done;
%     VarMem.Instrument.Numera.analyzer_transfer_type(i)= Instrument.Numera.analyzer_transfer_type;
%     VarMem.Instrument.Numera.sample_ready(i)= Instrument.Numera.sample_ready;
%     VarMem.Instrument.Numera.seq2_stop(i)= Instrument.Numera.seq2_stop;

    % Biomass corelation
    NIR= VarMem.Instrument.NIR(end);
     if NIR>= 0.52
         CDWcu = 22.187 *NIR - 5.0991 ;
   
     else
         CDWcu = 11.124 *NIR + 0.66116 ;
     end
     
     
     % No extra big NIR values
     if NIR >= 5
         CDWcu= 0.66116;
     end
     
     % No negative NIR values
     if NIR <= 0
         CDWcu= 0.66116;
     end     
     
   VarMem.Instrument.CDWcu(i)= CDWcu;
%   VarMem.Instrument.CDWcu(i)= NIR;
    
%     %% MPC solution
%     % Save the information of the MPC in VarMem vector
%     
    % Sugar added:
    VarMem.Instrument.S_added(i)= Instrument.S_added;
    
     

     
end

% Function for plotting all the results
function plotting_VarMem(VarMem,par,iteration)

    set(0,'defaultfigurecolor',[1 1 1]) %Make the figure white
    time= VarMem.ProcTime/3600;
    
    % Substrate information
   % S_timeseries= VarMem.measurements(:,3);
    time_HPLC= VarMem.Instrument.HPLC.timeHPLC/3600;
    
    
    figure(1)
    % magenta: ModSim model simulation
    % red: EKF
    % blue: measurements
    % Black: Set points
  
    subplot(2,2,1)
    plot(time,VarMem.Instrument.V,'-b')
    ylabel('Volume [L]')
    %ylim([0 2.5])
    %xlim([0 40])  
    xticklabels({})      
    box 'off'
    legend ('Measurement')
    legend box off
   

 % X
  subplot(2,2,2)
  plot(time,VarMem.Instrument.CDWcu,'-b')   
    ylabel('Biomass [g L^{-1}]')
    % xlim([0 40])
    %ylim([0 100])
   % xticklabels({})
    xlabel('Time [h]')
    box 'off'
    legend ('Measurement')
    legend box off
    
    subplot(2,2,3)
    plot(time_HPLC,VarMem.measurements(:,3),'-b')
    ylabel('Substrate [g L^{-1}]')
    ylim([0 50])
    %xlim([-40 0])
    xticklabels({})
    box 'off' 
    legend ('Measurement HPLC')
    legend box off
    
%   CO2
    subplot(2,2,4)
    plot(time,VarMem.Instrument.CO2_measurement,'-b')    
    %hold on
    ylabel('CO_2 [% L^{-1}]')
    %ylim([0 0.050])
    % xlim([0 40])
    %xticklabels({})
    xlabel('Time [h]')
    box 'off'
    legend ('Measurement')
    legend box off
       
    
    % INPUT FIGURE
    figure(2)
    subplot(1,2,1)
    hold on
    plot(time,VarMem.Instrument.Flow_in,'-b')
    plot(time,VarMem.Instrument.Flow_out,'-b')
    ylabel('Sugar Flow[L hr^{-1}]')
    box 'off'
    legend("Flow in", "Flow out")
    
    
    subplot(1,2,2)
    plot(time,VarMem.Instrument.S_added,'-b')
    ylabel('Sugar added [g]')
    %ylim([18 40])
    %xlim([0 40])
    xticklabels({})
    box 'off'
    legend("Total Sugar added")


end

function SugVal=ReadSugarValuesFromChromeleonFile(file)
% this file reads the amount values of all sugar components from the file
% created by Chromeleon 6, NOT FROM LUCULLUS.
% if it does not finds a sugar its sends back nan
% can be modified to products and other molecules

%fid = fopen('file03_00021821.txt');
fid = fopen(file);

CTitle = textscan(fid,'%19c',1);
%CTitle = textscan(fid,'%s',1,'Delimiter','\n');

if not(strcmp(CTitle,'Integration Results'))
    disp('ERROR: Wrong Chromeleon results file!')
    SugVal.gluc=nan;
    SugVal.man=nan;
    SugVal.xyl=nan;
    SugVal.ara=nan;
    keyboard
    return
    
end

% read the next two lines (but not used).
CsubTitle=textscan(fid,'%s',4,'Delimiter','\n');

%C=textscan(fid,'%s',1,'Delimiter','\n');
%C=textscan(fid,'%s %s %f %f %f %f %f %f');

%*[^\n] skips the reminder of the line after the Amount:
C=textscan(fid,'%s %s %s %*[^\n]');
fclose(fid);


% replace the comma with dot:
Data1 = strrep(C{3}, ',', '.');
Data = strrep(Data1, 'n.a.', 'nan');

% find the glucose line:
idx=find(strcmp(C{2},'Glucose'));

% convert the cell string to value:
SugVal.gluc=cellfun(@str2num,Data(idx))*0.18;
if isnan(SugVal.gluc)
    SugVal.gluc=0;
end

idx=find(strcmp(C{2},'Arabinose'));
if not(isempty(idx))
    SugVal.ara=cellfun(@str2num,Data(idx))*0.158;
else
    SugVal.ara=nan;
end

idx=find(strcmp(C{2},'Mannose'));
if not(isempty(idx))
    SugVal.man=cellfun(@str2num,Data(idx))*0.18;
else
     SugVal.man=nan;
end

idx=find(strcmp(C{2},'Xylose'));
if not(isempty(idx))
    SugVal.xyl=cellfun(@str2num,Data(idx))*0.158;
else
    SugVal.xyl=nan;
end

end

function [filesave,SampleName]=FindNewChromeleonFiles_HPLC(clk,par)
% this function read the cpu clock c clk at each iteration, and all the .txt files in the folder
% 'myfolder (change to the suitable chromeleon folder) and then only pick
% the .txt files that were created AFTER the clock registered c. 
% the output is file, the list of files to be read (cell array). 

filesave=[]; % 
TimeStamp=[];
SampleName=[];
%for debugging only!!:
%load c

% HERE, change with the correct Chromeleon folder
%if nargin<1
   %myfolder = 'C:\Users\Public\Documents\Lucullus\import\chromeleon'; % chromeleon folder! 
    %myfolder ='C:\Users\LAB\Desktop\ChromeleonResult'
    myfolder = par.Info.ChromeleonFolder;
    %end
%clk_0=par.clk_0;
% keyboard % for debugging only!
filesInfo=dir(myfolder);
expression = '.txt';

% Start the counting
n=1;

% no Chromeleon files, empty directory
if length(filesInfo)<4
    return
end

for z=3:length(filesInfo)   
    % get only the .txt files in a variable
    if strcmp(filesInfo(z).name(end-3:end),expression)
        
        name{n,:}=filesInfo(z).name;
        tid(n,:)=filesInfo(z).date;
        n=n+1;
    end
end

dato=datestr(tid,'ddmmyy');
%convert 'tid' to numbers:
% Days
for z=1:n-1
    day(z,1)=str2num(tid(z,1:2));
end

% Months
for z=1:n-1
  mon(z,1) = str2num(dato(z,3:4));
end

% Years
for z=1:n-1 
    year(z,1)=str2num(tid(z,8:11));
end

% Hours;
for z=1:n-1 % hours
    timer(z,1)=str2num(tid(z,13:14));
end

% Minutes
for z=1:n-1 % minutes
    minutter(z,1)=str2num(tid(z,16:17));
end

% Seconds
for z=1:n-1 % seconds
    sekunder(z,1)=str2num(tid(z,19:20));
end


ProcessNowSeconds=clk*[0 par.mth*60*60*24 60*60*24 60*60 60 1]';
ProcessStartSeconds=par.clk_0 * [0 par.mth*60*60*24 60*60*24 60*60 60 1]';

% Keep all the files that were created AFTER the clock
% day, month, year are the files, C is the clock.
%ct=1; % counter for the save files
for z=1:n-1
    TotalSeconds=[year(z) mon(z) day(z) timer(z) minutter(z) sekunder(z)]*[0 par.mth*60*60*24 60*60*24 60*60 60 1]';
    if TotalSeconds>ProcessNowSeconds-45
        SampleName= name{z,:};
        filesave=sprintf('%s%s',par.Info.ChromeleonFolder,SampleName);
        TimeStamp= TotalSeconds-ProcessStartSeconds;
        disp('NEW HPLC MEASUREMENTS ACQUIRED')
        return
    end
end

end

% Function to restart in case it is required
function restart(VarMem,par)

% In case the CO2 signal is not taking the correct initial value
VarMem.Instrument.CO2(1)= 0.04;
       

end

% Function to call the MHE
% Function to call Parameter Estimation
% Function to call the Metabolic model.