%% A2 pump 1 (A2)
% Set up the Import Options and import the data
opts = spreadsheetImportOptions("NumVariables", 2);
% Specify sheet and range
opts.Sheet = "Sheet1";
opts.DataRange = "A2:B173";

% Specify column names and types
opts.VariableNames = ["Var1", "Var2"];
opts.VariableTypes = ["double", "double"];

% Import the data
CalibrationpumpsA2 = readtable("C:\Code\Matlab\Master\Fall\Calibration\Calibration_pumps_A2.xlsx", opts, "UseExcel", false);


%% Clear temporary variables
clear opts

PumpPercent = CalibrationpumpsA2{2:end, 1}; % x
Flow_mL_min = -CalibrationpumpsA2{2:end, 2}; % y

Flow = Flow_mL_min*60/1000;


createFit(Flow, PumpPercent)


function [fitresult, gof] = createFit(Flow, PumpPercent)
%CREATEFIT(FLOW,PUMPPERCENT)
%  Create a fit.
%
%  Data for 'untitled fit 1' fit:
%      X Input : Flow
%      Y Output: PumpPercent
%  Output:
%      fitresult : a fit object representing the fit.
%      gof : structure with goodness-of fit info.
%
%  See also FIT, CFIT, SFIT.

%  Auto-generated by MATLAB on 26-Mar-2023 14:44:57


%% Fit: 'untitled fit 1'.
[xData, yData] = prepareCurveData( Flow, PumpPercent );


% Set up fittype and options.
ft = fittype( 'poly1' );
excludedPoints = excludedata( xData, yData, 'Indices', [62 66 70 84 106 147 153 155] );
opts = fitoptions( 'Method', 'LinearLeastSquares' );
opts.Lower = [-Inf 0];
opts.Robust = 'LAR';
opts.Upper = [Inf 0];
opts.Exclude = excludedPoints;


% Fit model to data.
[fitresult, gof] = fit( xData, yData, ft, opts );


figure( 'Name', 'Pump Calibration A2 2');
h = plot( fitresult, xData, yData, excludedPoints);

legend( h, 'Experimental', 'Excluded', strcat('Regression, y = ' , string(fitresult.p1), 'x'), 'Location', 'SouthEast', 'Interpreter', 'none' );
legend boxoff
% Label axes
xlim([-0.1,2.7])
xlabel( 'Flow [L/h]', 'Interpreter', 'none' );
ylabel( 'Pump [%]', 'Interpreter', 'none' );
title('Calibration Curve for Pump 1 A2')
h(1).MarkerSize = 10 ;
h(2).Color = [0.5 0.5 0.5];

% Errors
str = {[strcat('R^2 = ', string(gof.rsquare))], [strcat('RMSE = ' , string(gof.rmse))]};
annotation('textbox',  [0.144 0.66 0.3 0.25] , 'String', str, 'FitBoxToText','on', 'FontSize', 9, 'LineWidth', 0.4, 'LineStyle', '--')
grid off
end
