function model = model_DYCOPS(params)
addpath('C:\Code\Matlab\Casadi\casadi-windows-matlabR2016a-v3.5.5')
import casadi.*
model = struct;

% Fixed values for the parameters
mu_max = params.mu_max;	
Ks = params.Ks; 
kd =  params.kd; 
Y_XS = params.Y_XS; 
Y_CO2X = params.Y_CO2X; 

% Other
q_CO2 = params.q_CO2;
Sin = params.Sin;   % g/L

%% Model Equations
% Define variables
V_mod   = SX.sym('V');
X_mod   = SX.sym('X');
S_mod   = SX.sym('S');
CO2_mod = SX.sym('CO2');

x_mod   = [V_mod, X_mod, S_mod, CO2_mod]';

Fin_mod = SX.sym('Fin');
Fout_mod =  SX.sym('Fout'); %Fin_mod;


% Differential equations
dV_mod = Fin_mod - Fout_mod;
dX_mod = -(X_mod/V_mod)*Fin_mod + mu_max *(S_mod /(Ks + S_mod)) *X_mod - kd *X_mod ; %Biomass
dS_mod = (Sin *Fin_mod) / V_mod - mu_max *(S_mod /(Ks + S_mod)) *X_mod /Y_XS - (S_mod / V_mod) *Fin_mod; % Substrate
dCO2_mod = (mu_max *(S_mod /(Ks + S_mod)) *X_mod)/Y_CO2X - q_CO2*CO2_mod ;

xdot_mod = [dV_mod, dX_mod, dS_mod, dCO2_mod]';

% Define model variables
model.x    = x_mod;
model.xdot = xdot_mod;
model.u    = [Fin_mod, Fout_mod]'; 

end
