function MHEsys = MHEmodel_DYCOPS(x0, params)
addpath('C:\Code\Matlab\Casadi\casadi-windows-matlabR2016a-v3.5.5')

import casadi.*

MHEsys = struct;

%% States
V = MX.sym('V');
X = MX.sym('X');
S = MX.sym('S');
CO2 = MX.sym('CO2');

MHEsys.x = [V, X, S, CO2]';
%MHEsys.x0 = [1.5, 0.77519, 15, 0]'; % CHANGE bf : 1.5, 1.2, 20, 0
% param est
MHEsys.x0 =  x0; %[1.48,  0.8,  15,  0]';

MHEsys.x_lower = [1.2,   0,   0,   0]'; %%% !!! before:1.0
MHEsys.x_upper = [1.8, inf, inf, inf]'; %%% !!! before: 2.5

%% Inputs
Fin = MX.sym('Fin');
Fout = MX.sym('Fout'); %Fin;

MHEsys.u = [Fin, Fout]'; %MHEsys.u = [Fin, 
MHEsys.u0 = [0, 0]';

%% Measurements
% We assume that h(x)= Hx is a selector, such that y = x + noise
y_V = MX.sym('y_V');
y_X = MX.sym('y_X');
y_CO2 = MX.sym('y_CO2');

MHEsys.y = [y_V, y_X, y_CO2]';

%% Model Equations

% Here you can set fixed values for the parameters

mu_max = params.mu_max;	 %0.27399 ;%0.30711 ; 
Ks = params.Ks; %0.0050896; %0.0070378 ;
kd =  params.kd; %0.0013298; %0.0013996 ;
Y_XS = params.Y_XS; %0.12548 ; %0.39212  ;
Y_CO2X = params.Y_CO2X; %0.28222; %0.055265; 

% Other
q_CO2 = params.q_CO2;

% Sin = 100; % g/L
Sin = params.Sin;   % g/L


dV = Fin - Fout;
dX = -(X ./V) .*Fin + mu_max .*(S ./(Ks + S)) .*X - kd .*X; %Biomass
dS = (Sin .*Fin) ./ V - mu_max .*(S ./(Ks + S)) .*X ./Y_XS - (S ./ V) .*Fin; % Substrate
dCO2 = (mu_max .*(S ./(Ks + S)) .*X)./Y_CO2X - q_CO2*CO2; 

MHEsys.x_dot = [dV, dX, dS, dCO2]';

%% Model noise

% This section is to compute the Jacobian df(x,p)/dp

% Differential equations:
w_mu = MX.sym('w_mu'); 
w_ks = MX.sym('w_ks');
w_Yxs = MX.sym('w_Yxs'); 
w_YCO2 = MX.sym('w_YCO2'); 
w_V = MX.sym('w_V'); 
w_X = MX.sym('w_X'); 
w_S = MX.sym('w_S');
w_CO2 = MX.sym('w_CO2'); 
w_kd = MX.sym('w_kd');

w_mu0 = MX.sym('w_mu0'); 
w_ks0 = MX.sym('w_ks0');
w_Yxs0 = MX.sym('w_Yxs0'); 
w_YCO20 = MX.sym('w_YCO20'); 
w_V0 = MX.sym('w_V0'); 
w_X0 = MX.sym('w_X0'); 
w_S0 = MX.sym('w_S0');
w_CO20 = MX.sym('w_CO20'); 
w_kd0 = MX.sym('w_kd0');


dV_w = Fin - Fout + w_V;
dX_w = -(X ./V) .*Fin + (mu_max + w_mu) .*(S ./((Ks + w_ks) + S)) .*X - (kd + w_kd) .*X + w_X; %Biomass
dS_w = (Sin .*Fin) ./ V - (mu_max + w_mu) .*(S ./((Ks + w_ks) + S)) .*X ./(Y_XS + w_Yxs) - (S ./ V) .*Fin + w_S; % Substrate
dCO2_w = ((mu_max + w_mu) .*(S ./((Ks + w_ks) + S)) .*X)./(Y_CO2X + w_YCO2) - 2*CO2 + w_CO2;

dV_w0 = Fin - Fout + w_V0;
dX_w0 = -(X ./V) .*Fin + (mu_max + w_mu0) .*(S ./((Ks + w_ks0) + S)) .*X - (kd + w_kd0) .*X + w_X0; %Biomass
dS_w0 = (Sin .*Fin) ./ V - (mu_max + w_mu0) .*(S ./((Ks + w_ks0) + S)) .*X ./(Y_XS + w_Yxs0) - (S ./ V) .*Fin + w_S0; % Substrate
dCO2_w0 = ((mu_max + w_mu0) .*(S ./((Ks + w_ks0) + S)) .*X)./(Y_CO2X + w_YCO20) - 2*CO2 + w_CO20;

MHEsys.x_dot_w = [dV_w, dX_w, dS_w, dCO2_w]';
MHEsys.w_Q = [w_mu; w_ks; w_Yxs; w_YCO2; w_V; w_X; w_S; w_CO2; w_kd];

MHEsys.x_dot_w0 = [dV_w0, dX_w0, dS_w0, dCO2_w0]';
MHEsys.w_Q0 = [w_mu0; w_ks0; w_Yxs0; w_YCO20; w_V0; w_X0; w_S0; w_CO20; w_kd0];

G = jacobian(MHEsys.x_dot_w0, MHEsys.w_Q0);
Q_w = diag([w_mu; w_ks; w_Yxs; w_YCO2; w_V; w_X; w_S; w_CO2; w_kd]);
Q_out = G*Q_w*G';
MHEsys.Q_out = Q_out;

L1 = Function('Jtest',{X, S, w_mu, w_ks, w_Yxs, w_YCO2, w_V, w_X, w_S, w_CO2, w_kd, w_mu0, w_ks0, w_Yxs0, w_YCO20, w_V0, w_X0, w_S0, w_CO20, w_kd0},{Q_out});

MHEsys.Q_out = L1;

%% Weighting
Q0_x = [2.0946e-08, 1.0975e-05, 1.0852e-04, 2.1669e-05]'; % Covariance for initial values of x   % CHANGE

%P_in_01.0e+03 *[ 6.9095, 0.3019, 0.0960, 0.2148]
%P iter 2
        %9.9602         0         0         0
        % 0    2.8759    0.0000   -0.0024
        % 0         0   -0.0104    0.0000
        % 0         0         0   -0.7198
MHEsys.P0 = diag(Q0_x.^(-1/2));
R = [0.1, 0.1, 0.001]'; % 0.01, 0.1, 1 Covariance of measurement noise % CO2 =1 ! !     % CHANGE
MHEsys.V = diag(R.^(-1/2)); % R Helene prev: R = [0.01, 1, 1]'

%% Objective function
yVector = [y_V, y_X, y_CO2]'; % Measurements
xVector = [V, X, CO2]'; % Corresponding states

MHEsys.costfun = (yVector-xVector)'*MHEsys.V'*MHEsys.V*(yVector-xVector);
end