addpath('C:\Code\Matlab\Master\Spring')
addpath('C:\Code\Matlab\Master\Spring\Results')

% Get system parameters
system = system_param();

%% Sugar
filename_S = 'A2_sugar_20230329_HPLC.xlsx'; %% UPDATE WITH CORRECT FILE

sugar_table = readtable(filename_S, 'Sheet','Corrected_sugar');
sugar_results = table2array(sugar_table(:,2:3)); 

time_s = sugar_results(:,1);
sugar = sugar_results(:,2);
%% Offline CDW
filename = 'Result_29_03.xlsx'; %% UPDATE WITH CORRECT FILE

lab_result_table = readtable(filename, 'Sheet','Result_29_03');
lab_results = table2array(lab_result_table(:,4:10)); 

dryweight = [lab_results(1:7,4:6)-lab_results(1:7,1:3)]; % Triplicates

time_CDW = lab_results(1:7,7); 

cell_density = dryweight*1000/3; % g/L, 3 mL samples -> x g in 3mL  -> y g/L

mean_CDW = mean(cell_density,2);
std_CDW = std(cell_density,0, 2);


%% Get intended feed
intended = importdata('Values_feed_09_03_23.mat');

%% Get experimental data 
% load experimental data
% open .mat file first seperately in matlab before to see full structure
% use "clear da" and "clear data" in console for avoiding warnings
exp_data = load('20230329_MHE_None_Glucose_A2.mat', 'VarMem');

VarMem = exp_data.VarMem;
time= VarMem.ProcTime/3600;

% Substrate information
time_HPLC= VarMem.Instrument.HPLC.timeHPLC/3600;

% MHE
x_est= [VarMem.MHESim.V', VarMem.MHESim.X', VarMem.MHESim.S', VarMem.MHESim.CO2'];
%x_model = [VarMem.ModSim.V', VarMem.ModSim.X', VarMem.ModSim.S', VarMem.ModSim.CO2'];

x0 = x_est(1,:)';
time_lab = VarMem.ProcTime/3600;

%Feed 
Fin     = VarMem.Instrument.Flow_in';
Fout    = VarMem.Instrument.Flow_out';

us = [Fin, Fout];


%% MODEL
%% Integrate system model
model = model_DYCOPS(system);

dt         = system.dt;
N_points   = length(time);
model.T    = N_points/dt;
model.dt   = dt;

x_model = zeros(N_points,4);
x_model(1, :) = x0;

% GET INTEGRATOR
model.noise = 0;
[ode_sol]    = ODE_solver(model); 
xk = x0;
t_model = zeros(N_points,1);
% Integrate model
for i= 2:N_points
    % Get previous state
    xk = [x_model(i-1,1), x_model(i-1,2), x_model(i-1,3), x_model(i-1,4)];
    
    sim = ode_sol('x0', xk, 'p', [us(i-1, :)] ); % [us(i)] ); % must check to use us(i) or us(i-1)
    
    xk_mod = full(sim.xf); % y;
    
    % Update states
    x_model(i,:)   = xk_mod;   
    t_model(i) =  t_model(i-1) + dt;
end


%% RSME

% Offline CDW
t_CDW = zeros(length(mean_CDW),1);
mod_CDW = zeros(length(mean_CDW),1);
MHE_CDW = zeros(length(mean_CDW),1);


for i = 1:length(mean_CDW)
    i_CDW = find(time_lab <= time_CDW(i), 1, 'last');
    t_CDW(i) = time_lab(i_CDW);
    mod_CDW(i) = x_model(i_CDW,2);
    MHE_CDW(i) = x_est(i_CDW,2);
end


% Offline Sugar
% Get sugar values (datarows without NaN values)

S = sugar; %VarMem.measurements(:,3);

[row, col] = find(isnan(S));

S( row )  = [];
% Timestamp for sugar measurements
ts_S = time_s; %time_HPLC;
ts_S( row )  = [];

mod_S = zeros(length(S),1);
MHE_S = zeros(length(S),1);

for i = 1:length(ts_S)
    i_S = find(time_lab <= ts_S(i), 1, 'last');
    t_CDW(i) = time_lab(i_S);
    mod_S(i) = x_model(i_S, 3);
    MHE_S(i) = x_est(i_S, 3);
end


% Calculate Root Mean Squared Error -> MHE- X, MHE - CDW offline , MHE - S offline
%MS_model    = [ sum((x_model(:,2) - X).^2) sum((mod_CDW - mean_CDW).^2) sum((S_model -  S).^2)];
%MS_MHE    = [ sum((x_est(:,2) - X).^2) sum((MHE_CDW - mean_CDW).^2) sum((S_est -  S).^2)];

MS_model    = [ sum((mod_CDW - mean_CDW).^2) sum((mod_S -  S).^2)];
MS_MHE    = [ sum((MHE_CDW - mean_CDW).^2) sum((MHE_S -  S).^2)];

RMSE_model  = sqrt(MS_model./[length(mean_CDW)  length(S)]); %sqrt(sum(MS_model)/N_points);
RMSE_MHE    = sqrt(MS_MHE./[length(mean_CDW)  length(S)]);%sqrt(sum(MS_MHE)/N_points);
disp("Root mean squared error, model:")
disp(RMSE_model)
disp("Root mean squared error, MHE:")
disp(RMSE_MHE)

%% Plotting
% Get colors
addpath('C:\Code\Matlab\Master\Spring')
color = colorHex();

clf;

 set(0, 'DefaultTextFontName', 'Computer Modern',...
     'DefaultTextFontSize', 10,...
     'DefaultAxesFontName', 'Computer Modern',...
     'DefaultAxesFontSize', 10,...
     'DefaultLineLineWidth', 1.2)

set(gcf,'Position',[450 300 650 580])



tile  = tiledlayout(2,2,'TileSpacing','Compact','Padding','Compact');

%% Volume
nexttile
    hold on
    plot(time,x_model(:,1), '--','Color', 'k')
    plot(time,VarMem.Instrument.V,            '-.','Color', [0.5 0.5 0.5])
    plot(time, VarMem.MHESim.V,   '-','Color', color.b)
    yline(1.5,'--', 'Desired')

    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('V [L]')
    %legend('boxoff')
   % legend('Model','MHE','Measurement','location', 'best')

% X
nexttile
    hold on
    
    
    plot(time,VarMem.Instrument.CDWcu,            '-.','Color', [0.5 0.5 0.5])
    plot(time,x_model(:,2), '--','Color', 'k')
    plot(time, VarMem.MHESim.X,   '-','Color', color.b)
    plot(time_CDW, mean_CDW, 'x', 'Color', color.r)
    errorbar(time_CDW, mean_CDW, std_CDW, 'Color', color.p, 'LineStyle', 'none')
    % MANUAL SAMPLES
    %err = std_CDW;
    %errorbar(time,y,err,'ob')

    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('X [g/L]')
    legend('boxoff')
    legend('Measurement','Model', 'MHE', 'Offline','location', 'SouthEast')
    %legend('Model','Measurement','MHE','location', 'best')

%% Sugar
nexttile
    hold on
    plot(time,x_model(:,3), '--','Color', 'k')
    plot(time, VarMem.MHESim.S,   '-','Color', color.b)
    %plot(t_S, S,'x','Color', color.p)
    %plot(time_HPLC,VarMem.measurements(:,3), 'x','Color', color.r)
    plot(time_s,sugar, 'x','Color', color.r)
    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('S [g/L]')
    %legend('boxoff')
    %legend('Model','MHE','Offline','location', 'best')    %'off-line S',

%   CO2
nexttile
    hold on
    plot(time,x_model(:,4), '--','Color', 'k')
    plot(time,VarMem.Instrument.CO2_measurement,          '-.','Color', [0.5 0.5 0.5])
    plot(time, VarMem.MHESim.CO2,   '-','Color', color.b)
    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('CO2 [% of gas outlet]')
    %legend('boxoff')
    %legend('Model','MHE','Measurement','location', 'best')
    
sgtitle("MHE real-time run 1")

%% Feeding
fig = figure; clf
set(gcf,'Position',[450 300 550 200])
box off
    hold on
    plot(time, VarMem.Instrument.Flow_in,'Color', color.b)
    plot(time, VarMem.Instrument.Flow_out, '-.', 'Color', color.r)
    plot(intended.time, intended.feed, '--', 'Color', [0.5 0.5 0.5])
    xlabel('Time [h]')
    ylabel('Flow [L/h]')
    legend('F_{in}', 'F_{out}', 'Intended') 
    title("Flow in & out, real-time run 1", 'FontSize', 10,'FontWeight','Normal')
%% Feeding & O2

% create new figure
fig = figure; clf
tile  = tiledlayout(2,1,'TileSpacing','Compact','Padding','Compact');
nexttile
    box off
    hold on
    plot(time, VarMem.Instrument.Flow_in,'Color', color.b)
    plot(time, VarMem.Instrument.Flow_out, '-.', 'Color', color.r)
    plot(intended.time, intended.feed, '--', 'Color', [0.5 0.5 0.5])
    xlabel('Time [h]')
    ylabel('Flow [L/h]')
    legend('F_{in}', 'F_{out}', 'Intended') 
    title("Flow in & out")
nexttile
    box off
    hold on
    plot(time, VarMem.Instrument.pO2,'Color', color.b)
    xlabel('Time [h]')
    ylabel('pO2')
    legend('pO2') 
    title("Oxygen level in reactor")
     
sgtitle("MHE real-time run 1")
 
