%% General function to call OPC and to start the biochemical process
function [savefile]=main_meas_v1_MHE_A2(simulation)
% change later to Runme.m
% the main function for control 
% everything is saved in SAVEFILE in the EXPERIMENTS folder.
% "par.filter_step" determined how often (how many while loop) before the EKF is
% running. for example, if step = 5, the EKF will run calculation every
% iterationTime * 5, i.e. every fifth loop iteration. 
% default: Preferable step=1 (each time step)
% including reading the automativ HPLC measurement from the Chromeleon
% files in 'C:\Users\Public\Documents\Lucullus\import\chromeleon'
% The procedure check the time in the beginning of the experiments (clk)
% and only check for files created after this time. 
% The output is Instrument.HPLC.SugVal.Man/Xyl/Glu/Ara
% if you call the function with value, then simulation ==1 which implies
% not reading data from real-time

%% Import Casadi
addpath('C:\Matlab\Helene\Casadi\casadi-windows-matlabR2016a-v3.5.5')
import casadi.*
    
%% CHECK:
% check whether the NIR and the cdw are correctly measured and defined.

%% Get feeding profile
% !OBS par.iterationtime set to 60s and par.dt set to 1/60 as below
disp('Get feed profile')
load('Values_feed_09_03_23.mat')
Values_feeding_profile = Values_feed;
time_profile = Values_feeding_profile.time;
feed_profile = Values_feeding_profile.feed;

% warning('off', 'all')
%warning('on', 'all')

%% Connect to the OPC server
if nargin==1
     %% For computer in-silico simulations, uncomment the next line:
 %   load ('Data_20210908_EKF_sim')
    simulation=1;
else
    simulation=0;
    %% For real OPC time use this option
    % Connect to OPC Server
    da = opcda('localhost', 'Securecell.OpcBridgeServer.DA.1'); 
    connect(da); 
    BioReactor = 'A2';
    [readGrpA2, writeGrpA2, indexesA2] = findReactorTags(BioReactor, da);
    %[readGrpHPLC,~, indexesHPLC] = findSampleInfo(reactor, da);
    data = read(readGrpA2);
end

close all


%% Initialize the code
% The following creates the file name to be saved at each iteration (saving
% the data), according to the current date
  

clk=clock;
clk_0=fix(clk);  
      % INITIAL clock time update, 
clk_now=clock;
clk_now=fix(clk_now);
%% Define the initial loop
i=1;
VarMem=  [];

if simulation <1
    % Initialize the memory    
    % TODO 
    
    iteration= i;
    
    par=GetParValues(BioReactor,[],clk_0);
    % Get the instrument information
    [Instrument0]=GetInstData(data,indexesA2,clk_now,par,iteration);
    par=GetParValues(BioReactor,Instrument0,clk_0);
    
    Instrument0.V = par.V0; %Initial reactor volume (user defined) 
   
    % Define the initial conditions
    x0= [par.V0 par.X0 par.S0 par.CO20]'; % CHANGE DURING INITIAL RUN
    
    % Initializing for the first estimator
    
    
    %% Setup MHE
    % MHE model and MHE properties
    MHEsys = MHEmodel_DYCOPS(x0, par);
    MHEproblemProps = MHEprops_DYCOPS(MHEsys);
    
    % MHE horizon
    N_MHE = MHEproblemProps.N;
    
    % Estimator
    [MHEprob, estimator] = moving_horizon_estimation_DYCOPS(MHEsys, MHEproblemProps, x0');

    F_integrator = makePlantIntegrator_DYCOPS(MHEsys,MHEproblemProps);
    
    % Create an NLP solver
    w = MHEprob.w;
    g = MHEprob.g;

    MHE_initguess = MHEprob.w0;
    
    % Simulate system
    x_bar = x0;
    x_ode = x0;
    P = MHEsys.P0;
    
    % Horizon values
    y_V = zeros(N_MHE,1);
    y_X = zeros(N_MHE,1);
    y_CO2 = zeros(N_MHE,1);
    U_in =  zeros(N_MHE,2); 
    flag_sol = zeros(N_MHE,1);
    
    % Weights
    wg      = par.std_dev.^2;  % variance %= noise.w; % White gaussian noise

    w_mu = wg(1); w_ks = wg(2); w_kd = wg(3); w_Yxs = wg(4); w_YCO2 = wg(5);

    w_V = 1.0e-2;   w_X = 9.0e-2;   w_S =  1.0e-2;  w_CO2 = 1.0e-1;    

    w_mu0= 0; w_ks0 = 0; w_Yxs0 = 0; w_YCO20 = 0; w_V0 = 0; w_X0 = 0; w_S0 = 0; w_CO20 = 0; w_kd0 = 0;

    % Error covariance
    Q = diag(P).^(-2);
    Q =(diag(Q));
    
    %% Model integrator
    model = model_DYCOPS(par);
    model.dt   = par.dt;

    % GET INTEGRATOR
    [ode_sol]    = ODE_solver(model); 

    %% BEGIN SANITY CHECK LIST!
   % sprintf('Running sannity check... Please Wait\n----------------------------------\n')
    % check the following in Instrument0
    % Instrument.Balance_A2
   % if isempty
    % Instrument.Balance_B2
    % Balance C
    % Balance D
    % Instrument.NIR
    % Instrument.Air_flow= double(data(indexesA2.measurements.AirFlow).Value);
    % Instrument.CO2
    % Instrument.O2
    % Instrument.Pump_1
    % Instrument.Pump_2
    % Instrument.Acid_pump
    % Instrument.Base_pump
    % Instrument.pO2
    % Instrument.NowSampleBarCode
    % Instrument.NowSampleName
    % Instrument.NowSampleBioreactor
    
    
    % xhatukf = xhat; % only if using UKF
    %Q_ukf = diag([0.0001 0.0000004 0.000002 0.001]);
else
    
end

% if you stop the code in the middle of experiments and want to proceed 
%from same place, you can:
% 1. Run the code until here (debug here)
% 2. load the most recent SaveFile by the following command:
% load (SaveFile,'VarMem','i','Instrument0','Instrument','par','x0','MPC')
% For example:
% load ('20220928_EKF_PID_Glucose_A2.mat','VarMem','i','Instrument0','Instrument','par','x0','MPC')
% 3. continue, remove the debug point

%% Enter in the loop
 while true 
    tic
    
    % DATA EXTRACTION
    % Read the data
     
        if i==1 
           ProcTime= 0;
           iteration=1;
        else
           iteration=i;
           % We need to update the ProcTime always!!!   - Done in SaveMemory         

%            ProcTime= (clk_now*[0 0 60*60*24 60*60 60 1]'-par.clk_0*[0 0 60*60*24 60*60 60 1]');
%            Instrument.ProcTime= ProcTime;

        end       
        
        data = read(readGrpA2);
        %get HPLC data from OPC Lucullus
        %dataHPLC = read(readGrpHPLC);        
    
       % Extract data from the instrument
       [Instrument]=GetInstData(data,indexesA2,clk_now,par,iteration);
       
       %% CLEANING REVERSE FLUSH
       % WHAT DOES THIS DO?
       if iteration>1
           [C] = intersect(VarMem.Instrument.SampleDataBase(:,1),Instrument.NowSampleBarCode,'stable');
           if not(isempty(Instrument.NowSampleBarCode)) && isempty(C)
               % new sample was just taken
               disp('FLUSHING....')
               %write(writeGrpA2.Item(indexesA2.setpoints.Feed_Ext_3_Setpoint), {6})
               write(writeGrpA2.Item(indexesA2.setpoints.Feed_Ext_3_Setpoint), {10})
               % the value 5 is about 10mL in 30 second... about 10 dropse
           else
               write(writeGrpA2.Item(indexesA2.setpoints.Feed_Ext_3_Setpoint), {0})
           end
       end
%%
       % Get HPLC measurements from file, if exist:
       [Instrument,VarMem]=GetHPLCFromFile(Instrument,clk_now,par,VarMem,iteration);
       
       %% Instrumentation additional analysis additional
       [Instrument,par]= Analysis_instrumentation(Instrument,Instrument0,par,iteration,VarMem);
       
       %% Measurement vector with HPLC data and the outlier identification algorithm
       % [measurements, VarMem,Instrument]= HPLC_data(Instrument,VarMem,iteration,ProcTime);
       [measurements,VarMem,Instrument]= HPLC_data(Instrument,VarMem,iteration,Instrument.ProcTime,par);
       

%% Control action Feed
    if i > length(feed_profile)
        feed_i = 0;
    else
        % CSTR
        %feed_i = feed_profile(i); %L/h 
        feed_i = 0; % !! CHANGE
    end
    % Fin = Fout
    Controller.Fin= feed_i;

    % Use the calibration data to transform Flow into % pump   
    Controller.Fin2pump= par.pump.m1*feed_i+par.pump.b2;

     % Fin
     write(writeGrpA2.Item(indexesA2.setpoints.Feed_Ext_1_Setpoint), {Controller.Fin2pump});
     % Fout
     write(writeGrpA2.Item(indexesA2.setpoints.Feed_Ext_2_Setpoint), {Controller.Fin2pump});

     %% Current flow
     ui = [Instrument.Flow_in, Instrument.Flow_out]'; % CHECK ' or not
     
     %% Modelintegration
     [xk,ModSim] = Process_model(Instrument,x0,ode_sol, VarMem,iteration, ui);
     
     %% MHE
     yk = measurements';
     W = Q^(-1/2);

     W_Array(:,:,i) = W;
     
     
     % GET MEASUREMENTS
    if i <= N_MHE % before first horizon -> Filling up measurements
        y_V(i:end)    = yk(1);      
        y_X(i:end)    = yk(2);       
        y_CO2(i:end)  = yk(4);      
        U_in(i,:) = ui;
    else
        y_V(1:end-1) = y_V(2:end);
        y_V(end) = yk(1);

        y_X(1:end-1) = y_X(2:end);
        y_X(end) = yk(2);

        y_CO2(1:end-1) = y_CO2(2:end);
        y_CO2(end) = yk(4);

        U_in(1:end-1, :) = U_in(2:end, :);
        U_in(end, :) = ui; 
    end
    
    y = [y_V, y_X, y_CO2];

    if i ~= 1

        [xL, PL, PLEig] = arrivalCostUpdate_DYCOPS(MHEsys, MHEproblemProps, x, y, U_in, P, Q, F_integrator);

        x_residual(:,i-1) = xL-x(:,1);
        P_residual(:,:,i-1) = PL-P;
        P_eig(:,i-1) = PLEig;

        x_bar = xL;
        P = PL;
    end
    
    Y_in = vertcat(y_V,y_X,y_CO2); 
    x_bar_in = x_bar;
    P_in = reshape(P,[length(P).^2,1]);
    U_in_in = reshape(U_in,[N_MHE * 2, 1]);
    
    % Update W
    if iteration <= N_MHE
        for iter = 1:N_MHE
            W_Array_presample(:,:,iter) = W;
        end
        W_in = reshape(W_Array_presample(:,:,:),[N_MHE*(length(W).^2),1]);
    else
        W_in = reshape(W_Array(:,:,i-N_MHE+1:i),[N_MHE*(length(W).^2),1]);
    end
    
    % Estimation
    MHE_variables = vertcat(Y_in, U_in_in, x_bar_in, P_in, W_in);

    % Solve the NLP
    sol = estimator('x0', MHE_initguess, 'p', MHE_variables, 'lbx', MHEprob.lbw, 'ubx', MHEprob.ubw,...
                'lbg', MHEprob.lbg, 'ubg', MHEprob.ubg);
    w_opt = full(sol.x);
    MHE_initguess = w_opt;

    [x, d] = sortResultsMHE_DYCOPS(w_opt, MHEsys, MHEproblemProps);

    flag_sol(i) = estimator.stats.success;
    
    % Update data
    if iteration < N_MHE
        x_est(:,i)     = x(:,i); 
        MHESim.V       = x(1,i);
        MHESim.X       = x(2,i);
        MHESim.S       = x(3,i);
        MHESim.CO2     = x(4,i);
        %t = [t, time]; % REMOVE ?
    else
        x_est          = [x_est, x(:,end)]; 
        MHESim.V       = x(1,end);
        MHESim.X       = x(2,end);
        MHESim.S       = x(3,end);
        MHESim.CO2     = x(4,end);
        %t = [t, time]; % REMOVE ?
    end


    if iteration < N_MHE
        d_est(:,i) = d(:,i);
    else
        d_est = [d_est, d(:,end)];
    end

    for j = 1:size(x0,1)
    d_Array(j,i) = d(j,:)*d(j,:)';
    end

    P = P'*P;
    PArray(:,i) = diag(P);
    
    L1_cas = MHEsys.Q_out(x_est(2,end), x_est(3,end), w_mu, w_ks, w_Yxs, w_YCO2, w_V, w_X, w_S, w_CO2, w_kd, w_mu0, w_ks0, w_Yxs0, w_YCO20, w_V0, w_X0, w_S0, w_CO20, w_kd0);
    Q = full(L1_cas);
     

%% Save data into VarMem function
% Input is the Intrument0 also and initialize answer to not erase data
if i==1
    % First iteration
    VarMem=SaveVar2Memory(i,par,measurements,Instrument0,Instrument,Controller, ModSim, MHESim, VarMem);
else % send VarMem to the function
    VarMem=SaveVar2Memory(i,par,measurements,Instrument0,Instrument,Controller,ModSim,  MHESim, VarMem);
end

    
%%  Plotting with an external function
plotting_VarMem(VarMem,par,iteration);
usedTime = toc;
save(par.Info.SaveFile)

%remove the HPLC measurement before the next iteration, because each HPLC
%measurement only used once.
  Instrument.HPLC.gluc=[];
    Instrument.HPLC.xyl= [];
    Instrument.HPLC.man= [];
    Instrument.HPLC.ara= [];
     Instrument.HPLC.TimeStamp=[];
% Note: it must be Iteration time> TOC    g
% Otherwise it does not pause
 
% IMPORTANT: update of the new clock must be BEFORE the pause, otherwise
% any new HPLC file will not be recognized (because the comparison with the
% time stamp of the new file and the clock). 
% ONLY HERE!
clk_now=clock;
clk_now=fix(clk_now);

% Display current iteration
%it = strcat("iteration = ", num2str(i));
disp("Iteration")
disp(iteration)
pause(par.iterationTime-usedTime);



%pause(10) % CHANGE 

i = i+1; % advance one step


end

disconnect(da)
delete(da)


end

% Function to obtain the parameters of the instrument
function par=GetParValues(BioReactor,Instrument0,clk_0)
% All the parameters of the reaction are packed here!

% % information about the experiment, file and folder
  par.Info.ChromeleonFolder ='C:\Users\LAB\Desktop\ChromeleonResult\'; % chromeleon folder!
  par.Info.Substrate = 'Glucose';
  par.Info.Control = 'None';
  par.Info.Filter = 'MHE';
  par.Info.date = datestr(now,'yyyymmdd');
  par.Info.BioReactor=BioReactor;
  par.Info.SaveFile=sprintf('%s_%s_%s_%s_%s.mat',par.Info.date,par.Info.Filter,par.Info.Control,par.Info.Substrate,par.Info.BioReactor);
  par.Info.LoopMethod= 'None';
  %par.Info.ChromeFile=sprintf('%s\\%s_%s_%s_%s.mat',par.Info.ChromeleonFolder,par.Info.date,par.Info.Filter,par.Info.Control,par.Info.Substrate,par.Info.BioReactor);

  % The initial time in the system is:
% Compute it as hrs always
% Take into consideration the hours, minutes and seconds of today.
par.clk_0= clk_0;
switch clk_0(2)
    case {1,3,5,7,8,10,12}
        par.mth=31;
    case 2
        par.mth=28; % CAREFULL!!! it can be 29 in extended years
    case {4,6,9,11}
        par.mth=30;
end

% The initial time is in hrs
par.initial_time= clk_0*[0 par.mth*60*60*24 60*60*24 60*60 60 1]'/3600;
par.iterationTime =60; % in seconds

par.dt= 1/60; % the discrete time step in hours

% the delay time in minutes from the sample taken to the chromeleon wrote
% the file:
% OBSOLETE
par.NUMERA_HPLC_SamplingTime=55; % in minutes
par.NUMERA_HPLC_SamplingTime=par.NUMERA_HPLC_SamplingTime * 60; % convert to seconds

% "step" determinefd hsm often (how many while loop) before the EKF is
% running. for example, if step = 5, the EKF will run calculation every
% iterationTime * 5, i.e. every fifth loop iteration. 
% default: Preferable step=1 (each time step)
% Initial conditions:
par.V0= 1.5;
par.S0= 11; %15; % initial g/L sugar in the reactor
par.flowDensity = 1078.5; % DENSITY

if nargin>1 && not(isempty(Instrument0)) % if there is Instrument0   
% Initial measurement of CO2
par.CO2inAir_volpercent = Instrument0.CO2; 
par.CO2inAir_volpercent = 0.04; 


% Density of CO2 in air
CO2_density = 1.77;

% No CO2 at the beginning
CO20= 0;
CarbDiox = CO20/par.V0;
par.CO20= 0;

% safeguard of the balances A,B,C and D
% if the balances reset by mistake, it will be ignored.
% the baseline weight is first zero. if a reset happens, the baselineWeight
% will be changed in the Analysis_Instrumentation function.
if isfield(Instrument0,'Balance_A2') %  'Balance_A' CHANGE ->'Balance_A2' ? 
par.baselineWeight.A2=Instrument0.Balance_A2; %  'Balance_A' CHANGE ->'Balance_A2' ? 
par.baselineWeight.B2=Instrument0.Balance_B2; % par.baselineWeight.B -> B2
else
par.baselineWeight.C=Instrument0.Balance_C;
par.baselineWeight.D=Instrument0.Balance_D;
end

% Biomass corelation - valid for all NIR ranges
CDWcu = -15212.545*sin(2.0532*Instrument0.NIR  +2.5893) +15197.4245*sin(2.055*Instrument0.NIR +2.5887); 

%22.187 * Instrument0.NIR - 5.0991 ;
% CDWcu = %11.124 * Instrument0.NIR + 0.66116 ;


% No extra big NIR values (probe disconnected)
if Instrument0.NIR >= 5
    CDWcu= 0.66116;
end

% No negative NIR values (no biomasss in the system) and no negative CDWcu,
% CDWcu (0.0136281815728) => negative
if Instrument0.NIR <= 0.0136281815728
    CDWcu= 0;
end

% Biomass concentration
par.X0= CDWcu; % CDW in g/L (OD 1)

% Concentration of the feeding
par.Sin= 200; % g/L 

% set points: (NOT ACTIVE)

% Set point of the Volume if controller
par.sp.V= 1.5;

% Set point of the CDWcu if controller
par.sp.CDWcu= 15;

% airflow
par.q_air   = 2;       % 2 NL/min
par.q_CO2   = par.q_air * 1.8* 60 / 100; 

% Wild type C Glutamicum  ATCC13032
par.mu_max  = 0.2296;       % 0.19445;   % h^-1
par.Ks      = 0.0077556;    % 0.007;    % g L^-1
par.kd      = 0.0025183;    % 0.006;
par.Y_XS    = 0.42931;      % 0.42042; 
par.Y_CO2X  = 0.39387; 

par.std_dev = [0.0001, 9.4749e-05, 4.4482e-05 , 0.0050084, 0.00010611];

% Calibration value to merge NIR and CO2
par.alpha = 0.6655;
   

% Parameter for the pumps
% linear conversion of the flow to opening of external WM pumps
par.pump.m1= 45.2176; %45.0196; % Pump 1 %Prevoiusly 37.03
par.pump.b1= 0;

% Parameters for pump 2
par.pump.m2= 44.7667;%45.0196; % pump 2
par.pump.b2= 0;



end

end

% Function to find the OPC ports and the indexes
function [readGrp, writeGrp, indexes] = findReactorTags(reactor, da)

% Connect to OPC Server 
%da = opcda('localhost', 'Securecell.OpcBridgeServer.DA.1'); 
%connect(da); 
itmIDs = serveritems(da); 


reactorIDs = [];

% Find all values related to the reactor
for i = 1:length(itmIDs)
    str = [char(itmIDs(i)), ' '];   % Convert cell string to char, add space
                                    % for easier parsing (making sure we
                                    % always have 2 characters)
    if length(reactor) == 2
        if str(1:2) == reactor
            reactorIDs = [reactorIDs; itmIDs(i)];
        end                         
    elseif reactor == 'A' || reactor == 'B'
        if str(1) == reactor && str(2) ~= '2'
            reactorIDs = [reactorIDs; itmIDs(i)];
        end
    else
        if str(1) == reactor
            reactorIDs = [reactorIDs; itmIDs(i)];
        end
    end
end



%  {'Numera.Numera.SEQUENCE_1.TimeStampSec_NU'                   }
%     {'Numera.Numera.SEQUENCE_1.TimeStampHour_NU'  

if isempty(reactorIDs)
    error("No tags were found for reactor """+convertCharsToStrings(reactor)+""". Have you inserted a valid reactor name? Use characters and capital letters, like 'A'.")
end

% Additional instrumentation
balances = reactorIDs(contains(reactorIDs, 'Balance'));
buglab = reactorIDs(contains(reactorIDs, 'Buglab'));
HPLC= reactorIDs(contains(reactorIDs,'HPLC'));

% Take the time stamp sf the last sample of Numera to HPLC
HPLC=[HPLC ; 
    {'Numera.Numera.SEQUENCE_1.TimeStampYear_NU'};
    {'Numera.Numera.SEQUENCE_1.TimeStampMonth_NU'};
    {'Numera.Numera.SEQUENCE_1.TimeStampDay_NU'};
    {'Numera.Numera.SEQUENCE_1.TimeStampHour_NU'};
    {'Numera.Numera.SEQUENCE_1.TimeStampMinute_NU'};% 
    {'Numera.Numera.SEQUENCE_1.TimeStampSec_NU'}];


measurements = reactorIDs(contains(reactorIDs, 'Measured Values'));
controlOutputs = reactorIDs(contains(reactorIDs, 'Controller Output'));
setpoints = reactorIDs(contains(reactorIDs, 'Setpoints'));
sysDevice = reactorIDs(contains(reactorIDs, 'System Device.SystemSubDev'));

% Digging out the timers
HPLC = HPLC(2:end);
measurements = measurements(2:end);
controlOutputs = controlOutputs(2:end);
sysDevice = sysDevice(2:end);
setpoints = setpoints(2:end);

% Adding the scales and HPLC to measurements
Balance_A2 = balances(contains(balances, {'Balance A2.Balance Ports.Weight'}));
Balance_B2 = balances(contains(balances, {'Balance B2.Balance Ports.Weight'}));
Balance_C = balances(contains(balances, {'Balance C.Balance Ports.Weight'}));
Balance_D = balances(contains(balances, {'Balance D.Balance Ports.Weight'}));
measurements = [measurements; Balance_A2; Balance_B2;Balance_C;Balance_D; HPLC];

% Adding the buglab to measurements
buglab_measurement = buglab(contains(buglab, {'Result Raw'}));
measurements = [measurements; buglab_measurement];

% Create Read Group 
readGrp = addgroup(da, ['ReadGroup' ,reactor]); 


%% SEQUENCE TWO OF THE HPLC DATA
NUMERA=itmIDs(contains(itmIDs,'Numera.Numera.SEQUENCE_2'));
NUMERA=NUMERA(2:end);

% For HPLC reactor identification of SampleManagmenet 
% important to know from which reactor the file with the barcode was taken
SampleBarcode=itmIDs(contains(itmIDs,'SampleBarcode'));
SampleName=itmIDs(contains(itmIDs,'SampleName'));

% this loop is to secure the name of the reactor is registered only from the
% first field of the list of possible reactors (because the letters D, E, A
% can also be par of the names in ItemID): 
for z=1:length(SampleBarcode)
    
    SBC=SampleName(contains(SampleName{z}(1:2),reactor));
    if not(isempty(SBC))
        SBCidx=z;
    end
end
SampleName=SampleName(SBCidx);
SampleBarcode=SampleBarcode(SBCidx);

% Adding Numera Sequence 2 information
measurements = [measurements; NUMERA;SampleBarcode;SampleName];



% Add Tags 
readID = [measurements; controlOutputs; sysDevice];
itm = additem(readGrp, readID);
writeGrp = addgroup(da, ['WriteGroup', reactor]);
writeID = setpoints;
itm = additem(writeGrp, writeID);

% Making a structure with the indexes of each variable of interest
measurementIndexes = struct;
measurementIndexes.ExitO2 = find(contains(measurements, 'ExitO2'));
measurementIndexes.GasMix = find(contains(measurements, 'GasMix'));
measurementIndexes.OD = find(contains(measurements, 'OD'));
measurementIndexes.Temperature = find(contains(measurements, 'Temperature'));
measurementIndexes.Feed = find(contains(measurements, 'Feed'));
if length(measurementIndexes.Feed) > 1
    measurementIndexes.Feed_Ext_1 = find(contains(measurements, 'Ext_1'));
    measurementIndexes.Feed_Ext_2 = find(contains(measurements, 'Ext_2'));
    measurementIndexes.Feed_Ext_3 = find(contains(measurements, 'Ext_3'));
    measurementIndexes.Feed_Ext_4 = find(contains(measurements, 'Ext_4'));
    measurementIndexes.Feed_Ext_5 = find(contains(measurements, 'Ext_5'));
    % Make a list of only the IDs of "feed" pumps
    feedMeasurements = measurements(measurementIndexes.Feed);
    % Find the only index that is not "Ext"
    measurementIndexes.Feed = measurementIndexes.Feed ...
        (~contains(feedMeasurements, 'Ext'));
end
measurementIndexes.O2Flow = find(contains(measurements, 'O2Flow'));
measurementIndexes.N2Flow = find(contains(measurements, 'N2Flow'));
measurementIndexes.AirFlow = find(contains(measurements, 'AirFlow'));
measurementIndexes.ExitCO2 = find(contains(measurements, 'ExitCO2'));
measurementIndexes.pH = find(contains(measurements, 'Values.pH'));
measurementIndexes.GMFlow = find(contains(measurements, 'GMFlow'));
measurementIndexes.pO2 = find(contains(measurements, 'pO2'));
measurementIndexes.Antifoam = find(contains(measurements, 'Antifoam'));
measurementIndexes.Stirrer = find(contains(measurements, 'Stirrer'));
measurementIndexes.Balance_A2 = find(contains(measurements, 'Balance A2'));
measurementIndexes.Balance_B2 = find(contains(measurements, 'Balance B2'));
measurementIndexes.Balance_C = find(contains(measurements, 'Balance C'));
measurementIndexes.Balance_D = find(contains(measurements, 'Balance D'));


% Add the HPLC information
measurementIndexes.HPLC_gluc= find(contains(measurements,'Measurements_HPLC.Glucose'));
measurementIndexes.HPLC_ara= find(contains(measurements,'Measurements_HPLC.Arabinose'));
measurementIndexes.HPLC_xyl= find(contains(measurements,'Measurements_HPLC.Xylose'));
measurementIndexes.HPLC_man= find(contains(measurements,'Measurements_HPLC.Mannose'));
measurementIndexes.HPLC_year= find(contains(measurements,'TimeStampYear_NU'));
measurementIndexes.HPLC_month= find(contains(measurements,'TimeStampMonth_NU'));
measurementIndexes.HPLC_day= find(contains(measurements,'TimeStampDay_NU'));
measurementIndexes.HPLC_hour= find(contains(measurements,'TimeStampHour_NU'));
measurementIndexes.HPLC_minute= find(contains(measurements,'TimeStampMinute_NU'));
measurementIndexes.HPLC_sec= find(contains(measurements,'TimeStampSec_NU'));

% Add the Buglab
measurementIndexes.Buglab = find(contains(measurements, 'Buglab'));

% Add the Seq 2 from Numera
measurementIndexes.NumeraIndexes.analyzer_number= find(contains(measurements,'ANALYZER_NUMBER'));
measurementIndexes.NumeraIndexes.seq2_start =find(contains(measurements,'SEQ2_START'));
measurementIndexes.NumeraIndexes.seq2_state =find(contains(measurements,'SEQ2_STATE'));
measurementIndexes.NumeraIndexes.sampling_done =find(contains(measurements,'SAMPLING_DONE'));
measurementIndexes.NumeraIndexes.analyzer_transf_type =find(contains(measurements,'ANALYZER_TRANSF_TYPE'));
measurementIndexes.NumeraIndexes.sample_ready= find(contains(measurements,'SAMPLE_READY'));
measurementIndexes.NumeraIndexes.seq2_stop= find(contains(measurements,'SEQ2_STOP'));

measurementIndexes.SampleBarcode=find(contains(measurements,SampleBarcode));
measurementIndexes.SampleName=find(contains(measurements,SampleName));

controlOutputIndexes = struct;
controlOutputIndexes.GMFlow_Output = find(contains(controlOutputs, 'GMFlow_Output')) + length(measurements);
controlOutputIndexes.Stirrer_Output = find(contains(controlOutputs, 'Stirrer_Output')) + length(measurements);
controlOutputIndexes.AirFlow_Output = find(contains(controlOutputs, 'AirFlow_Output')) + length(measurements);
controlOutputIndexes.OD_Output = find(contains(controlOutputs, 'OD_Output')) + length(measurements);
controlOutputIndexes.Temperature_Output = find(contains(controlOutputs, 'Temperature_Output')) + length(measurements);
controlOutputIndexes.GasMix_Output = find(contains(controlOutputs, 'GasMix_Output')) + length(measurements);
controlOutputIndexes.pH_Output = find(contains(controlOutputs, 'pH_Output')) + length(measurements);
controlOutputIndexes.Feed_Output = find(contains(controlOutputs, 'Feed_Output')) + length(measurements);
controlOutputIndexes.pO2_Output = find(contains(controlOutputs, 'pO2_Output')) + length(measurements);
controlOutputIndexes.ExitCO2_Output = find(contains(controlOutputs, 'ExitCO2_Output')) + length(measurements);
controlOutputIndexes.N2Flow_Output = find(contains(controlOutputs, 'N2Flow_Output')) + length(measurements);
controlOutputIndexes.Antifoam_Output = find(contains(controlOutputs, 'Antifoam_Output')) + length(measurements);
controlOutputIndexes.O2Flow_Output = find(contains(controlOutputs, 'O2Flow_Output')) + length(measurements);
controlOutputIndexes.ExitO2_Output = find(contains(controlOutputs, 'ExitO2_Output')) + length(measurements);
controlOutputIndexes.PumpValAcid = find(contains(controlOutputs, 'Pump_Val_Acid')) + length(measurements);
controlOutputIndexes.PumpValBase = find(contains(controlOutputs, 'Pump_Val_Base')) + length(measurements);

systemIndexes = struct;
systemIndexes.StepTime = find(contains(sysDevice, 'StepTime')) + ...
    length(measurements) + length(controlOutputs);
systemIndexes.ProcTime = find(contains(sysDevice, 'ProcTime')) + ...
    length(measurements) + length(controlOutputs);

setpointIndexes = struct;
setpointIndexes.Antifoam_Setpoint = find(contains(setpoints, 'Antifoam_Setpoint'));
setpointIndexes.Temperature_Setpoint = find(contains(setpoints, 'Temperature_Setpoint'));
setpointIndexes.pO2_Setpoint = find(contains(setpoints, 'pO2_Setpoint'));
setpointIndexes.Stirrer_Setpoint = find(contains(setpoints, 'Stirrer_Setpoint'));
setpointIndexes.GMFlow_Setpoint = find(contains(setpoints, 'GMFlow_Setpoint'));
setpointIndexes.O2Flow_Setpoint = find(contains(setpoints, 'O2Flow_Setpoint'));
setpointIndexes.pH_Setpoint = find(contains(setpoints, 'pH_Setpoint'));
setpointIndexes.AirFlow_Setpoint = find(contains(setpoints, 'AirFlow_Setpoint'));
setpointIndexes.N2Flow_Setpoint = find(contains(setpoints, 'N2Flow_Setpoint'));
setpointIndexes.GasMix_Setpoint = find(contains(setpoints, 'GasMix_Setpoint'));
setpointIndexes.Feed_Setpoint = find(contains(setpoints, 'Feed'));

if length(setpointIndexes.Feed_Setpoint) > 1
    setpointIndexes.Feed_Ext_1_Setpoint = find(contains(setpoints, 'Ext_1'));
    setpointIndexes.Feed_Ext_2_Setpoint = find(contains(setpoints, 'Ext_2'));
    setpointIndexes.Feed_Ext_3_Setpoint = find(contains(setpoints, 'Ext_3'));
    setpointIndexes.Feed_Ext_4_Setpoint = find(contains(setpoints, 'Ext_4'));
    setpointIndexes.Feed_Ext_5_Setpoint = find(contains(setpoints, 'Ext_5'));

    % Make a list of only the IDs of "feed" pumps
    feedSetpoints = setpoints(setpointIndexes.Feed_Setpoint);
    
    % Find the only index that is not "Ext"
    setpointIndexes.Feed_Setpoint = setpointIndexes.Feed_Setpoint ...
        (~contains(feedSetpoints, 'Ext'));
end



indexes = struct;
indexes.system = systemIndexes;
indexes.setpoints = setpointIndexes;
indexes.measurements = measurementIndexes;
indexes.controlOutputs = controlOutputIndexes;
end

% Function to extract the information of the measurements
function [Instrument]=GetInstData(data,indexesA2,clk_now,par,iteration)
% clk is the clock information taken in the beginning of the expetiments
% here you can add what you want to Instrument struct (as fields)
% Define data for the balances, 

if ~isnan(indexesA2.measurements.Balance_A2)==1
    Instrument.Balance_A2= double(data(indexesA2.measurements.Balance_A2).Value);
    Instrument.Balance_B2= double(data(indexesA2.measurements.Balance_B2).Value);

end

if ~isnan(indexesA2.measurements.Balance_C)==1
    Instrument.Balance_C= double(data(indexesA2.measurements.Balance_C).Value);
    Instrument.Balance_D= double(data(indexesA2.measurements.Balance_D).Value);
end

%%%%%%%%%%%%%%%%%%%
%%%% NIR Acquisition: Smoothing, then linear conversion to Cell dry weight CDw
%%%%%%%%
%%%%
% Measurements of NIR, Buglab;
%Instrument.Buglab= double(data(indexesA2.measurements.Buglab).Value);
%acquisition
Instrument.NIR= double(data(indexesA2.measurements.OD).Value);
% smoothing and conversion is conducted in Analysis_instrumentation()

% Measurements of Air flow CO2 and O2 
% XXX Remove the 60 after discussion with Nadi
Instrument.Air_flow= double(data(indexesA2.measurements.AirFlow).Value);
Instrument.CO2 = double(data(indexesA2.measurements.ExitCO2).Value);
Instrument.O2= double(data(indexesA2.measurements.ExitO2).Value); % 

% Read the data of the pumps
Instrument.Pump_1 = double(data(indexesA2.measurements.Feed_Ext_1).Value);
Instrument.Pump_2 = double(data(indexesA2.measurements.Feed_Ext_2).Value); 
 
% NOT NEEDED!!
% Aqcuire the information from Numera
% Instrument.Numera.Analyzer_number = double(data(indexesA2.measurements.NumeraIndexes.analyzer_number).Value);
% Instrument.Numera.seq2_start= double(data(indexesA2.measurements.NumeraIndexes.seq2_start).Value);
% Instrument.Numera.seq2_state= double(data(indexesA2.measurements.NumeraIndexes.seq2_state).Value);
% Instrument.Numera.sampling_done= double(data(indexesA2.measurements.NumeraIndexes.sampling_done).Value);
% Instrument.Numera.analyzer_transfer_type= double(data(indexesA2.measurements.NumeraIndexes.analyzer_transf_type).Value);
% Instrument.Numera.sample_ready= double(data(indexesA2.measurements.NumeraIndexes.sample_ready).Value);
% Instrument.Numera.seq2_stop= double(data(indexesA2.measurements.NumeraIndexes.seq2_stop).Value);

% Instrument time
if iteration>1
        Instrument.t=iteration*par.iterationTime; %calculate the current time from i and par.iterationtime
        %Instrument.time= Instrument.time(iteration-1) + par.iterationTime;
        Instrument.ProcTime= fix(clock)*[0 par.mth*60*60*24 60*60*24 60*60 60 1]'-par.clk_0*[0 par.mth*60*60*24 60*60*24 60*60 60 1]';        
    else % i==1
        Instrument.t=0; % first iteration time is close to zero
        %Instrument.time= 0;
        Instrument.ProcTime= 0;
end

% Acid and Base outputs
Instrument.Acid_pump= double(data(indexesA2.controlOutputs.PumpValAcid).Value);
Instrument.Base_pump= double(data(indexesA2.controlOutputs.PumpValBase).Value);

% pO2 measurement
Instrument.pO2= double(data(indexesA2.measurements.pO2).Value);

%% Sample HANDLING:
% The samples taken by NUMERA: 
% 1. read the now sample
Instrument.NowSampleBarCode=(data(indexesA2.measurements.SampleBarcode).Value);
Instrument.NowSampleName=(data(indexesA2.measurements.SampleName).Value);
% get the sample time
NUmth=double(data(indexesA2.measurements.HPLC_month).Value);
NUd=double(data(indexesA2.measurements.HPLC_day).Value);
NUh=double(data(indexesA2.measurements.HPLC_hour).Value);
% IMPORTANT!! because NUMERA reports the hour in UTC, mhile Norway clock is
% CET+1 -- we need to add another hour.
% SUMMER TIME: +2 WINTER: +1
t1=datetime('today','TimeZone','Europe/Oslo');
utc2GMTOslo=tzoffset(t1);

NUh=NUh+hours(utc2GMTOslo);
NUm=double(data(indexesA2.measurements.HPLC_minute).Value);
NUs=double(data(indexesA2.measurements.HPLC_sec).Value);

Instrument.NowSampleProcTime=[0 NUmth NUd NUh NUm NUs]*[0 par.mth*60*60*24 60*60*24 60*60 60 1]'-par.clk_0*[0 par.mth*60*60*24 60*60*24 60*60 60 1]';
if Instrument.NowSampleProcTime<0 % in case you started matlab and it found older sample from lucullus.
    Instrument.NowSampleProcTime=0;
end

if isempty(Instrument.NowSampleBarCode)
   % Instrument.NowSampleBioreactor= '';
else
    Instrument.NowSampleBioreactor=Instrument.NowSampleName(1:2);
end


end

% Obtain further analysis of the instrumentation, except HPLC
function[Instrument,par] = Analysis_instrumentation(Instrument,Instrument0,par,iteration,VarMem)
       % We need to divide by 1000 to have it in kg, 
       % Minus sign because the balance reduces its weight when consumed
       if iteration==1
           % first iteration, no memory of the flow
           Instrument.Volume_in= 0;
           Instrument.Volume_out= 0;
           Instrument.Flow_in = 0;
           Instrument.Flow_out = 0;
           Instrument.CO2_measurement = 0;  
           Instrument.V= par.V0;
           
           % CO2 analysis
           % Initial condition of air in the system (no overvalue of CO2, in case of wrong calibration)           
           if Instrument.CO2>0.05
               CO2inAir_volpercent= 0.04;
           else
               CO2inAir_volpercent = Instrument.CO2;
           end                       
                      
           % CO2 density kg/m3 and in g/L
           CO2_density = 1.77; % g/L
           
           % Volumetric flow of CO2
           % The Percentage value the volume of CO2
           Volume_CO2 = (((Instrument.CO2-CO2inAir_volpercent)));           
           if Volume_CO2<0
               Volume_CO2=0;
           end
           
           % CO2 excreted in units to check
           Instrument.CO2_excreted = (Volume_CO2*CO2_density); % % CO2/hr
           
           % The follow is the CO2 in % for the estimator
           % The total volume of the reactor is V_total:
           V_total= 2.7;
           
           % XXX Here multiply by 60, and in the air flow ignore the 60
           %Instrument.CO2_measurement_EKF = 60 * Instrument.CO2_excreted/((V_total-Instrument.V).*100);
           Instrument.CO2_measurement = 60 * Instrument.CO2_excreted/((V_total-Instrument.V).*100);
          
           % At the beginning no CO2 produced
           Instrument.CER=0;
           Instrument.CO2_stoichiometric = 0;           
           
           % NIR data as such
           NIR= Instrument.NIR;
           % No smoothing of first point
           Instrument.NIR_smooth= NIR;
           CDWcu = -15212.545*sin(2.0532*NIR  +2.5893) +15197.4245*sin(2.055*NIR +2.5887); 
           % No extra big NIR values when disconnected
           if NIR >= 5
               CDWcu= 0.66116;
           end
           % No negative NIR values
            if NIR <= 0.0136281815728
               CDWcu= 0;
           end
           % Store the CDW information
           Instrument.CDWcu= CDWcu;            
           
           % At the beginning no Substrate added
           Instrument.S_added= 0;          
           
       else % iterations > 1
           
           % Compute the balances values
           if isfield(Instrument,'Balance_A2')==1
               % Zero the balances
               % this code is written so that the initial balance A2 is 0,
               % and is negatively decreasing
               % FILTER algorithm works like this:
               % 1. Balance A2 must be monotonically decreasing. 
               % 2. diff values less than 0.02 are ignored (keep the old
               % value)
               diffBalanceA2=Instrument.Balance_A2-VarMem.Instrument.Balance_A2(iteration-1);
               % This difference should either decrease or stay the same,
               % never increase.
               if diffBalanceA2>0
                   % values increased, keep the previous value
                   Instrument.Balance_A2=VarMem.Instrument.Balance_A2(iteration-1);
               elseif abs(diffBalanceA2)<0.05
                % the change is too small, it is noise. keep the previous
                % value
                  Instrument.Balance_A2=VarMem.Instrument.Balance_A2(iteration-1);
           
               end
               
               % Balance B2 (out balance)
                diffBalanceB2=Instrument.Balance_B2-VarMem.Instrument.Balance_B2(iteration-1);
               % This difference should either increase or stay the same,
               % never decrease.
               if diffBalanceB2<0
                   % values decreased, keep the previous value
                   Instrument.Balance_B2=VarMem.Instrument.Balance_B2(iteration-1);
               elseif abs(diffBalanceB2)<0.05
                % the change is too small, it is noise. keep the previous
                % value
                  Instrument.Balance_B2=VarMem.Instrument.Balance_B2(iteration-1);
               end
               
               if abs(diffBalanceA2)>100
                   % We have a reset of the balances. This needs a
                   % correction to the initial Intrument0
                    par.baselineWeight.A2= par.baselineWeight.A2 + diffBalanceA2;      
               end
               if abs(diffBalanceB2)>100
                   % We have a reset of the balances. This needs a
                   % correction to the initial Intrument0
                    par.baselineWeight.B2= par.baselineWeight.B2 + diffBalanceB2;
               end
               
               Instrument.Volume_in = -(Instrument.Balance_A2 - par.baselineWeight.A2)/par.flowDensity;        
               Instrument.Volume_out = (Instrument.Balance_B2 - par.baselineWeight.B2)/par.flowDensity;
               
           end
           
           
           if isfield(Instrument,'Balance_C')==1
                % Instrument.Balance_C=Instrument.Balance_C+par.baselineWeight;
                diffBalanceC=Instrument.Balance_C-VarMem.Instrument.Balance_C(iteration-1);
               % This difference should either decrease or stay the same,
               % never increase.
               if diffBalanceC>0 && abs(diffBalanceC)<100
                       Instrument.Balance_C=VarMem.Instrument.Balance_C(iteration-1);
                  % end
                  
               elseif abs(diffBalanceC)<0.1
                % the change is too small, it is noise. keep the previous
                % value
                  Instrument.Balance_C=VarMem.Instrument.Balance_C(iteration-1);
               end

              
               % Balance D (out balance)
                diffBalanceD=Instrument.Balance_D-VarMem.Instrument.Balance_D(iteration-1);
               % This difference should either decrease or stay the same,
               % never increase.
               if diffBalanceD<0
                   % values increased, keep the previous value
                   Instrument.Balance_D=VarMem.Instrument.Balance_D(iteration-1);
               elseif abs(diffBalanceD)<0.1
                % the change is too small, it is noise. keep the previous
                % value
                  Instrument.Balance_D=VarMem.Instrument.Balance_D(iteration-1);
               end
     
               if abs(diffBalanceC)>100
                   % We have a reset of the balances. This needs a
                   % correction to the initial Intrument0
                    par.baselineWeight.C= par.baselineWeight.C + diffBalanceC;      
               end
               
               if abs(diffBalanceD)>100
                   % We have a reset of the balances. This needs a
                   % correction to the initial Intrument0
                    par.baselineWeight.D= par.baselineWeight.D + diffBalanceD;
               end
               Instrument.Volume_in = -(Instrument.Balance_C - par.baselineWeight.C)/par.flowDensity;        
               Instrument.Volume_out = (Instrument.Balance_D -  par.baselineWeight.D)/par.flowDensity;
           end
           
             %regardless of the balances values, the volumes cannot be negative:
             if Instrument.Volume_in<0
                 Instrument.Volume_in=0;
             end
             
              if Instrument.Volume_out<0
                 Instrument.Volume_out=0;
             end
               
           % Volume estimation from the measurements
           Instrument.V= par.V0 + Instrument.Volume_in - Instrument.Volume_out;
           
           % Delta of time in hrs
           Delta= (Instrument.ProcTime-VarMem.ProcTime(iteration-1))/3600;                 
           
           % Compute the flows  
           % Current value minus the previous value divided by the delta in time.
           Instrument.Flow_in = (Instrument.Volume_in - VarMem.Instrument.Volume_in(iteration-1))/ Delta;
           Instrument.Flow_out = (Instrument.Volume_out - VarMem.Instrument.Volume_out(iteration-1))/ Delta;    
           
           % Eliminate data that is only noise
           if iteration>1             
               
               % from analysis of the balances (diff(balance_A2) we found
               % that the flow during 30 seconds is not less than 0.012. 
               % Eliminate noisy data from Flow in
               if Instrument.Flow_in < 0.012 || isnan(Instrument.Flow_in) % 5 mL/h
                   Instrument.Flow_in= 0;
               end               
               if Instrument.Flow_in>2.5 % saturation, because the pumps has max capacity, so this removes noise in the balances.
                   Instrument.Flow_in=2.5;
               end
               
               % Eliminate noisy data from Flow out
               if Instrument.Flow_out < 0.012 || isnan(Instrument.Flow_out)
                   Instrument.Flow_out= 0;
               end
               if Instrument.Flow_out>2.5 % saturation, because the pumps has max capacity, so this removes noise in the balances.
                   Instrument.Flow_out=2.5;
               end
           end
           
           % CO2 computation
           % Take First point
          % CO2inAir_volpercent= VarMem.Instrument.CO2(1);
       
           % CO2 density kg/m3 and in g/L
           CO2_density = 1.77; % g/L
           
           % Volumetric flow of CO2
           % The Percentage value the volume of CO2
           Volume_CO2 = (((Instrument.CO2-par.CO2inAir_volpercent)));           
           if Volume_CO2<0
               Volume_CO2=0;
           end
                                           
           % CO2 excreted in units to check
           Instrument.CO2_excreted = (Volume_CO2*CO2_density); % % CO2/hr
           
           % The follow is the CO2 in % for the estimator
           % The total volume of the reactor is V_total:
           V_total= 2.7;
           Instrument.CO2_measurement = 60 * Instrument.CO2_excreted/((V_total-Instrument.V).*100);
           if Instrument.CO2_measurement < 0
              % how is it possible?
               Instrument.CO2_measurement = 0 ;
           end
           
           % Stoichiometric CO2
           Instrument.CER = Instrument.Air_flow*(Volume_CO2*CO2_density); % g CO2/hr
           CarbDiox= VarMem.Instrument.CO2_stoichiometric(iteration-1);
           
           % Delta of time in hrs
           Delta= (Instrument.ProcTime-VarMem.ProcTime(iteration-1))/3600;                                
           
           % Make the integral gCO2/L 
           Instrument.CO2_stoichiometric = CarbDiox + Instrument.CER*Delta;    
       
           %%%% NIR Handling: Smoothing, then linear conversion to Cell dry weight CDw
           %%%%%%%%
           %%%%
          
           % smooth the whole vector in the NIR
           NIR_analysis = [VarMem.Instrument.NIR Instrument.NIR];
           
           % Gaussian filter
           Instrument.NIR_smooth= smoothdata(NIR_analysis,'lowess',80);
           Instrument.NIR_smooth=Instrument.NIR_smooth(end);
           
           NIR= Instrument.NIR_smooth;
           %if  NIR>= 0.52 % CDWcu = 22.187 *NIR - 5.0991 ; else CDWcu = 11.124 *NIR + 0.66116 ; end
           % New calibration
           CDWcu = -15212.545*sin(2.0532*NIR  +2.5893) +15197.4245*sin(2.055*NIR +2.5887); 
           % No extra big NIR values when disconnected
           if NIR >= 5
               CDWcu= 0.66116;
           end
           % No negative NIR or CDW values
           if NIR <= 0.0136281815728
               CDWcu= 0;
           end
           % Store the CDW information
           Instrument.CDWcu= CDWcu;
           
           % Substrate added in the bioprocess:
           Instrument.S_added = (Instrument.Volume_in)*par.Sin;

           
       end
end

% Function to obtain HPLC data from File
function [Instrument,VarMem]=GetHPLCFromFile(Instrument,clk,par,VarMem,iteration)

  %% Sample data to store in the Numera system
    if iteration ==1
        % nan is because intersect needs to compare to something
        VarMem.Instrument.SampleDataBase(1,1)={'0'};
        VarMem.Instrument.SampleDataBase(1,2)={nan};
        VarMem.Instrument.SampleDataBase(1,3)={nan};
        VarMem.Instrument.SampleDataBase(1,4)={nan};
    
    % 1. Is there a new sample??
    elseif isempty(Instrument.NowSampleBarCode)
        % no --> do nothing
        
    else
        %yes --> compare it to the database
        
        %2. Compare to SampleDataBase
        % the structure of the database is cell matrix m x 4, where m is the number
        % of samples taken from the beginning of the process. the columns are
        % Barcode ,  timestamp,  reactor,  name
         L=length(VarMem.Instrument.SampleDataBase(:,1));
        [C] = intersect(VarMem.Instrument.SampleDataBase(:,1),Instrument.NowSampleBarCode,'stable');
        if isempty(C)
            % new sample does not exist in the database. add the sample
            VarMem.Instrument.SampleDataBase(L+1,1)={Instrument.NowSampleBarCode};
            VarMem.Instrument.SampleDataBase(L+1,2)={Instrument.NowSampleProcTime};
            VarMem.Instrument.SampleDataBase(L+1,3)={Instrument.NowSampleBioreactor};
            VarMem.Instrument.SampleDataBase(L+1,4)={Instrument.NowSampleName};
        end
     
        
    end
%% Aqcuire HPLC measurements
%Read from the Chromeleon Folder, and search for a new file using the
%iteration clock clk

  [file,FileSampleBarCode]=FindNewChromeleonFiles_HPLC(clk,par);
  
  % need to remove the .txt from the barcode:
  FileSampleBarCode=FileSampleBarCode(1:end-4);
  
  % the time file was genrated is the end of HPLC measurement. Understood
  Instrument.HPLC.gluc=[];
  Instrument.HPLC.xyl=[];
  Instrument.HPLC.man=[];
  Instrument.HPLC.ara=[];
  
  if not(isempty(file))
     
     %if strcmp(Instrument.SampleName(1:2),par.Info.BioReactor)
     
     % Now compare the Sample Barcode the the database of all the samples
     % taken, and extract from the SampleDataBase he reactor and the
     % timestamp:
     [C,IA] = intersect(VarMem.Instrument.SampleDataBase(:,1),FileSampleBarCode,'stable');
     if not(isempty(C))
         
          % is this the correct reactor?
          IsBioReactor=VarMem.Instrument.SampleDataBase{IA,3};
          if strcmp(IsBioReactor,par.Info.BioReactor)
              %It is the correct!
          %extract the data from the file
          % the format is Instrument.HPLC.Gluc, Instrument.HPLC.Man,
          % Instrument.HPLC.Ara etc
          Instrument.HPLC = ReadSugarValuesFromChromeleonFile(file);
          Instrument.HPLC.TimeStamp= VarMem.Instrument.SampleDataBase{IA,2};
          end
     else
         disp('ERROR! new chromeleon file was created and found, but it is missing in the SampleDataBase!')
         %disp('debugging...')
         %keyboard
     end
      
  end
end

% Function to analyze HPLC data
function[measurements,VarMem,Instrument]= HPLC_data(Instrument,VarMem,iteration,ProcTime,par)

i= iteration;
% If you do not want to use the possible previous HPLC information
% First iteration is zero
if i==1
    Instrument.HPLC.gluc=[];
    Instrument.HPLC.xyl= [];
    Instrument.HPLC.man= [];
    Instrument.HPLC.ara= [];
end

% Define the vector state measurements
% Case 1) No measurement
% Case 2) With HPLC measurement
% If there is no HPLC measurement then the structure is empty
% Then we have 3 measurements    
if isempty(Instrument.HPLC.gluc)
    % Case 1 
    % Initialize the algorithm
    tidsec= [];
    measurements = [Instrument.V, Instrument.CDWcu, nan, Instrument.CO2_measurement];
    
    % Save the vectors as NaN
    VarMem.Instrument.HPLC.ara(i)=nan;
    VarMem.Instrument.HPLC.gluc(i)=nan;
    VarMem.Instrument.HPLC.man(i)=nan;
    VarMem.Instrument.HPLC.xyl(i)=nan;
%     VarMem.Instrument.HPLC.year(i)=nan;
%     VarMem.Instrument.HPLC.month(i)=nan;
%     VarMem.Instrument.HPLC.day(i)=nan;
%     VarMem.Instrument.HPLC.hour(i)=nan;
%     VarMem.Instrument.HPLC.minute(i)=nan;
%     VarMem.Instrument.HPLC.sec(i)=nan;
    VarMem.Instrument.HPLC.timeHPLC(i)=nan;
    VarMem.Instrument.HPLC.idxTimeStamp(i)=nan;
    VarMem.Instrument.HPLC.TimeStamp(i)= nan;
    VarMem.Instrument.HPLC.SampleIteration=nan;
else
    % Case 2
    % Define the HPLC measurement time in seconds        
    % The time stamp of the measurement from the Chromeleon file stamp, in
    % seconds.
    tidsec= Instrument.HPLC.TimeStamp; % Time of the Numera sample of HPLC, since the beginning of ProcTime. In seconds 
    % Confirm that the time of the sampling of the HPLC is lower than the ProcTime
    if Instrument.ProcTime < tidsec || tidsec<0
        % keyboard
        display('Previous HPLC measurement is found in Lucullus, or ProcTime is lower than HPLC measurement (not correct)')
        %keyboard
    end
    
        % Find the set of indexes that have a higher time than tidsec
        idxTimeBackTemp= find(VarMem.ProcTime>tidsec);
        % Here are two indeces: idxTimeBack is how many index to go back in
        % VarMem, this is for the backwards calculation of the EKF, and 
        % idxTimeStamp is where in VarMem the TimeStamp of the HPLC belongs
        Instrument.HPLC.idxTimeStamp= idxTimeBackTemp(1);
        Instrument.HPLC.idxTimeBack= length(VarMem.ProcTime) - idxTimeBackTemp(1);
        
        % Construct the vector of measurement with the HPLC data
        Instrument.HPLC.first_measurement= [VarMem.measurements(Instrument.HPLC.idxTimeStamp,1:2) Instrument.HPLC.gluc VarMem.measurements(Instrument.HPLC.idxTimeStamp,4)];            
        
        % Replace the specific vector in the Memory of measurements
        VarMem.measurements(Instrument.HPLC.idxTimeStamp,:)= Instrument.HPLC.first_measurement;
        VarMem.Instrument.HPLC.timeHPLC(Instrument.HPLC.idxTimeStamp) = tidsec;
        
        % saving the HPLC values that belong to the past, 
        % Here is the index where the information corresponds
        VarMem.Instrument.HPLC.ara(Instrument.HPLC.idxTimeStamp)=Instrument.HPLC.ara;
        VarMem.Instrument.HPLC.gluc(Instrument.HPLC.idxTimeStamp)=Instrument.HPLC.gluc;
        VarMem.Instrument.HPLC.man(Instrument.HPLC.idxTimeStamp)=Instrument.HPLC.man;
        VarMem.Instrument.HPLC.xyl(Instrument.HPLC.idxTimeStamp)=Instrument.HPLC.xyl;    
        VarMem.Instrument.HPLC.idxTimeStamp(Instrument.HPLC.idxTimeStamp)=Instrument.HPLC.idxTimeStamp;
        VarMem.Instrument.HPLC.TimeStamp(Instrument.HPLC.idxTimeStamp)= Instrument.HPLC.TimeStamp;
        VarMem.Instrument.HPLC.SampleIteration=iteration;
        % Please note that the present i-iteration has no HPLC data
        measurements= [Instrument.V, Instrument.CDWcu, nan, Instrument.CO2_measurement];
    %end
end    

end

% Function for saving the information

function VarMem=SaveVar2Memory(i,par,measurements,Instrument0,Instrument,Controller, ModSim, MHESim, VarMem)
    % VarMem is the variables in memory, all in struct with fields related
    
    % TIme
    % Fixed the time with the incorporation of VarMeM above
    if i>1
        VarMem.Instrument.t(i)=i*par.iterationTime; %calculate the current time from i and par.iterationtime
        VarMem.time(i)= VarMem.time(i-1) + par.iterationTime;
        
        Proctime= clock*[0 par.mth*60*60*24 60*60*24 60*60 60 1]'-par.clk_0*[0 par.mth*60*60*24 60*60*24 60*60 60 1]';
        VarMem.ProcTime(i)= Proctime;
    else % i==1
        VarMem(i).Instrument.t=0; % first iteration time is close to zero
        VarMem(1).time= 0;
        VarMem.ProcTime(i)= 0;
    end
    % HPLC:
    % it was already saved in the function HPLC_data, but it is repeated
    % here for the sake of the last iteration and plotting (removing this
    % will fail to plot)
    VarMem.Instrument.HPLC.ara(i)=nan;
    VarMem.Instrument.HPLC.gluc(i)=nan;
    VarMem.Instrument.HPLC.man(i)=nan;
    VarMem.Instrument.HPLC.xyl(i)=nan;
%     VarMem.Instrument.HPLC.year(i)=nan;
%     VarMem.Instrument.HPLC.month(i)=nan;
%     VarMem.Instrument.HPLC.day(i)=nan;
%     VarMem.Instrument.HPLC.hour(i)=nan;
%     VarMem.Instrument.HPLC.minute(i)=nan;
%     VarMem.Instrument.HPLC.sec(i)=nan;
    VarMem.Instrument.HPLC.timeHPLC(i)=nan;
    VarMem.Instrument.HPLC.idxTimeStamp(i)=nan;
    VarMem.Instrument.HPLC.TimeStamp(i)= nan;  
    
    % Infors towers
    VarMem.Instrument.Flow_in(i) = Instrument.Flow_in;
    VarMem.Instrument.Flow_out(i) = Instrument.Flow_out;    
    VarMem.Instrument.Pump_1(i) =Instrument.Pump_1;
    VarMem.Instrument.Pump_2(i) = Instrument.Pump_2;
    VarMem.Instrument.Air_flow(i) =Instrument.Air_flow;
    VarMem.Instrument.CO2(i)= Instrument.CO2;
    VarMem.Instrument.CO2_measurement(i)= Instrument.CO2_measurement;
    VarMem.Instrument.O2(i)=Instrument.O2;
   
     VarMem.Instrument.NIR(i) = Instrument.NIR;
    VarMem.Instrument.NIR_smooth(i) = Instrument.NIR_smooth;
    VarMem.Instrument.CDWcu(i) = Instrument.CDWcu;
    
    VarMem.Instrument.Volume_in(i) =Instrument.Volume_in;
    VarMem.Instrument.Volume_out(i) =Instrument.Volume_out;
    VarMem.Instrument.V(i)= Instrument0.V + Instrument.Volume_in - Instrument.Volume_out;
    VarMem.Instrument.CO2_stoichiometric(i)= Instrument.CO2_stoichiometric;    
    VarMem.Instrument.CER(i)= Instrument.CER;    
    
    % Save the available balances
    if isfield(Instrument,'Balance_A2')==1
        VarMem.Instrument.Balance_A2(i)= Instrument.Balance_A2;
        VarMem.Instrument.Balance_B2(i)= Instrument.Balance_B2;
    end
    
    if isfield(Instrument,'Balance_C')==1
        VarMem.Instrument.Balance_C(i)= Instrument.Balance_C;
        VarMem.Instrument.Balance_D(i)= Instrument.Balance_D;
    end
    
    % VarMem.Instrument.CarbDiox(i)=   Instrument.CarbDiox;    
    % VarMem.Instrument.Buglab(i)= Instrument.Buglab;        
    % The vector of state measurements
    VarMem.measurements(i,:)=measurements;    
        
    % Acid and Base outputs
    VarMem.Instrument.Acid_pump(i)= Instrument.Acid_pump;
    VarMem.Instrument.Base_pump(i)= Instrument.Base_pump;
    
    % pO2 measurement
    VarMem.Instrument.pO2(i)= Instrument.pO2;
    

    % Information from the Numera based on sequence 2
%     VarMem.Instrument.Numera.Analyzer_number(i) =Instrument.Numera.Analyzer_number;
%     VarMem.Instrument.Numera.seq2_start(i)= Instrument.Numera.seq2_start; 
%     VarMem.Instrument.Numera.seq2_state(i)= Instrument.Numera.seq2_state;
%     VarMem.Instrument.Numera.sampling_done(i)= Instrument.Numera.sampling_done;
%     VarMem.Instrument.Numera.analyzer_transfer_type(i)= Instrument.Numera.analyzer_transfer_type;
%     VarMem.Instrument.Numera.sample_ready(i)= Instrument.Numera.sample_ready;
%     VarMem.Instrument.Numera.seq2_stop(i)= Instrument.Numera.seq2_stop;

    % Biomass corelation
    NIR= VarMem.Instrument.NIR(end);
    CDWcu = -15212.545*sin(2.0532*NIR  +2.5893) +15197.4245*sin(2.055*NIR +2.5887); 
     
     
     % No extra big NIR values
     if NIR >= 5
         CDWcu= 0.66116;
     end
     
     % No negative NIR values
     if Instrument0.NIR <= 0.0136281815728
         CDWcu= 0;
     end   
     
   VarMem.Instrument.CDWcu(i)= CDWcu;
   VarMem.Instrument.S_added(i)= Instrument.S_added;
    
   % Log controller values  
   VarMem.Controller.Fin(i)= Controller.Fin;
   VarMem.Controller.Fout(i) =  Controller.Fin;
   VarMem.Controller.Pumpin(i) = Controller.Fin2pump;
   VarMem.Controller.Pumpout(i) = Controller.Fin2pump;
   
   % Model simulation
   VarMem.ModSim.V(i) = ModSim.V;
   VarMem.ModSim.X(i) = ModSim.X;
   VarMem.ModSim.S(i) = ModSim.S;
   VarMem.ModSim.CO2(i) = ModSim.CO2;
   
   % MHE estimates
   VarMem.MHESim.V(i) = MHESim.V;
   VarMem.MHESim.X(i) = MHESim.X;
   VarMem.MHESim.S(i) = MHESim.S;
   VarMem.MHESim.CO2(i) = MHESim.CO2;
     
end

% Function for plotting all the results
function plotting_VarMem(VarMem,par,iteration)

    set(0,'defaultfigurecolor',[1 1 1]) %Make the figure white
    time= VarMem.ProcTime/3600;
    
    % Substrate information
   % S_timeseries= VarMem.measurements(:,3);
    time_HPLC= VarMem.Instrument.HPLC.timeHPLC/3600;
    
    
    figure(1)
    % magenta: ModSim model simulation
    % red: EKF
    % blue: measurements
    % Black: Set points
  
    subplot(2,2,1)
    %hold on
    plot(time,VarMem.Instrument.V,'-b', time,VarMem.ModSim.V,'--r', time, VarMem.MHESim.V, '.-g')
    %plot(time(1:length(VarMem.ModSim.V)),VarMem.ModSim.V,'--r')
    %plot(time, VarMem.MHESim.V, '.-g')
    ylabel('Volume [L]')
    %ylim([0 2.5])
    %xlim([0 40])  
    xticklabels({})      
    box 'off'
    legend ('Measurement', 'Model', 'MHE')
    legend box off
   

 % X
  subplot(2,2,2)
  %hold on
  plot(time,VarMem.Instrument.CDWcu,'-b', time,VarMem.ModSim.X,'--r', time, VarMem.MHESim.X, '.-g')   
  %plot(time(1:length(VarMem.ModSim.X)),VarMem.ModSim.X,'--r') 
  %plot(time, VarMem.MHESim.X, '.-g')
    ylabel('Biomass [g L^{-1}]')
    % xlim([0 40])
    %ylim([0 100])
   % xticklabels({})
    xlabel('Time [h]')
    %box 'off'
    %legend ('Measurement', 'Model', 'MHE')
    %legend box off
    
    s_idx = find(VarMem.measurements(:,3)>=0);
    subplot(2,2,3)
    %hold on
    plot(time_HPLC,VarMem.measurements(:,3),'ob', time,VarMem.ModSim.S,'--r', time, VarMem.MHESim.S, '.-g')
    %plot(time_HPLC(s_idx),VarMem.measurements(s_idx,3),'-b')
    %plot(time(1:length(VarMem.ModSim.S)),VarMem.ModSim.S,'--r') 
    %plot(time, VarMem.MHESim.S, '.-g')
    ylabel('Substrate [g L^{-1}]')
    %ylim([0 50])
    %xlim([-40 0])
    xticklabels({})
    box 'off' 
    legend ('Measurement HPLC', 'Model', 'MHE')
    legend box off
    
%   CO2
    subplot(2,2,4)
    plot(time,VarMem.Instrument.CO2_measurement,'-b', time, VarMem.ModSim.CO2,'--r', time, VarMem.MHESim.CO2, '.-g') 
    %plot(time(1:length(VarMem.ModSim.CO2)),VarMem.ModSim.CO2,'--r') 
    %plot(time, VarMem.MHESim.CO2, '.-g')
    %hold on
    ylabel('CO_2 [% L^{-1}]')
    %ylim([0 0.050])
    % xlim([0 40])
    %xticklabels({})
    xlabel('Time [h]')
    %box 'off'
    %legend ('Measurement', 'Model', 'MHE')
    %legend box off
       
    
    % INPUT FIGURE
    figure(2)
    subplot(1,2,1)
    %hold on
    plot(time,VarMem.Instrument.Flow_in,'-b', time,VarMem.Instrument.Flow_out,'--r')
    %plot(time,VarMem.Instrument.Flow_out,'--r')
    ylabel('Sugar Flow[L hr^{-1}]')
    box 'off'
    legend("Flow in", "Flow out")
    
    
    subplot(1,2,2)
    plot(time,VarMem.Instrument.S_added,'-b')
    ylabel('Sugar added [g]')
    %ylim([18 40])
    %xlim([0 40])
    xticklabels({})
    box 'off'
    legend("Total Sugar added")


end

function SugVal=ReadSugarValuesFromChromeleonFile(file)
% this file reads the amount values of all sugar components from the file
% created by Chromeleon 6, NOT FROM LUCULLUS.
% if it does not finds a sugar its sends back nan
% can be modified to products and other molecules

%fid = fopen('file03_00021821.txt');
fid = fopen(file);

CTitle = textscan(fid,'%19c',1);
%CTitle = textscan(fid,'%s',1,'Delimiter','\n');

if not(strcmp(CTitle,'Integration Results'))
    disp('ERROR: Wrong Chromeleon results file!')
    SugVal.gluc=nan;
    SugVal.man=nan;
    SugVal.xyl=nan;
    SugVal.ara=nan;
    keyboard
    return
    
end

% read the next two lines (but not used).
CsubTitle=textscan(fid,'%s',4,'Delimiter','\n');

%C=textscan(fid,'%s',1,'Delimiter','\n');
%C=textscan(fid,'%s %s %f %f %f %f %f %f');

%*[^\n] skips the reminder of the line after the Amount:
C=textscan(fid,'%s %s %s %*[^\n]');
fclose(fid);


% replace the comma with dot:
Data1 = strrep(C{3}, ',', '.');
Data = strrep(Data1, 'n.a.', 'nan');

% find the glucose line:
idx=find(strcmp(C{2},'Glucose'));

% convert the cell string to value:
SugVal.gluc=cellfun(@str2num,Data(idx))*0.18;
if isnan(SugVal.gluc)
    SugVal.gluc=0;
end

idx=find(strcmp(C{2},'Arabinose'));
if not(isempty(idx))
    SugVal.ara=cellfun(@str2num,Data(idx))*0.158;
else
    SugVal.ara=nan;
end

idx=find(strcmp(C{2},'Mannose'));
if not(isempty(idx))
    SugVal.man=cellfun(@str2num,Data(idx))*0.18;
else
     SugVal.man=nan;
end

idx=find(strcmp(C{2},'Xylose'));
if not(isempty(idx))
    SugVal.xyl=cellfun(@str2num,Data(idx))*0.158;
else
    SugVal.xyl=nan;
end

end

function [filesave,SampleName]=FindNewChromeleonFiles_HPLC(clk,par)
% this function read the cpu clock c clk at each iteration, and all the .txt files in the folder
% 'myfolder (change to the suitable chromeleon folder) and then only pick
% the .txt files that were created AFTER the clock registered c. 
% the output is file, the list of files to be read (cell array). 

filesave=[]; % 
TimeStamp=[];
SampleName=[];
%for debugging only!!:
%load c

% HERE, change with the correct Chromeleon folder
%if nargin<1
   %myfolder = 'C:\Users\Public\Documents\Lucullus\import\chromeleon'; % chromeleon folder! 
    %myfolder ='C:\Users\LAB\Desktop\ChromeleonResult'
    myfolder = par.Info.ChromeleonFolder;
    %end
%clk_0=par.clk_0;
% keyboard % for debugging only!
filesInfo=dir(myfolder);
expression = '.txt';

% Start the counting
n=1;

% no Chromeleon files, empty directory
if length(filesInfo)<4
    return
end

for z=3:length(filesInfo)   
    % get only the .txt files in a variable
    if strcmp(filesInfo(z).name(end-3:end),expression)
        
        name{n,:}=filesInfo(z).name;
        tid(n,:)=filesInfo(z).date;
        n=n+1;
    end
end

dato=datestr(tid,'ddmmyy');
%convert 'tid' to numbers:
% Days
for z=1:n-1
    day(z,1)=str2num(tid(z,1:2));
end

% Months
for z=1:n-1
  mon(z,1) = str2num(dato(z,3:4));
end

% Years
for z=1:n-1 
    year(z,1)=str2num(tid(z,8:11));
end

% Hours;
for z=1:n-1 % hours
    timer(z,1)=str2num(tid(z,13:14));
end

% Minutes
for z=1:n-1 % minutes
    minutter(z,1)=str2num(tid(z,16:17));
end

% Seconds
for z=1:n-1 % seconds
    sekunder(z,1)=str2num(tid(z,19:20));
end


ProcessNowSeconds=clk*[0 par.mth*60*60*24 60*60*24 60*60 60 1]';
ProcessStartSeconds=par.clk_0 * [0 par.mth*60*60*24 60*60*24 60*60 60 1]';

% Keep all the files that were created AFTER the clock
% day, month, year are the files, C is the clock.
%ct=1; % counter for the save files
for z=1:n-1
    TotalSeconds=[year(z) mon(z) day(z) timer(z) minutter(z) sekunder(z)]*[0 par.mth*60*60*24 60*60*24 60*60 60 1]';
    if TotalSeconds>ProcessNowSeconds-45
        SampleName= name{z,:};
        filesave=sprintf('%s%s',par.Info.ChromeleonFolder,SampleName);
        TimeStamp= TotalSeconds-ProcessStartSeconds;
        disp('NEW HPLC MEASUREMENTS ACQUIRED')
        return
    end
end

end

% Function to restart in case it is required
function restart(VarMem,par)

% In case the CO2 signal is not taking the correct initial value
VarMem.Instrument.CO2(1)= 0.04;
       

end

% Function to call the Metabolic model / Model integration.
function model = model_DYCOPS(params)
addpath('C:\Matlab\Helene\Casadi\casadi-windows-matlabR2016a-v3.5.5')
import casadi.*

model = struct;

% Fixed values for the parameters
mu_max = params.mu_max;	
Ks = params.Ks; 
kd =  params.kd; 
Y_XS = params.Y_XS; 
Y_CO2X = params.Y_CO2X; 

% Other
q_CO2 = params.q_CO2;
Sin = params.Sin;   % g/L

%% Model Equations
% Define variables
V_mod   = SX.sym('V');
X_mod   = SX.sym('X');
S_mod   = SX.sym('S');
CO2_mod = SX.sym('CO2');

x_mod   = [V_mod, X_mod, S_mod, CO2_mod]';

Fin_mod = SX.sym('Fin');
Fout_mod =  SX.sym('Fout'); %Fin_mod;


% Differential equations
dV_mod = Fin_mod - Fout_mod;
dX_mod = -(X_mod/V_mod)*Fin_mod + mu_max *(S_mod /(Ks + S_mod)) *X_mod - kd *X_mod ; %Biomass
dS_mod = (Sin *Fin_mod) / V_mod - mu_max *(S_mod /(Ks + S_mod)) *X_mod /Y_XS - (S_mod / V_mod) *Fin_mod; % Substrate
dCO2_mod = (mu_max *(S_mod /(Ks + S_mod)) *X_mod)/Y_CO2X - q_CO2*CO2_mod ;

xdot_mod = [dV_mod, dX_mod, dS_mod, dCO2_mod]';

% Define model variables
model.x    = x_mod;
model.xdot = xdot_mod;
model.u    = [Fin_mod, Fout_mod]'; 

end

% ODE solver for model
function [ode_sol] = ODE_solver(system)
addpath('C:\Matlab\Helene\Casadi\casadi-windows-matlabR2016a-v3.5.5')    
import casadi.*
    
    %% System %%
    x = system.x;   % states
    dt = system.dt;
    
    xdot = system.xdot; % Differential equations
    u = system.u;   
    
    %% Continuous time dynamics
    ode = struct('x',x,'p',vertcat(u),'ode',dt*xdot);

    [ode_sol] = integrator('F', 'idas', ode);
    %disp(ode_sol)
end

% Function to Solve the model simulation
function [xk,ModSim] = Process_model(Instrument,x0,ode_sol, VarMem,iteration, ui)

    if iteration==1
        xk = x0;
    else
        xk = [VarMem.Instrument.V(iteration-1), VarMem.Instrument.CDWcu(iteration-1), VarMem.ModSim.S(iteration-1), VarMem.Instrument.CO2_measurement(iteration-1)];
    end
    us = ui;
    sim = ode_sol('x0', xk, 'p', [us] ); % [us(i)] ); 
    
    xk_mod = full(sim.xf); % y;

    % Save the output
    ModSim.V=xk_mod(1);
    ModSim.X=xk_mod(2);
    ModSim.S=xk_mod(3);
    ModSim.CO2=xk_mod(4);

end

% Following: All functions for MHE
% Properties of MHE 
function props = MHEprops_DYCOPS(MHEsys)
    props.d = 3;    % interpolation points
    props.T = 1;    
    props.N = 40;   % horizon
end

% Plant integrator MHE
function F_integrator = makePlantIntegrator_DYCOPS(ODEmodel,OCPprops)
    addpath('C:\Matlab\Helene\Casadi\casadi-windows-matlabR2016a-v3.5.5')
    import casadi.*

    x = ODEmodel.x;
    u = ODEmodel.u;
    x_dot = ODEmodel.x_dot;
    
    timeStep = OCPprops.T/OCPprops.N;
    
    % Variable inputs in a single variable
    p = [u];
    
    ode = struct('x',x,'p',p,'ode',x_dot);
    opts = struct('tf',timeStep);
    
    F_integrator = integrator('F','cvodes',ode,opts);
end

% Arrival cost
function [x_Lplus1, P_Lplus1, pkEig] = arrivalCostUpdate_DYCOPS(MHEsys, MHEprops, x, y, u, P, Q, integrator)   
    addpath('C:\Matlab\Helene\Casadi\casadi-windows-matlabR2016a-v3.5.5')
    import casadi.*

    x_L = x(:,1);
    xtPlus1 = x(:,2);
    nx = length(x_L);
    ny = length(MHEsys.y);
    u0 = u(1,:);    % CHANGE WITH FOÌN & FOUT?
    params = u0';
    P_L = P^(-1/2);
    W = Q^(-1/2);
    V = MHEsys.V;
    y_L = y(1,:)';
    
    timestep = MHEprops.T/MHEprops.N;
    
    Fk = integrator('x0',x_L,'p',params);
    
%% Finite differences
    
    res = integrator('x0', x_L, 'p', params);
    xf = full(res.xf);
    
    h = 0.0001;
    X_x = zeros(nx,nx);
    for i = 1:nx
        x0h = x_L;
        x0h(i) = x0h(i) + h;
        res = integrator('x0', x0h, 'p', params);
        X_x(:,i) = (full(res.xf)-xf)/h;
    end

    % ASSUMING h() is a selection matrix

    H_x = [1     0     0     0
           0     1     0     0
           0     0     0     1];
    
    
%% Other parameters
    
    x_tilde = xf - X_x*x_L;
    h_tilde = H_x*xf - H_x*x_L;
    
    
%% Make and QR factorize matrix

    W_bar = W;
    
    factorizable_matrix = [P_L, zeros(length(P_L),length(W_bar)); ...
                            -((V)*H_x), zeros(length(V),length(W_bar)); ...
                            -W_bar*X_x, W_bar];
                        
    [Q_matrix, R] = qr(factorizable_matrix);
    
    nx_np = length(P_L);
    R1 = R(1:nx_np,1:nx_np);
    R12 = R(1:nx_np, nx_np+1:2*nx_np);
    R2 = R(nx_np+1:2*nx_np, nx_np+1:2*nx_np);
    
    P_Lplus1 = R2;

     constantsVector = [ -P_L*[x_L]; ...
                        (V)*(y_L - h_tilde); ...
                        -W_bar*[x_tilde]];
                    
    rho = Q_matrix\constantsVector;
    
    rho2 = rho(nx_np+1:nx_np*2);
    
    x_and_p = -R2\rho2;
    
    x_Lplus1 = x_and_p(1:nx);

pkEig = diag(P_Lplus1);
    
end

% MHE model
function MHEsys = MHEmodel_DYCOPS(x0, params)
addpath('C:\Matlab\Helene\Casadi\casadi-windows-matlabR2016a-v3.5.5')
import casadi.*

MHEsys = struct;

%% States
V = MX.sym('V');
X = MX.sym('X');
S = MX.sym('S');
CO2 = MX.sym('CO2');

MHEsys.x = [V, X, S, CO2]';
%MHEsys.x0 = [1.5, 0.77519, 15, 0]'; % CHANGE bf : 1.5, 1.2, 20, 0
% param est
MHEsys.x0 =  x0; %[1.48,  0.8,  15,  0]';

MHEsys.x_lower = [1.2,   0,   0,   0]'; %%% !!! before:1.0
MHEsys.x_upper = [1.8, inf, inf, inf]'; %%% !!! before: 2.5

%% Inputs
Fin = MX.sym('Fin');
Fout = MX.sym('Fout'); %Fin;

MHEsys.u = [Fin, Fout]'; %MHEsys.u = [Fin, 
MHEsys.u0 = [0, 0]';

%% Measurements
% We assume that h(x)= Hx is a selector, such that y = x + noise
y_V = MX.sym('y_V');
y_X = MX.sym('y_X');
y_CO2 = MX.sym('y_CO2');

MHEsys.y = [y_V, y_X, y_CO2]';

%% Model Equations

% Here you can set fixed values for the parameters

mu_max = params.mu_max;	
Ks = params.Ks; 
kd =  params.kd; 
Y_XS = params.Y_XS; 
Y_CO2X = params.Y_CO2X; 

% Sin = 100; % g/L
% Param est
Sin = params.Sin;   % g/L


dV = Fin - Fout;
dX = -(X ./V) .*Fin + mu_max .*(S ./(Ks + S)) .*X - kd .*X; %Biomass
dS = (Sin .*Fin) ./ V - mu_max .*(S ./(Ks + S)) .*X ./Y_XS - (S ./ V) .*Fin; % Substrate
dCO2 = (mu_max .*(S ./(Ks + S)) .*X)./Y_CO2X - 2*1.8*60*CO2/100; 

MHEsys.x_dot = [dV, dX, dS, dCO2]';

%% Model noise

% This section is to compute the Jacobian df(x,p)/dp

% Differential equations:
w_mu = MX.sym('w_mu'); 
w_ks = MX.sym('w_ks');
w_Yxs = MX.sym('w_Yxs'); 
w_YCO2 = MX.sym('w_YCO2'); 
w_V = MX.sym('w_V'); 
w_X = MX.sym('w_X'); 
w_S = MX.sym('w_S');
w_CO2 = MX.sym('w_CO2'); 
w_kd = MX.sym('w_kd');

w_mu0 = MX.sym('w_mu0'); 
w_ks0 = MX.sym('w_ks0');
w_Yxs0 = MX.sym('w_Yxs0'); 
w_YCO20 = MX.sym('w_YCO20'); 
w_V0 = MX.sym('w_V0'); 
w_X0 = MX.sym('w_X0'); 
w_S0 = MX.sym('w_S0');
w_CO20 = MX.sym('w_CO20'); 
w_kd0 = MX.sym('w_kd0');


dV_w = Fin - Fout + w_V;
dX_w = -(X ./V) .*Fin + (mu_max + w_mu) .*(S ./((Ks + w_ks) + S)) .*X - (kd + w_kd) .*X + w_X; %Biomass
dS_w = (Sin .*Fin) ./ V - (mu_max + w_mu) .*(S ./((Ks + w_ks) + S)) .*X ./(Y_XS + w_Yxs) - (S ./ V) .*Fin + w_S; % Substrate
dCO2_w = ((mu_max + w_mu) .*(S ./((Ks + w_ks) + S)) .*X)./(Y_CO2X + w_YCO2) - 2*CO2 + w_CO2;

dV_w0 = Fin - Fout + w_V0;
dX_w0 = -(X ./V) .*Fin + (mu_max + w_mu0) .*(S ./((Ks + w_ks0) + S)) .*X - (kd + w_kd0) .*X + w_X0; %Biomass
dS_w0 = (Sin .*Fin) ./ V - (mu_max + w_mu0) .*(S ./((Ks + w_ks0) + S)) .*X ./(Y_XS + w_Yxs0) - (S ./ V) .*Fin + w_S0; % Substrate
dCO2_w0 = ((mu_max + w_mu0) .*(S ./((Ks + w_ks0) + S)) .*X)./(Y_CO2X + w_YCO20) - 2*CO2 + w_CO20;

MHEsys.x_dot_w = [dV_w, dX_w, dS_w, dCO2_w]';
MHEsys.w_Q = [w_mu; w_ks; w_Yxs; w_YCO2; w_V; w_X; w_S; w_CO2; w_kd];

MHEsys.x_dot_w0 = [dV_w0, dX_w0, dS_w0, dCO2_w0]';
MHEsys.w_Q0 = [w_mu0; w_ks0; w_Yxs0; w_YCO20; w_V0; w_X0; w_S0; w_CO20; w_kd0];

G = jacobian(MHEsys.x_dot_w0, MHEsys.w_Q0);
Q_w = diag([w_mu; w_ks; w_Yxs; w_YCO2; w_V; w_X; w_S; w_CO2; w_kd]);
Q_out = G*Q_w*G';
MHEsys.Q_out = Q_out;

L1 = Function('Jtest',{X, S, w_mu, w_ks, w_Yxs, w_YCO2, w_V, w_X, w_S, w_CO2, w_kd, w_mu0, w_ks0, w_Yxs0, w_YCO20, w_V0, w_X0, w_S0, w_CO20, w_kd0},{Q_out});

MHEsys.Q_out = L1;

%% Weighting
Q0_x = [2.0946e-08, 1.0975e-05, 1.0852e-04, 2.1669e-05]'; % Covariance for initial values of x   % CHANGE

%P_in_01.0e+03 *[ 6.9095, 0.3019, 0.0960, 0.2148]
%P iter 2
        %9.9602         0         0         0
        % 0    2.8759    0.0000   -0.0024
        % 0         0   -0.0104    0.0000
        % 0         0         0   -0.7198
MHEsys.P0 = diag(Q0_x.^(-1/2));
R = [0.1, 0.1, 0.001]'; % 0.01, 0.1, 1 Covariance of measurement noise % CO2 =1 ! !     % CHANGE
MHEsys.V = diag(R.^(-1/2));

%% Objective function
yVector = [y_V, y_X, y_CO2]'; % Measurements
xVector = [V, X, CO2]'; % Corresponding states

MHEsys.costfun = (yVector-xVector)'*MHEsys.V'*MHEsys.V*(yVector-xVector);
end

% MHE solver
% A function initializing the optimal control problem
% Based on an implementation of direct collocation by Joel Andersson
function [MHEprop, solver] = moving_horizon_estimation_DYCOPS(MHEmodel, MHEprops, x0)
addpath('C:\Matlab\Helene\Casadi\casadi-windows-matlabR2016a-v3.5.5')
import casadi.*

% Degree of interpolating polynomial
intPolDeg = MHEprops.d;

% Get collocation points
tau_root = [0 collocation_points(intPolDeg, 'legendre')];

% Coefficients of the collocation equation
C = zeros(intPolDeg+1,intPolDeg+1);

% Coefficients of the continuity equation
D = zeros(intPolDeg+1, 1);

% Coefficients of the quadrature function
B = zeros(intPolDeg+1, 1);

% Construct polynomial basis
for j=1:intPolDeg+1
  % Construct Lagrange polynomials to get the polynomial basis at the collocation point
  coeff = 1;
  for r=1:intPolDeg+1
    if r ~= j
      coeff = conv(coeff, [1, -tau_root(r)]);
      coeff = coeff / (tau_root(j)-tau_root(r));
    end
  end
  % Evaluate the polynomial at the final time to get the coefficients of the continuity equation
  D(j) = polyval(coeff, 1.0);

  % Evaluate the time derivative of the polynomial at all collocation points to get the coefficients of the continuity equation
  pder = polyder(coeff);
  for r=1:intPolDeg+1
    C(j,r) = polyval(pder, tau_root(r));
  end

  % Evaluate the integral of the polynomial to get the coefficients of the quadrature function
  pint = polyint(coeff);
  B(j) = polyval(pint, 1.0);
end
% Time horizon
T = MHEprops.T;

% Control discretization
N = MHEprops.N;
h = T/N;

% Declare model variables
y = MHEmodel.y;
x = MHEmodel.x;
u = MHEmodel.u;
V = MHEmodel.V;
nx = length(x);
ny = length(y);
nu = length(u);

% Model equations
xdot = MHEmodel.x_dot;

% Noise term, inherited from the model structure
L = MHEmodel.costfun;

% Continuous time dynamics
f = Function('f', {x, y, u}, {xdot, L});

% Constraints
lbx = MHEmodel.x_lower;
ubx = MHEmodel.x_upper;

lbw = ones(nx,1)*-inf;  %noise
ubw = ones(nx,1)*inf;

% lbw = ones(nx,1).*[-0.2; -0.3; -0.6; -0.2];
% ubw = ones(nx,1).*[0.2; 0.3; 0.6; 0.2];

% Make x0 a column matrix
x0 = reshape(x0,nx,1);

%% Start with an empty NLP

decVar = {}; % Total, final decision variable vector
decVar_Xk = {};
decVar_Wk = {}; % small wi in opt problem
decVar_Xkj = {};
%w={};

decVar0 = []; % Total, final initial guess vector
decVar_Xk_0 = [];
decVar_Wk_0 = [];
decVar_Xkj_0 = [];

%w0 = [];
lbDecVar = [];
ubDecVar = [];

lbXk = [];
ubXk = [];
lbXkj = [];
ubXkj = [];
lbWk = [];
ubWk = [];

J = []; %sum((y - x)'*P*(y - x));
g={};   % Constraints
lbg = [];
ubg = [];

% Variable inputs to MHE

Y_in = MX.sym('Y', N*ny);
Y = reshape(Y_in,[N,ny]);

U_in = MX.sym('U', N*nu); %MX.sym('U', N, nu); 
U = reshape(U_in,[N,nu]);

P_mat_in = MX.sym('P_mat',(nx)^2);
P_mat = reshape(P_mat_in,[nx,nx']);

W_in = MX.sym('W',(N*(nx)^2));
W = reshape(W_in,[nx,N*nx]);

% Initial conditions
X0 = MX.sym('X0', nx);
decVar_Xk = {decVar_Xk{:}, X0}; % Concatinating init. cond into dec_Var_Xk
lbXk = [lbXk; lbx];
ubXk = [ubXk; ubx];
decVar_Xk_0 = [decVar_Xk_0; x0];

Xk = X0;    % Start value

% Variable initial value
X0_var = MX.sym('X0_var',nx);

J = ([X0 - X0_var])'*P_mat'*P_mat*([X0 - X0_var]);

% Formulate the NLP
for k=0:N-1
    
    % New NLP variable for the state noise (USALLY DENOTED BY W)
    Wk = MX.sym(['W_' num2str(k)], nx); % Creating uncertainty number k
    decVar_Wk = {decVar_Wk{:}, Wk};
    lbWk = [lbWk; lbw];
    ubWk = [ubWk; ubw];
    decVar_Wk_0 = [decVar_Wk_0; zeros(nx,1)];
    

    % State at collocation points
    Xkj = {};
    for j=1:intPolDeg   % For all collocation points d
        Xkj{j} = MX.sym(['X_' num2str(k) '_' num2str(j)], nx);   % Creating state number j interpolating between k and k+1: Two variables for two states
        decVar_Xkj = {decVar_Xkj{:}, Xkj{j}};
        lbXkj = [lbXkj; lbx];
        ubXkj = [ubXkj; ubx];
        decVar_Xkj_0 = [decVar_Xkj_0; x0];
    end

    % Loop over collocation points
    Xk_end = D(1)*Xk;
    
    for j=1:intPolDeg
       % Expression for the state derivative at the collocation point
       xp = C(1,j+1)*Xk;
       for r=1:intPolDeg
           xp = xp + C(r+1,j+1)*Xkj{r};
       end

       Yk = Y(k+1, :);
       Uk = U(k+1, :);

       % Append collocation equations
       [fj, qj] = f(Xkj{j}, Yk, Uk);
       g = {g{:}, h*fj - xp};
       lbg = [lbg; zeros(nx,1)];
       ubg = [ubg; zeros(nx,1)];

       % Add contribution to the end state
       Xk_end = Xk_end + D(j+1)*Xkj{j};

       % Add contribution to quadrature function
       J = J + B(j+1)*qj*h;
    end

    % New NLP variable for state at end of interval
    Xk = MX.sym(['X_' num2str(k+1)], nx);
    decVar_Xk = {decVar_Xk{:}, Xk};
    lbXk = [lbXk; lbx];
    ubXk = [ubXk; ubx];
    decVar_Xk_0 = [decVar_Xk_0; x0];

    % Add equality constraint
    g = {g{:}, (Xk_end + Wk) - Xk};
    lbg = [lbg; zeros(nx,1)]; % CHANGE FOR FIN & FOUT ?
    ubg = [ubg; zeros(nx,1)];

    % Add contribution for state noise
    W_k = W(:,(1+nx*k:nx+nx*k));
    J = J + (Wk'*(W_k'*W_k)*Wk);
end

decVar = {decVar_Xk{:}, decVar_Wk{:}, decVar_Xkj{:}};
decVar0 = [decVar_Xk_0; decVar_Wk_0; decVar_Xkj_0];

lbDecVar = [lbXk; lbWk; lbXkj];
ubDecVar = [ubXk; ubWk; ubXkj];


MHEprop.J = J;
MHEprop.w = decVar;
MHEprop.g = g;
MHEprop.w0 = decVar0;
MHEprop.lbw = lbDecVar;
MHEprop.ubw = ubDecVar;
MHEprop.lbg = lbg;
MHEprop.ubg = ubg;

% Calculating corresponding times to each collocation point
repColPoints = [0, repmat([tau_root(2:end) 1]*h,1,N)]';
baseTimes = [0; reshape((0:h:T-h).*ones(1+intPolDeg,1),[N*(intPolDeg+1),1])];

MHEprop.timeseries = repColPoints+baseTimes;    % Timeseries

opts = struct('warn_initial_bounds',false, ...
    'print_time',false, ...
    'eval_errors_fatal',true, ...
    'ipopt',struct('print_level',1) ...
    );

MHE_NLP = struct('f', J, 'x', vertcat(decVar{:}), 'g', vertcat(g{:}), 'p', vertcat(Y_in, U_in, X0_var, P_mat_in, W_in));

% options = struct;
solver = nlpsol('solver', 'ipopt', MHE_NLP, opts);


end

function [x, d] = sortResultsMHE_DYCOPS(results, MHEmodel, MHEprops)
    
    N = MHEprops.N;

    intPolDeg = MHEprops.d;
    
    nx = length(MHEmodel.x);
    
    xvec = results(1:nx*(N+1));
    x = reshape(xvec,[nx,N+1]); 
    
    dvec = results(nx*(N+1)+1:nx*(N+1)+(N*nx));
    d = reshape(dvec,[nx,N]);    

end
