clear all
close all
clc
addpath('C:\Code\Matlab\Casadi\casadi-windows-matlabR2016a-v3.5.5')
import casadi.*

addpath('C:\Code\Matlab\Master\Spring')

addpath('C:\Code\Matlab\Master\Spring\Results')
%addpath('C:\Code\Matlab\Master\Spring\Simulations')
addpath('C:\Code\Matlab\Master\Spring\Parameter_estimation\Estimation')

% Get colors
color = colorHex();
% Get system parameters
system = system_param();
q_CO2 = system.q_CO2;
% Get newest simulation data
%read_sim_data();
% USE LINEAR TUNING = 1, else sinusoidal
linear_tuning=0;
%% Get intended feed
intended = importdata('Values_feed_09_03_23.mat');
%% Offline CDW
filename = 'CDW_20230314.xlsx';

lab_result_table = readtable(filename, 'Sheet','Result_14_03');
lab_results = table2array(lab_result_table(:,4:12)); 

dryweight = [lab_results(2:7,4:6)-lab_results(2:7,1:3)]; % Triplicates

time_CDW = lab_results(2:7,9); 

cell_density = dryweight*1000/3; % g/L, 3 mL samples -> x g in 3mL  -> y g/L

mean_CDW = mean(cell_density,2);
std_CDW = std(cell_density,0, 2);


%% Data retrieval and treatment
mu_max = system.mu_max;	 %0.27399 ;%0.30711 ; 
Ks = system.Ks; %0.0050896; %0.0070378 ;
kd =  system.kd; %0.0013298; %0.0013996 ;
Y_XS = system.Y_XS; %0.12548 ; %0.39212  ;
Y_CO2X = system.Y_CO2X; %0.28222; %0.055265; 

std_dev = system.std_dev;
 
% load experimental data
% This data is made in \Parameter_estimation\Estimation
exp_data = importdata('data_20230314.mat');

% Coose intended data
if linear_tuning
    meas    = [exp_data.V, exp_data.CDW_lin, exp_data.S_tab, exp_data.CO2];
else
    meas    = [exp_data.V, exp_data.X, exp_data.S_tab, exp_data.CO2];
end

ts      = exp_data.ts;    % [s]
Fin      = exp_data.F_in;
Fout     = exp_data.F_out;

t_tot = ts;
tF = ts;

N_points        = length(ts);
Values_output   = meas(1:N_points,:);

Volume = Values_output(:,1);
X      = Values_output(:,2);
CO2    = Values_output(:,4);

% Get sugar values (datarows without NaN values)
[row, col] = find(isnan(Values_output));

S = Values_output(:, 3);
S( row )  = [];
% Timestamp for sugar measurements
ts_S = ts;
ts_S( row )  = [];

% Get feed usage

us = [Fin, Fout]; % [Fin]
% Save runtime !
step = 60; % experimental data was taken every 60 s system.dt;
dt = 1/60; %1/60;

runtime = zeros(length(ts),1);

% Get data for MHE
%t_S = Values_output.tm(1:5:end);
% S = Values_output.S_n(1:5:end);

%t_model = Values_model.tm;
%x_model = [Values_model.V Values_model.X Values_model.S Values_model.CO2];

%rng(1)

%% Setup MHE
x0 = meas(1,:)';
x0(3) = S(1);


MHEsys = MHEmodel_DYCOPS(x0, system);
MHEproblemProps = MHEprops_DYCOPS(MHEsys);

Sin = system.Sin;

N_MHE = MHEproblemProps.N;

[MHEprob, estimator] = moving_horizon_estimation_DYCOPS(MHEsys, MHEproblemProps, x0');

F_integrator = makePlantIntegrator_DYCOPS(MHEsys,MHEproblemProps);

% Create an NLP solver
w = MHEprob.w;
g = MHEprob.g;

MHE_initguess = MHEprob.w0;

% Simulate system
x_bar = x0;
x_ode = x0;
P = MHEsys.P0;
%Q0_x = (1e-11)*[0 0.1878 0 0.0028]';
%P = diag(Q0_x.^(-1/2));

sim = struct;
time = 0;
t = [];
y_V = zeros(N_MHE,1);
y_X = zeros(N_MHE,1);
y_CO2 = zeros(N_MHE,1);
U_in =  zeros(N_MHE,2); 
flag_sol = zeros(N_MHE,1);

wg      = std_dev.^2;  % variance %= noise.w; % White gaussian noise

w_mu = wg(1);%(3.2473e-06)^2;
w_ks = wg(2);%(3.9208e-06)^2;
w_kd = wg(3);%(4.4948e-06)^2;
w_Yxs = wg(4);%(3.5809e-06)^2;
w_YCO2 = wg(5);%(2.2161e-06)^2;

% now: w_add = [8e-5   2e-2     5e-1    1]; 
w_V = 1.0e-2;   % 1.0e-1; old Andrea value
w_X = 9.0e-2; %1e-1;%1.0e-1;   % 1.0e-2;
w_S = 1.0e-2;%1.0;  % 1.0e-2;
w_CO2 = 1.0;    % 1.0e-1; 

w_mu0= 0;
w_ks0 = 0;
w_Yxs0 = 0;
w_YCO20 = 0;
w_V0 = 0;
w_X0 = 0;
w_S0 = 0;
w_CO20 = 0;
w_kd0 = 0;

Q = diag(P).^(-2);
%Q = [2.0946e-08, 1.0975e-05, 1.0852e-04, 2.1669e-05]'; % Initial covariance matrix for P0+
%Q = (diag(Q0_x));
Q =(diag(Q));

%% Integrate system model
model = model_DYCOPS(system);

model.T    = N_points/dt;
model.dt   = dt;

x_model = zeros(N_points,4);
x_model(1, :) = x0;

% GET INTEGRATOR
model.noise = 0;
[ode_sol]    = ODE_solver(model); 
xk = x0;
t_model = zeros(N_points,1);
% Integrate model
for i= 2:N_points
    % Get measurements, sugar is just previous state
    %xk = [Volume(i-1), X(i-1), x_model(i-1,3), CO2(i-1)];
    xk = [x_model(i-1,1), x_model(i-1,2), x_model(i-1,3), x_model(i-1,4)];
    
    sim = ode_sol('x0', xk, 'p', [us(i-1, :)] ); % [us(i)] ); % must check to use us(i) or us(i-1)
    
    xk_mod = full(sim.xf); % y;
    
    % Update states
    x_model(i,:)   = xk_mod;   
    t_model(i) =  t_model(i-1) + dt;
end


%% Start simulation

% Time simulation
start = tic;
tStart = cputime;

yk = x0;
t = [0];
for i = 1:N_points %24 h
    trun = tic;
    fprintf('>>> Iteration: %d \n',i)
    yk = meas(i,:)';
    W = Q^(-1/2);

    W_Array(:,:,i) = W;

    % GET MEASUREMENTS
    if i <= N_MHE % before first horizon -> Filling up measurements
        y_V(i:end)    = yk(1);       %xs((N_batch-N_MHE+i):(N_batch+i-1),1); % Get batch data
        y_X(i:end)    = yk(2);       %xs((N_batch-N_MHE+i):(N_batch+i-1),2);
        y_CO2(i:end)  = yk(4);      %xs((N_batch-N_MHE+i):(N_batch+i-1),4);
        U_in(i,:) = us(i,:);%us(i);
    else
        y_V(1:end-1) = y_V(2:end);
        y_V(end) = yk(1);

        y_X(1:end-1) = y_X(2:end);
        y_X(end) = yk(2);

        y_CO2(1:end-1) = y_CO2(2:end);
        y_CO2(end) = yk(4);

        U_in(1:end-1, :) = U_in(2:end, :);
        U_in(end, :) = us(i,:); % us(i)
    end

    y = [y_V, y_X, y_CO2];

    if i ~= 1

        [xL, PL, PLEig] = arrivalCostUpdate_DYCOPS(MHEsys, MHEproblemProps, x, y, U_in, P, Q, F_integrator);

        x_residual(:,i-1) = xL-x(:,1);
        P_residual(:,:,i-1) = PL-P;
        P_eig(:,i-1) = PLEig;

        x_bar = xL;
        P = PL;
    end

    Y_in = vertcat(y_V,y_X,y_CO2); 
    x_bar_in = x_bar;
    P_in = reshape(P,[length(P).^2,1]);
    U_in_in = reshape(U_in,[N_MHE * 2, 1]);

    %% This needs fixing

    if i <= N_MHE
        for iter = 1:N_MHE
            W_Array_presample(:,:,iter) = W;
        end
        W_in = reshape(W_Array_presample(:,:,:),[N_MHE*(length(W).^2),1]);
    else
        W_in = reshape(W_Array(:,:,i-N_MHE+1:i),[N_MHE*(length(W).^2),1]);
    end

    %%
    MHE_variables = vertcat(Y_in, U_in_in, x_bar_in, P_in, W_in);

    % Solve the NLP
    sol = estimator('x0', MHE_initguess, 'p', MHE_variables, 'lbx', MHEprob.lbw, 'ubx', MHEprob.ubw,...
                'lbg', MHEprob.lbg, 'ubg', MHEprob.ubg);
    w_opt = full(sol.x);
    MHE_initguess = w_opt;

    [x, d] = sortResultsMHE_DYCOPS(w_opt, MHEsys, MHEproblemProps);

    flag_sol(i) = estimator.stats.success;

%% If simulation fails
if false 
    disp(i)
    disp(time)
    % PLOT
end

    %% Update data
    if i < N_MHE
        xk          = x(:,i);
        x_est(:,i)  = xk; 
        t = [t, time];
    else
        xk          = x(:,end);
        x_est       = [x_est, xk];
        t = [t, time];
    end


    if i < N_MHE
        d_est(:,i) = d(:,i);
    else
        d_est = [d_est, d(:,end)];
    end

    for j = 1:size(x0,1)
    d_Array(j,i) = d(j,:)*d(j,:)';
    end

    P = P'*P;
    PArray(:,i) = diag(P);

    %if time >= 22 
    %    w_ks = (0.1839208)^2;
    %    w_YCO2 = (0.22161)^2;   
    %end
     %disp(x_est)   

    L1_cas = MHEsys.Q_out(x_est(2,end), x_est(3,end), w_mu, w_ks, w_Yxs, w_YCO2, w_V, w_X, w_S, w_CO2, w_kd, w_mu0, w_ks0, w_Yxs0, w_YCO20, w_V0, w_X0, w_S0, w_CO20, w_kd0);
    Q = full(L1_cas);

    if i < N_points
        uk = us(i+1);
    else
        uk = 0;
    end
    %time = time + 1/step;
    time = time + dt;

    runtime(i) = toc(trun);
end

tEnd = cputime - tStart
endtime = toc(start)

disp("Max time")
disp(max(runtime))
disp("Min time")
disp(min(runtime))

disp("Avg runtime")
disp(sum(runtime) / length(runtime))

% Add first state and last time
%x_est = [x0,x_est]';
%% RSME
x_est= x_est';
% Offline CDW
t_CDW = zeros(length(mean_CDW),1);
mod_CDW = zeros(length(mean_CDW),1);
MHE_CDW = zeros(length(mean_CDW),1);

for i = 1:length(mean_CDW)
    %i_CDW = find(ts <= mean_CDW(i), 1, 'last');
    i_CDW = find(ts <= time_CDW(i), 1, 'last');
    t_CDW(i) = ts(i_CDW);
    mod_CDW(i) = x_model(i_CDW,2);
    MHE_CDW(i) = x_est(i_CDW,2);
end


% Offline Sugar
S_model = x_model(:,3);
S_model( row )  = [];
%S_model = [zeros(1,length(x_model(:,2))-length(S)), S_model']';
S_est = x_est(:,3);
S_est( row )  = [];
%S_est = [zeros(1,length(x_est(:,2))-length(S)), S_est']';

%S_long = [zeros(1,length(x_est(:,2))-length(S)), S']';

% Calculate Root Mean Squared Error -> MHE- X, MHE - CDW offline , MHE - S offline
MS_model    = [ sum((x_model(:,2) - X).^2) sum((mod_CDW - mean_CDW).^2) sum((S_model -  S).^2)];
MS_MHE    = [ sum((x_est(:,2) - X).^2) sum((MHE_CDW - mean_CDW).^2) sum((S_est -  S).^2)];

RMSE_model  = sqrt(MS_model./[N_points length(mean_CDW)  length(S)]); %sqrt(sum(MS_model)/N_points);
RMSE_MHE    = sqrt(MS_MHE./[N_points length(mean_CDW)  length(S)]);%sqrt(sum(MS_MHE)/N_points);
disp("Root mean squared error, model:")
disp(RMSE_model)
disp("Root mean squared error, MHE:")
disp(RMSE_MHE)

%% Printing Paper
color = colorHex();

t = t(1:end-1)';

clf;

 set(0, 'DefaultTextFontName', 'Computer Modern',...
     'DefaultTextFontSize', 10,...
     'DefaultAxesFontName', 'Computer Modern',...
     'DefaultAxesFontSize', 10,...
     'DefaultLineLineWidth', 1.2)

set(gcf,'Position',[450 300 650 580])


tile  = tiledlayout(2,2,'TileSpacing','Compact','Padding','Compact');

%% Volume
nexttile
    hold on
    plot(t_model, x_model(:,1), '--','Color', 'k')
    plot(t,    x_est(:,1),   '-','Color', color.b)
    plot(ts,      Volume,            '-.','Color', [0.5 0.5 0.5])
    yline(1.5,'--', 'Desired')

    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('V [L]')
    %legend('boxoff')
   % legend('Model','MHE','Measurement','location', 'best')

%% Biomass
nexttile
    hold on
    plot(t_model, x_model(:,2), '--','Color', 'k')
    plot(ts,      X,            '-.','Color', [0.5 0.5 0.5])
    plot(t,    x_est(:,2),   '-','Color', color.b)
    plot(time_CDW, mean_CDW, 'x', 'Color', color.r)
    errorbar(time_CDW, mean_CDW, std_CDW, 'Color', color.p, 'LineStyle', 'none')
    % MANUAL SAMPLES
    %err = std_CDW;
    %errorbar(time,y,err,'ob')

    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('X [g/L]')
    legend('boxoff')
    legend('Model', 'Measurement', 'MHE','Offline','location', 'best')

%% Sugar
nexttile
    hold on
    plot(t_model, x_model(:,3), '--','Color', 'k')
    plot(t,       x_est(:,3),   '-','Color', color.b)
    %plot(t_S, S,'x','Color', color.p)
    plot(ts_S,    S,            'x','Color', color.r)
    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('S [g/L]')
    %legend('boxoff')
    %legend('Model','MHE','Offline','location', 'best')    %'off-line S',

%% CO2
nexttile
    hold on
    plot(t_model, x_model(:,4), '--','Color', 'k')
    plot(ts,      CO2,          '-.','Color', [0.5 0.5 0.5])
    plot(t,       x_est(:,4),   '-','Color', color.b)

    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('CO2 [% of gas outlet]')
    %legend('boxoff')
    %legend('Model','MHE','Measurement','location', 'best')
    
sgtitle("MHE tuning")
%% Feeding

% create new figure
fig = figure; clf
    box off
    hold on
    plot(t, Fin,'Color', color.b)
    plot(t, Fout', '-.', 'Color', color.r)
    xlabel('Time [h]')
    ylabel('Flow [L/h]')
    legend('F_{in}', 'F_{out}') 
    title("Flow in & out")

%% Feeding
fig = figure; clf
set(gcf,'Position',[450 300 550 200])
box off
    colororder({'k','k'})
    hold on
    plot(t, Fin,'Color', color.b)
    plot(intended.time, intended.feed, '--', 'Color', [0.5 0.5 0.5])
    plot(t, Fout, '-.', 'Color', color.r)
    ylabel('Flow [L/h]')
    xlabel('Time [h]')
    legend('F_{in}', 'Intended', 'F_{out}') 
    title('Flow in & out, parameter estimation', 'FontSize', 10,'FontWeight','Normal')
%% Feeding & O2 consumption
% load experimental data
exp_data = load('20230314_None_None_Glucose_A2.mat', 'VarMem');

VarMem = exp_data.VarMem;

% create new figure
fig = figure; clf
tile  = tiledlayout(2,1,'TileSpacing','Compact','Padding','Compact');
nexttile
    box off
    hold on
    plot(VarMem.Instrument.t/3600, VarMem.Instrument.Flow_in,'Color', color.b)
    plot(VarMem.Instrument.t/3600, VarMem.Instrument.Flow_out, '-.', 'Color', color.r)
    xlabel('Time [h]')
    ylabel('Flow [L/h]')
    title("Flow in & out")
    xlim([0,22])
nexttile
    box off
    hold on
    plot(VarMem.Instrument.t/3600, VarMem.Instrument.pO2,'Color', color.b)
    xlabel('Time [h]')
    ylabel('pO2')
    legend('pO2') 
    title("Oxygen level in reactor")
    xlim([0,22])
     
sgtitle("MHE tuning")


