addpath('C:\Code\Matlab\Master\Spring')
addpath('C:\Code\Matlab\Master\Spring\Results')

color = colorHex();

%% Get CDW sample data
filename = 'CDW_20230314.xlsx';

lab_result_table = readtable(filename, 'Sheet','Result_14_03');
lab_results = table2array(lab_result_table(:,4:12)); 

dryweight = [lab_results(:,4:6)-lab_results(:,1:3)]; % Triplicates

time_CDW = lab_results(:,9); 

cell_density = dryweight*1000/3; % g/L, 3 mL samples -> x g in 3mL  -> y g/L

mean_CDW = mean(cell_density,2);
std_CDW = std(cell_density,0, 2);


%% Excel data - from missing time period
filename = 'A2_outp_20230314_extended.xlsx'; % Experiment 09/03


Table_excel = readtable(filename);

t_exc = Table_excel.Time_h_(1:26*60);
CO2_exc = Table_excel.ExitCO2(1:26*60);
OD_exc = Table_excel.OD(1:26*60);
pO2_exc = Table_excel.pO2(1:26*60);
acid_exc = Table_excel.Pump_Val_Acid(1:26*60);
base_exc = Table_excel.Pump_Val_Base(1:26*60);

NIR_exc = OD_exc;

for i = 2:length(NIR_exc)
    % Gaussian filter
    NIR_smooth= smoothdata(OD_exc(1:i),'lowess',80);
    NIR_exc(i) = NIR_smooth(end);
end

NIR_exc(NIR_exc <0 ) = 0;
CDW_exc = -15212.545*sin(2.0532*NIR_exc  +2.5893) +15197.4245*sin(2.055*NIR_exc +2.5887);


figure(5)
subplot(2,1,1)
    hold on
    colororder({'k','k'})
    plot(t_exc, OD_exc, 'Color', color.b)
    ylabel('Absorbance [CU]')
    yyaxis right
    plot(t_exc, pO2_exc, 'Color', color.r)
    ylabel('Oxygen')
    xlabel('Time [h]')
    legend('OD', 'pO_2',  'Location', 'best')
subplot(2,1,2)
    hold on
    plot(t_exc, acid_exc, 'Color', color.b)
    plot(t_exc, base_exc, 'Color', color.r)
    ylabel('')
    xlabel('Time [h]')
    legend('Acid', 'Base',  'Location', 'best')

figure(6)
hold on
    plot(t_exc, CO2_exc,'Color', color.b)
    xlabel('t [h]')
    ylabel('% of outlet')
    title('Offgas')
    yyaxis right
    plot(t_exc, NIR_exc, 'Color', color.r)
    ylabel('Absorbance [CU]')
    legend('CO_2', 'NIR', 'Location', 'NorthWest')%'Position',[0.8 0.5 0.15 0.08]);
    

%% Get experimental data 
% load experimental data
exp_data = load('20230314_None_None_Glucose_A2.mat', 'VarMem');

meas_    = exp_data.VarMem.measurements;

% time 
tF  = exp_data.VarMem.Instrument.t'/3600;
ts_      = exp_data.VarMem.time;    % [s]

N = length(ts_);
% Remove nan_values /
meas    = meas_(1:N,:);
ts      = ts_(1:N)'/3600;   %sec -> h

sugar = exp_data.VarMem.Instrument.HPLC.gluc';

%% Extra measurements
NIR =  exp_data.VarMem.Instrument.NIR_smooth(:,1:N)'; % NIR
pO2 =  exp_data.VarMem.Instrument.pO2(:, 1:N)'; %pO2

acid =  exp_data.VarMem.Instrument.Acid_pump(:,1:N)';
base =  exp_data.VarMem.Instrument.Base_pump(:,1:N)';


%% Flows
V_in    = exp_data.VarMem.Instrument.Volume_in(:,1:N)';
V_out   = exp_data.VarMem.Instrument.Volume_out(:,1:N)';

F_in    = exp_data.VarMem.Instrument.Flow_in(:,1:N)';
F_out   = exp_data.VarMem.Instrument.Flow_out(:,1:N)';

pump_in    = exp_data.VarMem.Instrument.Flow_in(:,1:N)';    % in decimal not percentage!
pump_out   = exp_data.VarMem.Instrument.Flow_out(:,1:N)';   % in decimal not percentage!

F_in_ctr    = exp_data.VarMem.Controller.Fin(:,1:N)';
F_out_ctr   = exp_data.VarMem.Controller.Fout(:,1:N)';

pump_in_ctr    = exp_data.VarMem.Controller.Pumpin(:,1:N)';
pump_out_ctr   = exp_data.VarMem.Controller.Pumpout(:,1:N)';



%% Get Sugar data
% Excel + process
m_gluc = 180.156;   % g/mol
dil = 10;           % dilution

filename = 'A2_sugar_20230314_HPLC.xlsx';

sugar_table = readtable(filename, 'Sheet','Sheet2');
sugar_array= table2array(sugar_table(:,3:4)); 

s_HPLC = sugar_array(1:end-1,1);  %mmol/L
s_HPLC = s_HPLC*0.001*dil*m_gluc;

t_S = sugar_array(1:end-1,2);

% Remove faulty data (12:12 sample)
s_HPLC = [s_HPLC(1)', s_HPLC(3:end)']';
t_S = [t_S(1)', t_S(3:end)']';


%% MATCH DATA

start_process_exc =  find(t_exc <= 2.750138889, 1, 'last');

t_start_glitch = 6.600138889;% Excel



i_s_exc_glitch = find(ts <= t_start_glitch, 1, 'last') + 1; % Excel - start one after glitch
num_missing_data = 172;
i_e_exc_glitch = i_s_exc_glitch + num_missing_data -1 ; % Excel - stop 1 before glitch end

i_s_process_glitch = 230; % start iteration
i_e_process_glitch = 231; % end

ts_new = 1:1:num_missing_data;         % missing time points 172 -> 2.863266111
ts_new = ts_new*60/3600 + ts(i_s_process_glitch); % add from latest timepoint [h]

ts = [ts(1:i_s_process_glitch)', ts_new, ts(i_e_process_glitch:end)' + (num_missing_data*60/3600) ]'; %[h]
tF = [tF(1:i_s_process_glitch)', ts_new, tF(i_e_process_glitch:end)' + (num_missing_data*60/3600) ]'; %[h]
% Update data
% MUST MAKE NEW matrices as sizes are fixed!
% Add bulk of missing data
NIR = [NIR(1:i_s_process_glitch)', NIR_exc(i_s_exc_glitch:i_e_exc_glitch)', NIR(i_e_process_glitch:end)']';

% NEED TO UPDATE VOLUME
% print(meas(230:231,1)) % 1.5008, 1.5008
V = [meas(1:i_s_process_glitch, 1)', ones(1,num_missing_data)*meas(230,1), meas(i_e_process_glitch:end, 1)']';
CDW   = [ meas(1:i_s_process_glitch, 2)', CDW_exc(i_s_exc_glitch:i_e_exc_glitch)',  meas(i_e_process_glitch:end, 2)']';
CO2  = [ meas(1:i_s_process_glitch, 4)', CO2_exc(i_s_exc_glitch:i_e_exc_glitch)', meas(i_e_process_glitch:end, 4)']';
sugar = [meas(1:i_s_process_glitch, 3)', ones(1,num_missing_data)*NaN, meas(i_e_process_glitch:end, 3)']';
% Update meas
meas = zeros(length(V), 4);
meas(:,1) = V;
meas(:,2) = CDW;
X = CDW;
meas(:,3) = NaN;
meas(:,4) = CO2;

% Flows / refernece
F_in  = [F_in(1:i_s_process_glitch, 1)', zeros(1,num_missing_data),F_in(i_e_process_glitch:end, 1)']';
F_out  = [F_out(1:i_s_process_glitch, 1)', zeros(1,num_missing_data),F_out(i_e_process_glitch:end, 1)']';
F_in_ctr  = [F_in_ctr(1:i_s_process_glitch, 1)', zeros(1,num_missing_data),F_in_ctr(i_e_process_glitch:end, 1)']';
F_out_ctr  = [F_out_ctr(1:i_s_process_glitch, 1)', zeros(1,num_missing_data),F_out_ctr(i_e_process_glitch:end, 1)']';
pump_in  = [pump_in(1:i_s_process_glitch, 1)', zeros(1,num_missing_data),pump_in(i_e_process_glitch:end, 1)']';
pump_out  = [pump_out(1:i_s_process_glitch, 1)', zeros(1,num_missing_data),pump_out(i_e_process_glitch:end, 1)']';
pump_in_ctr  = [pump_in_ctr(1:i_s_process_glitch, 1)', zeros(1,num_missing_data),pump_in_ctr(i_e_process_glitch:end, 1)']';
pump_out_ctr  = [pump_out_ctr(1:i_s_process_glitch, 1)', zeros(1,num_missing_data),pump_out_ctr(i_e_process_glitch:end, 1)']';

%pO2( i_start_glitch: i_end_glitch) =  pO2_exc(i_start_glitch_exc:i_end_glitch_exc);
pO2 = [pO2(1:i_s_process_glitch)', pO2_exc(i_s_exc_glitch:i_e_exc_glitch)', pO2(i_e_process_glitch:end)']';
acid = [acid(1:i_s_process_glitch)', acid_exc(i_s_exc_glitch:i_e_exc_glitch)', acid(i_e_process_glitch:end)']';
i_acid = find(acid <=  23.0582, 1, 'last');
acid(i_acid) = acid(i_acid+1);
base = [base(1:i_s_process_glitch)', base_exc(i_s_exc_glitch:i_e_exc_glitch)', base(i_e_process_glitch:end)']';

%% Old CDW conversion
CDW_Andrea = zeros(length(NIR), 1);
CDW_Helene = zeros(length(NIR), 1);
for i=1:length(NIR)
    NIR_i = NIR(i);
    if NIR_i < 0.9
        CDW_Andrea(i) = 11.124*NIR_i + 0.66116;
    else
        CDW_Andrea(i) = 22.187*NIR_i -5.0991;
    end
    
    if NIR_i < 0.6
        CDW_Helene(i) = 7.959*NIR_i -0.116566;
    else
        CDW_Helene(i) = 22.2695*NIR_i -8.196;
    end
end

%% ACID BASE
% Convert g acid/base added to volume
rho_acid = 1053; % g/L 
rho_base = 1170; % g/L
v_acid = zeros(length(NIR), 1);
v_base = zeros(length(NIR), 1);

for i = 2:length(NIR)
    v_acid(i) = (acid(i)-acid(i-1))/rho_acid;
    v_base(i) = (base(i)-base(i-1))/rho_base;
end

% Add volume difference to V measurement
%meas(:,1) = meas(:,1) + v_acid + v_base;   % REMOVED BECAUSE a third input
%must be added to model dv = Fin - Fout + acid_base;

%% Find Sugar measurements ( remove nan rows )
[row, col] = find(isnan(sugar));

% Remove nan rows by index => all data is same size
%S = meas(:, 3);
sugar( row )  = [];
ts_S = tF;
ts_S( row )  = [];
tS = [ts_S(1)',ts_S(3:4)', t_S(5:6)', ts_S(5:end)']'; % Add missing time stamps
S = [s_HPLC', sugar(10:end)']'; % HPLC - 12 data points, but sugar is missing 2 data points sp --> 10

[n_rows, n_cols] = size(tF);

%% Flow calculations
% Allocate space
F_in_calc = zeros(n_rows,1); F_out_calc = zeros(n_rows,1);

V_in  = [V_in(1:i_s_process_glitch, 1)', zeros(1,num_missing_data),V_in(i_e_process_glitch:end, 1)']';
V_out  = [V_out(1:i_s_process_glitch, 1)', zeros(1,num_missing_data),V_out(i_e_process_glitch:end, 1)']';
for i = 1:n_rows-1
    tstep = (tF(i+1,1) - tF(i,1)); % h
  
    dV_in    = V_in(i+1) - V_in(i);
    dV_out   = V_out(i+1) - V_out(i);
    
    F_in_calc(i,1) = dV_in/tstep;     % L/h     
    F_out_calc(i,1)= dV_out/tstep;    % L/h 
end

%% If setpoint
X_goal= 15;

%% Plotting
% Get colors
addpath('C:\Code\Matlab\Master\Spring')
color = colorHex();

clf;

 set(0, 'DefaultTextFontName', 'Computer Modern',...
     'DefaultTextFontSize', 10,...
     'DefaultAxesFontName', 'Computer Modern',...
     'DefaultAxesFontSize', 10,...
     'DefaultLineLineWidth', 1.2)

set(gcf,'Position',[450 300 650 580])

figure(1)
subplot(4,1,1), %plot states
    hold on;
    %colororder({'k','k'})
    plot(ts, V, 'Color', color.b)
    xlabel('t [h]')
    ylabel('Volume [L]')
    ylim([1.4 1.7])
    %yyaxis right
    %plot(tF, F_in, 'Color', color.r, 'LineStyle','-.')
    %plot(tF, F_out, 'Color', color.p,'LineStyle','--')
    %ylabel('L/h')
    %ylim([-0.1 0.1])
    %legend({'V: Volume', 'F_{in}', 'F_{out}'}, 'Location', 'NorthWest');%'Position',[0.8 0.9 0.15 0.08]);
    title('Volume')
    legend('V')
    legend('boxoff')
    xlim([0, ts(end)]);
    

subplot(4,1,2), %plot states
    hold on
    plot(ts, X, 'Color', color.b)
    plot(time_CDW, mean_CDW, 'x', 'Color', color.r)
    errorbar(time_CDW, mean_CDW, std_CDW, 'Color', color.r,'LineStyle', 'none')
    yline(X_goal,'--', 'X_{goal}')
    %ylim([0,X_goal+3])
    xlabel('t [h]')
    ylabel('Concentration [g/L]')
    legend({'X: Cell', 'Offline samples'},'Location', 'NorthWest')%'Position',[0.8 0.7 0.15 0.08]);
    legend('boxoff')
    title('Biomass production')
    xlim([0, ts(end)]);
    
subplot(4,1,3), %plot states
    hold on
    
    scatter(tS, S, 'Marker', 'x', 'MarkerFaceColor', color.b ,'MarkerEdgeColor', color.b )
    plot(tS, S,'Color',color.p, 'LineStyle','-')
    %yline(X_goal,'--', 'X_{goal}')
    %ylim([0,X_goal+3])
    xlabel('t [h]')
    ylabel('Concentration [g/L]')
    legend({'S: Glucose offline samples'},'Location', 'NorthEast')%'Position',[0.8 0.7 0.15 0.08]);
    legend('boxoff')
    title('Sugar concentration')
    xlim([0, ts(end)]);


 subplot(4,1,4), %plot states
    hold on;
    plot(ts, CO2,'Color', color.b)
    xlabel('t [h]')
    ylabel('% of gas outlet')
    legend({'CO_2'}, 'Location', 'NorthWest')%'Position',[0.8 0.5 0.15 0.08]);
    legend('boxoff')
    title('Offgas')
    xlim([0, ts(end)]);
    %xlim([0 58])
% subplot(4,1,4), %plot states
%     hold on;
%     colororder({'k','k'})
%     stairs(tm, us, 'Color', color.r, 'LineWidth',0.9)
%     xlabel('t [h]')
%     ylabel('L/h')
%     yyaxis right
%     ylabel('Cost')
%     plot(tm, cost, '--', 'Color',color.b)
%     legend('Fin', 'Cost', 'Location', 'best')%'Position',[0.8 0.25 0.15 0.08]);
%     title('Feed')
sgtitle(" 14/03/23 Experimental data")
figure(2)
subplot(1,2,1)
    hold on
    plot(tF, F_in, 'Color', color.b)
    plot(tF, F_out, 'Color', color.r)
    plot(tF, F_in_calc, 'Color', color.o)
    plot(tF, F_out_calc, 'Color', color.p)
    plot(tF, F_in_ctr,'-.', 'Color',  [.7 .7 .7])
    plot(tF, F_out_ctr, '-.', 'Color', [0 0 0])
    ylabel('L/h')
    xlabel('Time [h]')
    legend('Fin', 'Fout', 'Fin calc', 'Fout calc','Fin intended', 'Fout intended', 'Location', 'best')
    legend('boxoff')
subplot(1,2,2)
    hold on
    plot(tF, pump_in*100, 'Color', color.b)
    plot(tF, pump_out*100, 'Color', color.r)
    plot(tF, pump_in_ctr,'--', 'Color', color.p)
    plot(tF, pump_out_ctr, '--', 'Color', color.o)
    ylabel('%')
    xlabel('Time [h]')
    legend('Pump in', 'Pump out', 'Pin intended', 'Pout intended', 'Location', 'best')
sgtitle('Inlet and outlet')  
   
figure(3)
subplot(2,1,1)
    hold on
    colororder({'k','k'})
    plot(tF, F_in, 'Color', color.b)
    plot(tF, F_out, '--', 'Color', color.o)
    ylabel('Flow [L/h]')
    yyaxis right
    plot(tF, pO2, 'Color', color.r)
    ylabel('Oxygen')
    xlabel('Time [h]')
    ylim([10 90])
    legend('F_{in}', 'F_{out}', 'pO_2',  'Location', 'SouthWest')
    legend('boxoff')
subplot(2,1,2)
    hold on
    plot(tF, acid, 'Color', color.b)
    plot(tF, base, 'Color', color.r)
    ylabel('[g]')
    xlabel('Time [h]')
    legend('Acid', 'Base',  'Location', 'best')
    legend('boxoff')
sgtitle('Various measurements')
figure(4) 
    hold on
    colororder({'k','k'})
    plot(tF, NIR, 'Color', color.b)
    ylabel('Absorbance [CU]')
    yyaxis right
    plot(ts, X, 'Color', color.r)
    plot(tF, CDW_Andrea, '--','Color', color.o)
    plot(tF, CDW_Helene, '--','Color', color.p)
    plot(t_exc(start_process_exc:end) - 2.750138889, CDW_exc(start_process_exc:end))
    plot(time_CDW, mean_CDW, 'x', 'Color', color.r)
    errorbar(time_CDW, mean_CDW, std_CDW, 'LineStyle', 'none')
    ylabel('CDW [g/L]')
    xlabel('Time [h]')
    legend('NIR', 'NIR to CDW', 'CDW Andrea', 'CDW Helene linear', 'NIT to CDW excel', 'Offline samples', 'Location', 'best')
    legend('boxoff')
    title('Comparison NIR & NIR to CDW') 
    
%% Feed PO2
% create new figure
fig = figure; clf
tile  = tiledlayout(2,1,'TileSpacing','Compact','Padding','Compact');
nexttile
    box off
    hold on
    plot(tF, F_in,'Color', color.b)
    plot(tF, F_out,'-.', 'Color', color.r)
    xlabel('Time [h]')
    ylabel('Flow [L/h]')
    title("Flow in & out")
nexttile
    box off
    hold on
    plot(tF, pO2,'Color', color.b)
    xlabel('Time [h]')
    ylabel('pO2')
    legend('pO2') 
    title("Oxygen level in reactor")
     
sgtitle("MHE tuning")



clf;

 set(0, 'DefaultTextFontName', 'Computer Modern',...
     'DefaultTextFontSize', 10,...
     'DefaultAxesFontName', 'Computer Modern',...
     'DefaultAxesFontSize', 10,...
     'DefaultLineLineWidth', 1.2)

set(gcf,'Position',[450 300 650 580])



tile  = tiledlayout(2,2,'TileSpacing','Compact','Padding','Compact');

%% Volume
nexttile
    hold on
    plot(ts, V,   '-','Color', color.b)
    yl = yline(1.5,'--', 'Desired');
    yl.LabelHorizontalAlignment = 'left';
    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('V [L]')
    %legend('boxoff')
   % legend('Model','MHE','Measurement','location', 'best')

% X
nexttile
    hold on
    plot(ts, X, 'Color', color.b)
    plot(time_CDW, mean_CDW, 'x', 'Color', color.r)
    errorbar(time_CDW, mean_CDW, std_CDW, 'Color', color.p,'LineStyle', 'none')
    % MANUAL SAMPLES
    %err = std_CDW;
    %errorbar(time,y,err,'ob')

    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('X [g/L]')
    legend('boxoff')
    legend('Measurement','Offline','location', 'best')
    %legend('Model','Measurement','MHE','location', 'best')

%% Sugar
nexttile
    hold on
    plot(tS, S,'x','Color', color.r)
    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('S [g/L]')
    %legend('boxoff')
    %legend('Model','MHE','Offline','location', 'best')    %'off-line S',

%   CO2
nexttile
    hold on
    plot(ts, CO2,   '-','Color', color.b)
    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('CO2 [% of gas outlet]')
    %legend('boxoff')
    %legend('Model','MHE','Measurement','location', 'best')
    
sgtitle("Parameter estimation")


%% Linear CDW
CDW_lin = NIR;
% MY ORIGINAL
CDW_lin(NIR > 0.9) = 22.2695*CDW_lin(NIR > 0.9) - 8.1916;
CDW_lin(NIR <= 0.9) = 7.959*CDW_lin(NIR <= 0.9) - 0.11656;


%% Save updated data
% find corresponding sugar points
for i = 1:length(S)
    i_S = find(ts <= tS(i), 1, 'last');
    meas(i_S,3) = S(i);
    disp(ts(i_S))
    disp(tS(i))
    % OK EFFECTIVELY EQUAL times
end



V = meas(:,1); X = meas(:,2); S_tab = meas(:,3); CO2 = meas(:,4);
Table_20230314 = table(ts, V, X, S_tab, CO2, F_in, F_out, CDW_lin);

data_20230314= table2struct(Table_20230314, "ToScalar", true);

%save('Values_feed.mat', 'Values_feed')
save('data_20230314.mat', 'data_20230314')



