addpath('C:\Code\Matlab\Master\Spring')
%% Get experimental data 
% load experimental data
% open .mat file first seperately in matlab before to see full structure
% use "clear da" and "clear data" in console for avoiding warnings
exp_data = load('Results/20230215_EKF_MPC_Glucose_A2.mat', 'VarMem');

meas_    = exp_data.VarMem.measurements;
V_in    = exp_data.VarMem.Instrument.Volume_in';
V_out   = exp_data.VarMem.Instrument.Volume_out';
tF  = exp_data.VarMem.Instrument.t'/3600;
ts_      = exp_data.VarMem.time;    % [s]

N = length(ts_);
% Remove nan_values / Collect measurements only at points where sugar measurements exists
meas    = meas_(1:N,:);
ts      = ts_(1:N)'/3600;   %sec -> h

%Find nan rows
[row, col] = find(isnan(meas));

% Remove nan rows by index => all data is same size
S = meas(:, 3);
S( row )  = [];
ts_S = ts;
ts_S( row )  = [];

% Sort data
V = meas(:,1);
X = meas(:,2);
CO2 = meas(:, 4);

[n_rows, n_cols] = size(tF);

% Allocate space
F_in = zeros(n_rows,1); F_out = zeros(n_rows,1);

for i = 1:n_rows-1
    tstep = (tF(i+1,1) - tF(i,1)); % h
  
    dV_in    = V_in(i+1) - V_in(i);
    dV_out   = V_out(i+1) - V_out(i);
    
    F_in(i,1) = dV_in/tstep;     % L/h     
    F_out(i,1)= dV_out/tstep;    % L/h 
end

%% If setpoint
X_goal= 10;
%% Plotting
% Get colors
addpath('C:\Code\Matlab\Master\Spring')
color = colorHex();

clf;

 set(0, 'DefaultTextFontName', 'Computer Modern',...
     'DefaultTextFontSize', 10,...
     'DefaultAxesFontName', 'Computer Modern',...
     'DefaultAxesFontSize', 10,...
     'DefaultLineLineWidth', 1.2)

set(gcf,'Position',[450 300 650 580])

figure(1)
subplot(4,1,1), %plot states
    hold on;
    colororder({'k','k'})
    plot(ts, V, 'Color', color.b)
    xlabel('t [h]')
    ylabel('Volume [L]')
    ylim([1.4 1.7])
    yyaxis right
    plot(tF, F_in, 'Color', color.r, 'LineStyle','-.')
    plot(tF, F_out, 'Color', color.p,'LineStyle','--')
    ylabel('L/h')
    %ylim([-0.1 0.1])
    legend({'V: Volume', 'F_{in}', 'F_{out}'}, 'Location', 'NorthWest');%'Position',[0.8 0.9 0.15 0.08]);
    title('Volume')
    xlim([0, ts(end)]);
    

subplot(4,1,2), %plot states
    plot(ts, X, 'Color', color.b)
    yline(X_goal,'--', 'X_{goal}')
    %ylim([0,X_goal+3])
    xlabel('t [h]')
    ylabel('Concentration [g/L]')
    legend({'X: Cell'},'Location', 'NorthWest')%'Position',[0.8 0.7 0.15 0.08]);
    title('Biomass production')
    xlim([0, ts(end)]);
    
subplot(4,1,3), %plot states
    hold on
    
    scatter(ts_S, S, 'Marker', 'x', 'MarkerFaceColor', color.b ,'MarkerEdgeColor', color.b )
    plot(ts_S, S,'Color',color.p, 'LineStyle','-')
    %yline(X_goal,'--', 'X_{goal}')
    %ylim([0,X_goal+3])
    xlabel('t [h]')
    ylabel('Concentration [g/L]')
    legend({'S: Glucose'},'Location', 'NorthWest')%'Position',[0.8 0.7 0.15 0.08]);
    title('Sugar concentration')
    xlim([0, ts(end)]);


 subplot(4,1,4), %plot states
    hold on;
    plot(ts, CO2,'Color', color.b)
    
    xlabel('t [h]')
    ylabel('% of outlet')
    legend({'CO_2'}, 'Location', 'NorthWest')%'Position',[0.8 0.5 0.15 0.08]);
    title('Offgas')
    xlim([0, ts(end)]);
    %xlim([0 58])
% subplot(4,1,4), %plot states
%     hold on;
%     colororder({'k','k'})
%     stairs(tm, us, 'Color', color.r, 'LineWidth',0.9)
%     xlabel('t [h]')
%     ylabel('L/h')
%     yyaxis right
%     ylabel('Cost')
%     plot(tm, cost, '--', 'Color',color.b)
%     legend('Fin', 'Cost', 'Location', 'best')%'Position',[0.8 0.25 0.15 0.08]);
%     title('Feed')
sgtitle(" 15/02/23 Experimental data")